/*
 * C
 *
 * Copyright 2014-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

/* Toolchain includes */
#include <ctype.h>
#include <string.h>

/* wf200 includes */
#include "sl_wfx.h"
#include "sl_wfx_host.h"

/* External includes */
#include "LLECOM_WIFI_impl.h"
#include "NETIF_interface.h"
#include "NET_ADAPT_delegate_task.h"

/* Internal includes */
#include "WIFI_WF200_driver.h"
#include "wlanif.h"

#include "fsl_debug_console.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////Macro definitions//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern sl_wfx_context_t wifi;

extern void vBusCommStart(void);
extern void vBusCommStop(void);

extern int   memcpy_s(void *_Restrict_dst, size_t, const void *_Restric_src, size_t);
extern int   strncpy_s(char *_Restrict_dst, size_t, const char *_Restrict_src, size_t);
extern size_t    strnlen_s(const char *_Restrict, size_t);

/**@brief Define this value to enable assert */
#define ENABLE_ASSERT

/**@brief Assert macro */
#ifdef ENABLE_ASSERT
#define WIFI_WF200_ASSERT(x) do { if (!(x)) { WIFI_WF200_ASSERT_TRACE("%s, %d\n", __FILE__, __LINE__); while(1); } } while (0)
#else
#define WIFI_WF200_ASSERT(x) do { if (!(x)) { WIFI_WF200_ASSERT_TRACE("%s, %d\n", __FILE__, __LINE__); } } while (0)
#endif

/**@brief Log priority levels */

#define WIFI_WF200_LOG_DEBUG      0
#define WIFI_WF200_LOG_INFO       1
#define WIFI_WF200_LOG_WARNING    2
#define WIFI_WF200_LOG_ERROR      3
#define WIFI_WF200_LOG_ASSERT     4
#define WIFI_WF200_LOG_NONE       5

/**@brief Current log level */
#define WIFI_WF200_LOG_LEVEL WIFI_WF200_LOG_NONE  // Do not log anything, printf's go to ETM channel and if there is no debugger attached, hangs the system.

#ifndef WIFI_WF200_LOG_LEVEL
    #error "WIFI_WF200_LOG_LEVEL must be defined"
#endif

/**@brief Debug logger */

#if (WIFI_WF200_LOG_DEBUG >= WIFI_WF200_LOG_LEVEL)
    #define WIFI_WF200_DEBUG_TRACE        PRINTF("[wf200 Wifi Driver][DEBUG] ");PRINTF
#else
    #define WIFI_WF200_DEBUG_TRACE(...)   ((void) 0)
#endif

/**@brief Info logger */

#if (WIFI_WF200_LOG_INFO >= WIFI_WF200_LOG_LEVEL)
    #define WIFI_WF200_INFO_TRACE         PRINTF("[wf200 Wifi Driver][INFO] ");PRINTF
#else
    #define WIFI_WF200_INFO_TRACE(...)    ((void) 0)
#endif

/**@brief Warning logger */

#if (WIFI_WF200_LOG_WARNING >= WIFI_WF200_LOG_LEVEL)
    #define WIFI_WF200_WARNING_TRACE      PRINTF("[wf200 Wifi Driver][WARNING] ");PRINTF
#else
    #define WIFI_WF200_WARNING_TRACE(...) ((void) 0)
#endif

/**@brief Error logger */

#if (WIFI_WF200_LOG_ERROR >= WIFI_WF200_LOG_LEVEL)
    #define WIFI_WF200_ERROR_TRACE        PRINTF("[wf200 Wifi Driver][ERROR] ");PRINTF
#else
    #define WIFI_WF200_ERROR_TRACE(...)   ((void) 0)
#endif

/**@brief Assert logger */

#if (WIFI_WF200_LOG_ASSERT >= WIFI_WF200_LOG_LEVEL)
    #define WIFI_WF200_ASSERT_TRACE       PRINTF("[wf200 Wifi Driver][ASSERT] ");PRINTF
#else
    #define WIFI_WF200_ASSERT_TRACE(...)  ((void) 0)
#endif

/** @brief Max size of the AP scan buffers */
#define WIFI_WF200_MAX_SCAN_BUFFERS         32

/** @brief Timeouts */
#define WF200_DISCONNECT_TIMEOUT_MS             5000
#define WF200_CONNECT_TIMEOUT_MS                5000
#define WF200_START_TIMEOUT_MS                  5000
#define WF200_STOP_TIMEOUT_MS                   5000
#define WF200_SCAN_TIMEOUT_MS                  10000

/**@brief Security mode bitmask*/
#define WFM_SECURITY_MODE_OPEN_BITMASK 			0x00
#define WFM_SECURITY_MODE_WEP_BITMASK			0x01
#define WFM_SECURITY_MODE_WPA_PSK_BITMASK		0x42
#define WFM_SECURITY_MODE_WPA_EAP_BITMASK		0x82
#define WFM_SECURITY_MODE_WPA2_PSK_BITMASK		0x44
#define WFM_SECURITY_MODE_WPA2_EAP_BITMASK		0x84
#define WFM_SECURITY_MODE_WPA2_WPA1_PSK_BITMASK	0x46
#define WFM_SECURITY_MODE_WPA2_WPA1_EAP_BITMASK	0x86

/** @brief Max connections that can be made to the Soft AP */
#define WIFI_SOFT_AP_MAX_CONNECTIONS         4

/** @brief Beacon interval for the Soft AP */
#define WIFI_SOFT_AP_BEACON_INTERVAL       100

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////Type definitions//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/** @brief wf200 connection mode */
typedef enum {
    WIFI_WF200_NOT_INITIALIZED,    /**< Module not initialized */
    WIFI_WF200_IDLE,               /**< Idle mode */
    WIFI_WF200_CONNECTING,         /**< Connecting to an AP */
    WIFI_WF200_CONNECT_OK,         /**< Connected to an AP */
    WIFI_WF200_SOFT_AP_INIT,       /**< Initialized Soft AP */
    WIFI_WF200_SOFT_AP_UP,         /**< Soft AP is up and ready for stations to connect */
    WIFI_WF200_DISCONNECTING       /**< Currently disconnecting */
} WIFI_WF200_connection_state_t;

typedef enum {
    WF200_MODE_NULL = 0,  /**< null mode */
    WF200_MODE_STA,       /**< WiFi station mode */
    WF200_MODE_AP,        /**< WiFi soft-AP mode */
    WF200_MODE_MAX
}WIFI_WF200_mode_t;

/** @brief wf200 driver configuration structure */
typedef struct
{
    WIFI_WF200_connection_state_t connect_state;

} WIFI_WF200_driver_t;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////Global Variables//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/** @brief Private pointer on network interface */
static NETIF_t * gpst_network_interface;

/** @brief Global WF200 driver configuration */
static WIFI_WF200_driver_t g_wf200_driver;

/** @brief List of available access points */
extern sl_wfx_scan_result_ind_body_t gpst_available_ap_list[WIFI_WF200_MAX_SCAN_BUFFERS];

/** @brief Event group for wifi management */
extern EventGroupHandle_t gst_wf200_event_group;

/** @brief store mac address of connected station */
static uint8_t gpuc_sta_mac_addr[SL_WFX_MAC_ADDR_SIZE];

/** @brief store ssid of connected station */
static sl_wfx_ssid_def_t gpuc_sta_ssid;

/** @brief store channel of connected station */
static short gpuc_sta_channel;

/** @brief store rssi of connected station */
static float gpuc_sta_rssi;

/** @brief store security mode of connected station */
static uint8_t gpuc_sta_security_mode;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////Functions Prototypes////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief this function translates the security mode from LLNET mode to wf200 mode
 *
 * @param[in]  _i_llnet_security_mode input security mode
 * @param[out] _wf200_security_type output security type
 * @param[out] _wf200_cipher_type output cipher type
 *
 * @return true if the security mode was converted, false otherwise
 */
static bool convert_llnet_to_wf200_security_mode_f(int _i_llnet_security_mode,
		sl_wfx_security_mode_t * _wf200_security_type);

/**
 * @brief this function initializes the wifi module
 *
 * @return true if success else false
 */
static bool initialize_wifi_f(void);

/**
 * @brief this function joins a WiFi AP or mounts a SoftAP
 *
 * @param[in] _pc_ssid ssid of the network
 * @param[in] _i_ssid_length ssid length
 * @param[in] _pc_passphrase password needed for network connection
 * @param[in] _i_passphrase_length password length
 * @param[in] _i_security_mode security mode
 * @param[in] _mode AP or SoftAP
 *
 * @return true if success else false
 */
static bool join_ap_or_softap_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length,
        int _i_security_mode, WIFI_WF200_mode_t _mode);

/**
 * @brief this function leaves a WiFi network
 *
 * @param[in] mode AP or SoftAP
 *
 * @return true if success else false
 */
static bool leave_ap_or_softap_f(WIFI_WF200_mode_t mode);

/**
 * @brief this function translates the wf200 security mode bitmask to llnet security mode
 *
 * @param[out] _pi_llnet_security_mode output llnet security mode
 * @param[in]  _wf200_security_mode_bitmask input security mode bitmask type
 *
 * @return true if the security mode was converted, false otherwise
 */
static bool convert_wf200_sec_mode_bitmask_to_llnet_security_mode_f(uint8_t _wf200_security_mode_bitmask, int * _pi_llnet_security_mode);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////Private Functions//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool convert_wf200_sec_mode_bitmask_to_llnet_security_mode_f(uint8_t _wf200_security_mode_bitmask,
		int * _pi_llnet_security_mode)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    switch(_wf200_security_mode_bitmask)
	{
	case WFM_SECURITY_MODE_OPEN_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_OPEN;
		break;
	case WFM_SECURITY_MODE_WEP_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_WEP128;
		break;
	case WFM_SECURITY_MODE_WPA_PSK_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_WPA1;
		break;
	case WFM_SECURITY_MODE_WPA_EAP_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_WPA1;
		break;
	case WFM_SECURITY_MODE_WPA2_PSK_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_WPA2;
		break;
	case WFM_SECURITY_MODE_WPA2_EAP_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_WPA2;
		break;
	case WFM_SECURITY_MODE_WPA2_WPA1_PSK_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_MIXED;
		break;
	case WFM_SECURITY_MODE_WPA2_WPA1_EAP_BITMASK:
		*_pi_llnet_security_mode = SECURITY_MODE_MIXED;
		break;
    default:
        /* Not supported security mode */
        WIFI_WF200_ERROR_TRACE("Security mode bitmask not supported (bitmask = %d)\n", _wf200_security_mode_bitmask);
        return false;
	}
	return true;
}

static bool convert_llnet_to_wf200_security_mode_f(int _i_llnet_security_mode,
		sl_wfx_security_mode_t * _wf200_security_type)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    switch (_i_llnet_security_mode)
    {
        case SECURITY_MODE_OPEN:
            *_wf200_security_type = WFM_SECURITY_MODE_OPEN;
        break;
        case SECURITY_MODE_WEP64:
            *_wf200_security_type = WFM_SECURITY_MODE_WEP;
        break;
        case SECURITY_MODE_WEP128:
            *_wf200_security_type = WFM_SECURITY_MODE_WEP;
        break;
        case SECURITY_MODE_WPA1:
            *_wf200_security_type = WFM_SECURITY_MODE_WPA2_WPA1_PSK; //Use WPA1 or WPA2
        break;
        case SECURITY_MODE_WPA2:
            *_wf200_security_type = WFM_SECURITY_MODE_WPA2_PSK;
        break;
        case SECURITY_MODE_MIXED:
            *_wf200_security_type = WFM_SECURITY_MODE_WPA2_WPA1_PSK;
        break;

        default:
            /* Not supported security mode */
            WIFI_WF200_ERROR_TRACE("Security mode not supported (mode = %d)\n", _i_llnet_security_mode);
            return false;
            break;
    }

    return true;
}

static bool initialize_wifi_f(void)
{
    sl_status_t status;

    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Initialize wifi module */
    status = sl_wfx_init(&wifi);

	switch(status) {
	case SL_STATUS_OK:
		WIFI_WF200_DEBUG_TRACE("WF200 init successful\r\n");
		return true;
		break;
    case SL_STATUS_WIFI_INVALID_KEY:
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Firmware keyset invalid\r\n");
		return false;
		break;
    case SL_STATUS_WIFI_FIRMWARE_DOWNLOAD_TIMEOUT:
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Firmware download timeout\r\n");
		return false;
		break;
    case SL_STATUS_TIMEOUT:
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Poll for value timeout\r\n");
		return false;
		break;
    case SL_STATUS_FAIL:
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Error\r\n");
		return false;
		break;
    default :
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Unknown error\r\n");
		return false;
        break;
    }
}
static bool deinit_wifi_f(void)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    if(sl_wfx_deinit() != SL_STATUS_OK)
	{
		WIFI_WF200_ERROR_TRACE("Failed to init WF200: Unknown error\r\n");
		return false;
	}
	return true;
}
static bool join_ap_or_softap_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length,
        int _i_security_mode, WIFI_WF200_mode_t _mode)
{

    sl_wfx_security_mode_t authmode;
    sl_status_t status;
    EventBits_t uxBits;

    int ssid_len;
    int password_len;
    int i = 0;

    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    ssid_len = strlen(_pc_ssid);
    password_len = strlen(_pc_passphrase);

    /* Check wifi mode */
    if (_mode == WF200_MODE_STA)
    {
        // Some Java applications (Ex. AWS IoT) do issue join without scanning first.
        // WF200 requires scan and then join, therefore, issue scan first.
        short my_ps_ap_count = 0;
        if ( WIFI_WF200_get_ap_count_f( &my_ps_ap_count ) == false )
        {
            return false;
        }

        /* Convert security mode */
        if (!convert_llnet_to_wf200_security_mode_f(_i_security_mode, &authmode))
        {
            WIFI_WF200_ERROR_TRACE("Cannot convert security mode\n");
            return false;
        }
        /* Set the currnet connection status */
        g_wf200_driver.connect_state = WIFI_WF200_CONNECTING;

        if(sl_wfx_send_join_command((uint8_t const *)_pc_ssid, ssid_len, NULL, UNKNOW_CHANNEL, authmode, PREVENT_ROAMING,
                FRAME_PROTECTION, (uint8_t const *)_pc_passphrase, password_len, NULL, 0) != SL_STATUS_OK)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Cannot join the specified Wi-Fi network\n");
			return false;
        }
        uxBits = xEventGroupWaitBits(gst_wf200_event_group, SL_WFX_CONNECT_IND_ID, pdTRUE, pdTRUE, WF200_CONNECT_TIMEOUT_MS / portTICK_PERIOD_MS);

        if((uxBits & SL_WFX_CONNECT_IND_ID) == 0)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Join the specified Wi-Fi network failed with TIMEOUT\n");
			/*If join fails and in order to do another softap mount first a disconnect has to be done*/
			leave_ap_or_softap_f(WF200_MODE_STA);
			return false;
        }else
        {
            /*set and link up the interface for the sta mode*/
            set_link_up(STATION_MODE);
        }

        /* Set the current connection status */
        g_wf200_driver.connect_state = WIFI_WF200_CONNECT_OK;

        for(i = 0; i <= WIFI_WF200_MAX_SCAN_BUFFERS; i++)
        {
            // This store the string length. This is different from _i_ssid_length that store the
            // Buffer size. ssid_strlen < _i_ssid_length.
            size_t ssid_strlen = strnlen_s(_pc_ssid, _i_ssid_length);
            if(0 == ssid_strlen)
            {
                WIFI_WF200_WARNING_TRACE("[%s] SSID length is zero\n", __func__);
            }
            //First check that size match
            if( ssid_strlen == gpst_available_ap_list[i].ssid_def.ssid_length)
            {
                /*Check if the joined network is between scaned access points. If so, store the access point info.*/
                if( 0 == memcmp((char const*)gpst_available_ap_list[i].ssid_def.ssid,
                                _pc_ssid,
                                gpst_available_ap_list[i].ssid_def.ssid_length))
                {
                                WIFI_WF200_DEBUG_TRACE("Found wifi to join:%s\n", _pc_ssid);
                    if(memcpy_s((char*)gpuc_sta_ssid.ssid,
                                 sizeof(gpuc_sta_ssid.ssid),
                                 (char const *)gpst_available_ap_list[i].ssid_def.ssid,
                                 sizeof(gpuc_sta_ssid.ssid)))
                    {
                        WIFI_WF200_ERROR_TRACE("Copy ssid fail at %s %d\n", __FILE__, __LINE__);
                    }
                    gpuc_sta_ssid.ssid_length = gpst_available_ap_list[i].ssid_def.ssid_length;
                    memcpy(gpuc_sta_mac_addr, gpst_available_ap_list[i].mac, SL_WFX_MAC_ADDR_SIZE);
                    gpuc_sta_channel = gpst_available_ap_list[i].channel;
                    gpuc_sta_rssi = (int16_t)(((gpst_available_ap_list[i].rcpi)/2)-110);
                    gpuc_sta_security_mode = *(uint8_t*)(&gpst_available_ap_list[i].security_mode);
                    break;
                }
            }
        }

        //Check that we found the network with the matching SSID.
        if(WIFI_WF200_MAX_SCAN_BUFFERS == i)
        {
            WIFI_WF200_ERROR_TRACE("No wifi found with SSID %s\n", _pc_ssid);
            leave_ap_or_softap_f(_mode);
        }
    }
    else if (_mode == WF200_MODE_AP)
    {
        /* Convert security mode */
        if (!convert_llnet_to_wf200_security_mode_f(_i_security_mode, &authmode))
        {
            WIFI_WF200_ERROR_TRACE("Cannot convert security mode\n");
            return false;
        }
        /* Set the current connection status for Soft-AP*/
        g_wf200_driver.connect_state = WIFI_WF200_SOFT_AP_INIT;
        status = sl_wfx_start_ap_command(6, (uint8_t *)_pc_ssid, ssid_len, HIDDEN_SSID_AP, CLIENT_ISOLATION, authmode, WFM_MGMT_FRAME_PROTECTION_OPTIONAL,
													(uint8_t *) _pc_passphrase, password_len, NULL, 0, NULL,0);
        if (status != SL_STATUS_OK)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
            WIFI_WF200_ERROR_TRACE("Cannot start the Soft-AP mode\n");
            return false;
        }
        uxBits = xEventGroupWaitBits(gst_wf200_event_group, SL_WFX_START_AP_IND_ID, pdTRUE, pdTRUE, WF200_START_TIMEOUT_MS / portTICK_PERIOD_MS);

        if((uxBits & SL_WFX_START_AP_IND_ID) == 0)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Soft-AP start failed\n");
			return false;
        }else
        {
            /*set and link up the interface for ap mode*/
            set_link_up(SOFTAP_MODE);
        }
        /* Set the current connection status for Soft-AP*/
        g_wf200_driver.connect_state = WIFI_WF200_SOFT_AP_INIT;
    }
    else
    {
        WIFI_WF200_ERROR_TRACE("Unsupported mode parameter = %d\n", _mode);
        return false;
    }
    return true;
}

static bool leave_ap_or_softap_f(WIFI_WF200_mode_t mode)
{

	sl_status_t status = SL_STATUS_FAIL;
    EventBits_t uxBits;

    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Check the requested mode */
    if (mode == WF200_MODE_STA)
    {
        /* Set the current connection status */
        g_wf200_driver.connect_state = WIFI_WF200_DISCONNECTING;

        /* Disconnect from AP */
		status = sl_wfx_send_disconnect_command();

        if(status != SL_STATUS_OK)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Disconnecting from AP failed\n");
			return false;
        }
        uxBits = xEventGroupWaitBits(gst_wf200_event_group, SL_WFX_DISCONNECT_IND_ID, pdTRUE, pdTRUE, WF200_DISCONNECT_TIMEOUT_MS / portTICK_PERIOD_MS);

        if((uxBits & SL_WFX_DISCONNECT_IND_ID) == 0)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Failed to disconnect from Wi-Fi network!\n");
			return false;
        }else
        {
            /*set and link down the interface for sta mode*/
            set_link_down(STATION_MODE);
        }

		/*If disconnected from the access point clear the info.*/
        memset(gpuc_sta_mac_addr, 0x00, sizeof gpuc_sta_mac_addr);
        memset(&gpuc_sta_ssid, 0x00, sizeof gpuc_sta_ssid);
        gpuc_sta_channel = 0;
        gpuc_sta_rssi = 0;
        gpuc_sta_security_mode = 0;
    }
    else if (mode == WF200_MODE_AP)
    {
        /* Stop the Soft-AP */
        if (sl_wfx_stop_ap_command() != SL_STATUS_OK)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
            WIFI_WF200_ERROR_TRACE("Stopping Soft-AP failed\n");
            return false;
        }
        uxBits = xEventGroupWaitBits(gst_wf200_event_group, SL_WFX_STOP_AP_IND_ID, pdTRUE, pdTRUE, WF200_STOP_TIMEOUT_MS / portTICK_PERIOD_MS);

        if((uxBits & SL_WFX_STOP_AP_IND_ID) == 0)
        {
			g_wf200_driver.connect_state = WIFI_WF200_IDLE;
			WIFI_WF200_ERROR_TRACE("Soft AP failed to disconnect!\n");
			return false;
        }else
        {
            /*set and link down the interface for ap mode*/
            set_link_down(SOFTAP_MODE);
        }

    }
    else
    {
        WIFI_WF200_ERROR_TRACE("Unsupported mode parameter = %d\n", mode);
        g_wf200_driver.connect_state = WIFI_WF200_IDLE;
        return false;
    }

    /* Set the current connection status */
    g_wf200_driver.connect_state = WIFI_WF200_IDLE;

    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////Public Functions///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

NETIF_status_t WIFI_WF200_init_f(void)
{
    NETIF_status_t e_return = NETIF_ERROR;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    // Start WF200 communication task.  This task is defined in SiLabs WF200 driver.
    vBusCommStart();

//#define WF200_HW_VERIFICATION
#ifdef WF200_HW_VERIFICATION

    // This section is only used for initial hardware testing.  It will be removed
    // from release/production code.
    sl_wfx_context_t sl_wfx_context = { 0 };
    sl_wfx_init( &sl_wfx_context );
    
    // If scan command is sent successfully, sl_wfx_scan_result_callback should 
    // be called.
    sl_wfx_send_scan_command(WFM_SCAN_MODE_ACTIVE, NULL, 0, NULL, 0, NULL, 0, NULL) ;

    return 0;

#endif

    /* Initialize the driver state */
    g_wf200_driver.connect_state = WIFI_WF200_NOT_INITIALIZED;

    /* Register interface among netif module */
   if (NETIF_get_registered_interface_f(&gpst_network_interface) != NETIF_ERROR)
    {
		/* Update interface name */
		strncpy(gpst_network_interface->pc_name, STA_WLANIF_NAME, NETIF_MAX_NAME_LENGTH);

		/* Update flag interface */
		gpst_network_interface->puc_flag[NETIF_IS_UP] = 0;
		gpst_network_interface->puc_flag[NETIF_IS_LOOPBACK] = 0;
		gpst_network_interface->puc_flag[NETIF_IS_POINT_TO_POINT] = 0;

		/* Register callback for network interface */
		gpst_network_interface->pp_apply_configuration_f = WIFI_WF200_apply_configuration_f;
		gpst_network_interface->pp_start_f               = WIFI_WF200_start_f;
		gpst_network_interface->pp_stop_f                = WIFI_WF200_stop_f;
		gpst_network_interface->pp_initialize_f          = WIFI_WF200_initialize_f;
		gpst_network_interface->pp_is_started_f          = WIFI_WF200_netif_is_started_f;

		/* Initialization OK */
	   e_return = NETIF_NO_ERROR;
    }
    else
    {
        WIFI_WF200_ERROR_TRACE("Error registering network interface\n");
    }

    return e_return;
}

void WIFI_WF200_deinit_f(void)
{
	// Stop WF200 communication task.  This task is defined in SiLabs WF200 driver.
	vBusCommStop();
}

NETIF_status_t WIFI_WF200_initialize_f(void)
{
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Initialize buffers */
    memset(gpst_available_ap_list, 0x0, sizeof gpst_available_ap_list);

    /* Initialize wifi module */
    if (!initialize_wifi_f())
    {
        WIFI_WF200_ERROR_TRACE("Initialize wifi failed\n");
        return NETIF_ERROR;
    }

    /* Set the current connection status */
    g_wf200_driver.connect_state = WIFI_WF200_IDLE;

    return NETIF_NO_ERROR;
}
NETIF_status_t WIFI_WF200_apply_configuration_f(void)
{

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Take entire configuration from network interface and apply it */
	/*TO DO. The WF200 driver does not offer an API or support to apply the configuration for static IP
	 * or to use an DHCP client.
	 * */
    return NETIF_NO_ERROR;
}

NETIF_status_t WIFI_WF200_start_f(void)
{
    //WIFI_WF200_ASSERT((g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED) || (g_wf200_driver.connect_state == WIFI_WF200_IDLE));
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Start the module if it was not started before or if it was stopped */
    if (g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED)
    {
        /* Start the wifi module in default configuration */
        if (!initialize_wifi_f())
        {
            WIFI_WF200_ERROR_TRACE("Start wf200 failed\n");
            return NETIF_ERROR;
        }

        /* Set the current connection status */
        g_wf200_driver.connect_state = WIFI_WF200_IDLE;
    }

    return NETIF_NO_ERROR;
}

NETIF_status_t WIFI_WF200_stop_f(void)
{
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    /* Stop wifi module */
    if (!deinit_wifi_f())
    {
        WIFI_WF200_ERROR_TRACE("Stopping wifi failed\n");
        return NETIF_ERROR;
    }

    /* Module must be initialized again for further usage */
    g_wf200_driver.connect_state = WIFI_WF200_NOT_INITIALIZED;

    return NETIF_NO_ERROR;
}

NETIF_status_t WIFI_WF200_netif_is_started_f(unsigned char * _puc_netif_started)
{
    WIFI_WF200_ASSERT(_puc_netif_started != NULL);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    if (g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED)
    {
        *_puc_netif_started = 0;
    }
    else
    {
        *_puc_netif_started = 1;
    }

    return NETIF_NO_ERROR;
}

bool WIFI_WF200_is_started_f(void)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    if (g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED)
    {
        return false;
    }

    return true;
}

bool WIFI_WF200_join_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length, int _i_security_mode)
{
    WIFI_WF200_DEBUG_TRACE("(%s) ssid = %s\n", __func__, _pc_ssid);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pc_ssid != NULL);
    WIFI_WF200_ASSERT(_i_ssid_length > 0);
    WIFI_WF200_ASSERT(_pc_passphrase != NULL);
    WIFI_WF200_ASSERT(_i_passphrase_length > 0);
    WIFI_WF200_ASSERT(_i_security_mode >= 0);

    return join_ap_or_softap_f(_pc_ssid, _i_ssid_length, _pc_passphrase, _i_passphrase_length, _i_security_mode, WF200_MODE_STA);
}

bool WIFI_WF200_leave_f(void)
{
	 WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);

    return leave_ap_or_softap_f(WF200_MODE_STA);
}

bool WIFI_WF200_get_bssid_f(unsigned char * _puc_bssid, int _i_bssid_length)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_puc_bssid != NULL);
    WIFI_WF200_ASSERT(_i_bssid_length >= sizeof(gpst_available_ap_list[0].mac));

    if(g_wf200_driver.connect_state != WIFI_WF200_CONNECT_OK)
    {
		WIFI_WF200_ERROR_TRACE("WF200 is not connected to a network!\n");
		return false;
    }
    memcpy(_puc_bssid, gpuc_sta_mac_addr, _i_bssid_length);

    return true;
}

bool WIFI_WF200_get_channel_f(short * _ps_channel)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_ps_channel != NULL);

    if(g_wf200_driver.connect_state != WIFI_WF200_CONNECT_OK)
    {
		WIFI_WF200_ERROR_TRACE("WF200 is not connected to a network!\n");
		return false;
    }

    *_ps_channel = gpuc_sta_channel;

    return true;
}

bool WIFI_WF200_get_rssi_f(float * _pf_rssi, int _i_rssi_length)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pf_rssi != NULL);

    if(g_wf200_driver.connect_state != WIFI_WF200_CONNECT_OK)
    {
		WIFI_WF200_ERROR_TRACE("WF200 is not connected to a network!\n");
		return false;
    }

    *_pf_rssi = gpuc_sta_rssi;

    return true;
}

bool WIFI_WF200_get_security_mode_f(short * _ps_mode)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_ps_mode != NULL);
    bool b_return = true;

    if(g_wf200_driver.connect_state != WIFI_WF200_CONNECT_OK)
    {
		WIFI_WF200_ERROR_TRACE("WF200 is not connected to a network!\n");
		return false;
    }

    b_return = convert_wf200_sec_mode_bitmask_to_llnet_security_mode_f(gpuc_sta_security_mode, (int *)_ps_mode);
	if(!b_return)
	{
		WIFI_WF200_ERROR_TRACE("Failed to convert the wf200 security mode bitmask to llnet security type\n");
		return b_return;
	}

    return b_return;
}

bool WIFI_WF200_get_ssid_f(char * _pc_ssid, int _i_ssid_length)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pc_ssid != NULL);

    if(g_wf200_driver.connect_state != WIFI_WF200_CONNECT_OK)
    {
		WIFI_WF200_ERROR_TRACE("WF200 is not connected to a network!\n");
		return false;
    }

	if(strncpy_s(_pc_ssid,
                _i_ssid_length,
                (char const *)gpuc_sta_ssid.ssid,
                _i_ssid_length - 1))
    {
        WIFI_WF200_ERROR_TRACE("Copy ssid fail at %s\n", __func__);
        return false;
    }

    return true;
}

bool WIFI_WF200_get_wps_modes_f(short * _ps_wps_modes)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_ps_wps_modes != NULL);

    /* Not supported by the wf200 APIs */
    *_ps_wps_modes = WPS_MODE_NONE;
    return true;
}

bool WIFI_WF200_enable_softap_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pc_ssid != NULL);
    WIFI_WF200_ASSERT(_i_ssid_length > 0);
    WIFI_WF200_ASSERT(_pc_passphrase != NULL);
    WIFI_WF200_ASSERT(_i_passphrase_length > 0);
    WIFI_WF200_DEBUG_TRACE("(%s) ssid = %s\n", __func__, _pc_ssid);

    return join_ap_or_softap_f(_pc_ssid, _i_ssid_length, _pc_passphrase, _i_passphrase_length, SECURITY_MODE_WPA2, WF200_MODE_AP);
}

bool WIFI_WF200_set_name_softap_f(const char * _pc_name, int _i_name_length)
{
	WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pc_name != NULL);
    WIFI_WF200_DEBUG_TRACE("(%s) ssid = %s\n", __func__, _pc_name);

    /* Not supported by the wf200 APIs */

    return true;
}

bool WIFI_WF200_disable_softap_f(void)
{
	WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);
    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    return leave_ap_or_softap_f(WF200_MODE_AP);
}

bool WIFI_WF200_get_ap_count_f(short * const _ps_ap_count)
{
	WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    EventBits_t uxBits;
    int index = 0;

    if ( g_wf200_driver.connect_state == WIFI_WF200_NOT_INITIALIZED )
    {
       return false;
    }
    
    WIFI_WF200_ASSERT(_ps_ap_count != NULL);

    /* Clear buffers */
    memset((void *) &gpst_available_ap_list, 0, sizeof(gpst_available_ap_list));

    if(sl_wfx_send_scan_command(WFM_SCAN_MODE_ACTIVE, NULL, 0, NULL, 0, NULL, 0, NULL) == SL_STATUS_FAIL)
    {
        WIFI_WF200_ERROR_TRACE("Error while sending scan network command\n");
        return false;
    }

    uxBits = xEventGroupWaitBits(gst_wf200_event_group, SL_WFX_SCAN_COMPLETE_IND_ID, pdTRUE, pdTRUE, WF200_SCAN_TIMEOUT_MS / portTICK_PERIOD_MS);
    
    if ((uxBits & SL_WFX_SCAN_COMPLETE_IND_ID) != 0)
    {
        for(index = 0; index < WIFI_WF200_MAX_SCAN_BUFFERS; index++)
        {
			if (gpst_available_ap_list[index].channel != 0)
			{
				WIFI_WF200_DEBUG_TRACE( "# %2d %2d  %03d %02X:%02X:%02X:%02X:%02X:%02X  %s  %u\n",
						index, gpst_available_ap_list[index].channel,
						((int16_t)(gpst_available_ap_list[index].rcpi - 220)/2),
						gpst_available_ap_list[index].mac[0], gpst_available_ap_list[index].mac[1],
						gpst_available_ap_list[index].mac[2], gpst_available_ap_list[index].mac[3],
						gpst_available_ap_list[index].mac[4], gpst_available_ap_list[index].mac[5],
						gpst_available_ap_list[index].ssid_def.ssid,
						*(uint8_t*)(&gpst_available_ap_list[index].security_mode));
			}else
			{
                if(0 == index)
                {
                    // If first network is empty it means no network detected.
                    WIFI_WF200_INFO_TRACE("No network detected\n");
                }
				break;
			}

        }
        if(WIFI_WF200_MAX_SCAN_BUFFERS == index)
        {
			WIFI_WF200_WARNING_TRACE("Ap scan list is full! %d wifi network detected\n", WIFI_WF200_MAX_SCAN_BUFFERS);
        }

        *_ps_ap_count = (short) (index);
    }
    
    // Timeout on xEventGroupWaitBits
    else
    {
        WIFI_WF200_WARNING_TRACE("SL_WFX_SCAN_COMPLETE_IND_ID was not set. xEventGroupWaitBits return event 0x%x\n", uxBits);
        return false;
    }

    return true;
}

bool WIFI_WF200_get_ap_ssid_f(unsigned int _ui_index, unsigned char * _puc_ssid, int _i_ssid_length)
{
	WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
    bool b_return;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_puc_ssid != NULL);

    if ((_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0]))) && (_i_ssid_length > 0))
    {
        b_return = true;

        if(memcpy_s((char *) _puc_ssid,
                     _i_ssid_length,
                     (const char *)(gpst_available_ap_list[_ui_index].ssid_def.ssid),
                     _i_ssid_length - 1))
        {
            WIFI_WF200_ERROR_TRACE("Copy ssid fail at %s line %d\n", __FILE__, __LINE__);
            b_return = false;
        }
        _puc_ssid[_i_ssid_length - 1] = '\0';

    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_ap_bssid_f(unsigned int _ui_index, unsigned char * _puc_bssid, int _i_bssid_length)
{
	WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
    bool b_return;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_puc_bssid != NULL);

    if ((_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0]))) && (_i_bssid_length > 0))
    {
        int copy_length;

        b_return = true;

        if ((unsigned int) _i_bssid_length < sizeof(gpst_available_ap_list[_ui_index].mac))
        {
            copy_length = _i_bssid_length;
        }
        else
        {
            copy_length = sizeof(gpst_available_ap_list[_ui_index].mac) - 1;
        }

        strncpy((char *) _puc_bssid, (char *) gpst_available_ap_list[_ui_index].mac, (size_t) copy_length);
    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_ap_channel_f(unsigned int _ui_index, int * _pi_channel)
{
	WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
	bool b_return = true;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pi_channel != NULL);


    if (_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0])))
    {
        b_return = true;

        *_pi_channel = gpst_available_ap_list[_ui_index].channel;
    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_ap_rssi_f(unsigned int _ui_index, float * _pf_rssi, int _i_rssi_length)
{
    WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
    bool b_return = true;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pf_rssi != NULL);

    if ((_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0]))) && (_i_rssi_length > 0))
    {
        b_return = true;

        *_pf_rssi = (gpst_available_ap_list[_ui_index].rcpi/2) - 110; /*Equation to get the signal level in dBm*/
    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_ap_security_mode_f(unsigned int _ui_index, int * _pi_mode)
{
    WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
    bool b_return = true;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pi_mode != NULL);

    if (_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0])))
    {
		uint8_t ap_security_mode_bitmask = *(uint8_t*)(&gpst_available_ap_list[_ui_index].security_mode);
		b_return = convert_wf200_sec_mode_bitmask_to_llnet_security_mode_f(ap_security_mode_bitmask, _pi_mode);
		if(!b_return)
		{
			WIFI_WF200_ERROR_TRACE("Failed to convert the wf200 security mode bitmask to llnet security type\n");
			return b_return;
		}
    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_ap_wps_modes_f(unsigned int _ui_index, int * _pi_wps_cap)
{
    WIFI_WF200_DEBUG_TRACE("(%s) index = %u\n", __func__, _ui_index);
    bool b_return = true;

    WIFI_WF200_ASSERT(g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED);
    WIFI_WF200_ASSERT(_pi_wps_cap != NULL);


    if (_ui_index < (sizeof(gpst_available_ap_list) / sizeof(gpst_available_ap_list[0])))
    {
        b_return = true;

        *_pi_wps_cap = WPS_MODE_NONE;
    }
    else
    {
        b_return = false;
    }

    return b_return;
}

bool WIFI_WF200_get_client_state_f(void)
{
    WIFI_WF200_DEBUG_TRACE("(%s)\n", __func__);

    if (g_wf200_driver.connect_state != WIFI_WF200_NOT_INITIALIZED)
    {
        return true;
    }
    else
    {
        return false;
    }
}
