/*
 * C
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

/**
 * @file fault_handlers.c
 *
 * @brief This file contains the fault exception handlers.
 *
 * @details The CMSIS names are used.
 *
 * There are two modes:
 * - verbose
 * - lite
 *
 * In verbose mode, each fault exception handler will:
 * - print the stack frame,
 * - print its name,
 * - print the associated fault status register,
 * - try to make an analysis of the fault cause,
 * - make an infinite loop.
 *
 * In lite mode, each fault exception handler will:
 * - print its name,
 * - make an infinite loop.
 *
 * Define the @ref VERBOSE_MODE macro to 1 to activate verbose mode.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>

#include "MIMXRT595S_cm33.h"
#include "LLMJVM.h"

#include "fsl_debug_console.h"

#include "fault_handlers.h" // will check prototypes and import public types

#include "mej_log.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/**
 * @brief Define this macro to if you are using Keil's ARMCC toolchain. If you are using IAR or GCC toolchains, do not define it.
 */
#define COMPILE_WITH_ARMCC

/**
 * @brief Define this macro to enable verbose handlers. Verbose handlers will consume  more memory than lite handlers.
 */
#define VERBOSE_MODE 1

/**
 * @brief Make an infinite while loop.
 */
#define INFINITE_LOOP()		{while(1);}

// CMSIS doesn't provide macro to get bits from BFSR (which is the 2nd byte in CFSR), so we create our own macros
#define BFSR_IBUSERR_Pos			(0)
#define BFSR_IBUSERR_Msk			(1 << BFSR_IBUSERR_Pos)

#define BFSR_PRECISERR_Pos			(1)
#define BFSR_PRECISERR_Msk			(1 << BFSR_PRECISERR_Pos)

#define BFSR_IMPRECISERR_Pos		(2)
#define BFSR_IMPRECISERR_Msk		(1 << BFSR_IMPRECISERR_Pos)

#define BFSR_BFARVALID_Pos			(15)
#define BFSR_BFARVALID_Msk			(1 << BFSR_BFARVALID_Pos)

// CMSIS doesn't provide macro to get bits from UFSR (which is the upper half-word in CFSR), so we create our own macros
#define UFSR_UNDEFINSTR_Pos			(0)
#define UFSR_UNDEFINSTR_Msk			(1 << UFSR_UNDEFINSTR_Pos)

#define UFSR_INVSTATE_Pos			(1)
#define UFSR_INVSTATE_Msk			(1 << UFSR_INVSTATE_Pos)

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

#if VERBOSE_MODE == 1

/**
 * @brief C function to print the stacked registers (== the stack frame) along with EXEC_RETURN.
 *
 * @warning this function cannot be static because it won't be seen by assembly code during link edition.
 * You should not called this function directly.
 */
static void printer(context_state_frame_t *frame, uint32_t msp, uint32_t psp);

#endif // VERBOSE_MODE == 1

// -----------------------------------------------------------------------------
// Project functions
// -----------------------------------------------------------------------------

#if VERBOSE_MODE == 1

// See the header file for the function documentation
void handle_hard_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	uint32_t hfsr;
	printer(frame, msp, psp);

	hfsr = SCB->HFSR;
	MEJ_LOG_ERROR("Hard Fault Status Register =\t%X", hfsr);

	if(hfsr & SCB_HFSR_FORCED_Msk) {
		MEJ_LOG_ERROR("This hard fault is: FORCED");
	}

	LLMJVM_dump();
	LLMJVM_checkIntegrity();

	volatile int i = 42;

	while (i == 42);
}

// See the header file for the function documentation
void handle_memory_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	printer(frame, msp, psp);

	INFINITE_LOOP();
}

// See the header file for the function documentation
void handle_bus_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	uint32_t cfsr;
	uint8_t bfsr;
	uint32_t bfar;
	bool still_valid;
	printer(frame, msp, psp);

	cfsr = SCB->CFSR;
	MEJ_LOG_ERROR("Configurable Fault Status Register =\t%.8X", cfsr);

	bfsr = (cfsr & SCB_CFSR_BUSFAULTSR_Msk) >> SCB_CFSR_BUSFAULTSR_Pos;
	MEJ_LOG_ERROR("Bus Fault Status Register =\t%.2X", bfsr);

	if(bfsr & BFSR_IBUSERR_Msk) {
		MEJ_LOG_ERROR("Instruction access error");
		// This case has not been experimented yet
	}

	if(bfsr & BFSR_PRECISERR_Msk) {
		MEJ_LOG_ERROR("Precise data access error");

		// Faulting instruction
		// --> see the PC of the printed stack frame

		// Address of faulting data access
		bfar = SCB->BFAR;
		MEJ_LOG_ERROR("Address of faulting data access (BFAR) = %.8X", bfar);

		still_valid = (cfsr & BFSR_BFARVALID_Msk);
		if(!still_valid) {
			MEJ_LOG_WARNING("BFAR is no longer valid!");
		}
	}

	if(bfsr & BFSR_IMPRECISERR_Msk) {
		MEJ_LOG_ERROR("Imprecise data access error");
	}

	INFINITE_LOOP();
}

// See the header file for the function documentation
void handle_usage_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	uint32_t cfsr;
	uint16_t ufsr;
	// TODO Activate usage for for division by 0 and unaligned access? (see page 384)

	printer(frame, msp, psp);

	cfsr = SCB->CFSR;
	MEJ_LOG_ERROR("Configurable Fault Status Register =\t%X", cfsr);

	ufsr = (cfsr & SCB_CFSR_USGFAULTSR_Msk) >> SCB_CFSR_USGFAULTSR_Pos;
	MEJ_LOG_ERROR("Usage Fault Status Register =\t%X", ufsr);

	if(ufsr & UFSR_UNDEFINSTR_Msk) {
		MEJ_LOG_ERROR("Attempt to execute an undefined instruction");
	}

	if(ufsr & UFSR_INVSTATE_Msk) {
		MEJ_LOG_ERROR("Attempt to switch to invalid mode (like ARM state)");
	}

	INFINITE_LOOP();
}

#else // VERBOSE_MODE == 1

// See the header file for the function documentation
void handle_hard_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	MEJ_LOG_ERROR("Hard Fault");
	INFINITE_LOOP();
}


// See the header file for the function documentation
void handle_memory_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	MEJ_LOG_ERROR("Memory Fault");
	INFINITE_LOOP();
}


// See the header file for the function documentation
void handle_bus_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	MEJ_LOG_ERROR("Bus Fault");
	INFINITE_LOOP();
}


// See the header file for the function documentation
void handle_usage_fault(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	MEJ_LOG_ERROR("Usage Fault");
	INFINITE_LOOP();
}

#endif // VERBOSE_MODE == 1

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

#if VERBOSE_MODE == 1

// See the section 'Internal function definitions' for the function documentation
static void printer(context_state_frame_t *frame, uint32_t msp, uint32_t psp) {
	MEJ_LOG_INFO("---------------------------------------------------------------------");
	// Show stack pointers
	MEJ_LOG_INFO("Current SP = 0x%08X", frame);
	MEJ_LOG_INFO("MSP = 0x%08X", msp);
	MEJ_LOG_INFO("PSP = 0x%08X", psp);
	MEJ_LOG_INFO("");

	// Show stacked registers == stack frame (see section 8.1.3 and figure 12.4)
	MEJ_LOG_INFO("Stack frame:");
	MEJ_LOG_INFO("R0 =\t0x%08X", frame->r0);
	MEJ_LOG_INFO("R1 =\t0x%08X", frame->r1);
	MEJ_LOG_INFO("R2 =\t0x%08X", frame->r2);
	MEJ_LOG_INFO("R3 =\t0x%08X", frame->r3);
	MEJ_LOG_INFO("R12 =\t0x%08X", frame->r12);
	MEJ_LOG_INFO("LR =\t0x%08X", frame->lr);
	MEJ_LOG_INFO("");

	MEJ_LOG_INFO("---------------------------------------------------------------------");
}

#endif // VERBOSE_MODE == 1

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

