/*
 * C
 *
 * Copyright 2014-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

/**
 * @file
 * @brief LLNET_CHANNEL 2.1.0 implementation over BSD-like API.
 * @author MicroEJ Developer Team
 * @version 1.1.1
 * @date 7 February 2020
 */

#include <LLNET_CHANNEL_impl.h>

#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "LLNET_CONSTANTS.h"
#include "async_select.h"
#include "LLNET_ERRORS.h"
#include "LLNET_Common.h"
#if LLNET_AF & LLNET_AF_IPV6
#include <ifaddrs.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#endif
#include "WIFI_WF200_driver.h"
#include "fsl_trng.h"

#ifdef LLNET_IGNORE_SIGPIPE
#include <signal.h>
#endif

#ifdef __cplusplus
	extern "C" {
#endif


#define NET_DELEGATE_TASK_PRIORITY		    (12) /** Should be > tskIDLE_PRIORITY & < configTIMER_TASK_PRIORITY */
#define NET_DELEGATE_STACK_SIZE				( 8192 / sizeof( StackType_t ) )


/**
 * @brief main function of net manager task
 */
void xNetDelegateTaskFunction(void * pvParameters)
{
	NET_run_f();
}

int32_t LLNET_CHANNEL_IMPL_initialize_custom(void)
{

    trng_config_t   trngConfig;
 
    /* Initialize TRNG. Used TCP/IP stack to generate random port numbers. */
    RESET_PeripheralReset(kRNG_RST_SHIFT_RSTn);
    TRNG_GetDefaultConfig(&trngConfig);
    /* Set sample mode of the TRNG ring oscillator to Von Neumann, for better random data.
     * It is optional.*/
    trngConfig.sampleMode = kTRNG_SampleModeVonNeumann;
    if ( kStatus_Success != TRNG_Init(TRNG0, &trngConfig))
    {
 	    PRINTF("-E- TRNG0 init failed.\r\n");
 	   return SNI_IGNORED_RETURNED_VALUE;
    }
#ifdef LLNET_DEBUG    
    else
    {
        PRINTF("TRNG0 init success.\r\n");
    }
#endif
    
	if(NETIF_NO_ERROR != NETIF_init_f()){
		return J_NETWORK_NOT_INITIALIZED;
	}

	if(NET_NO_ERROR != NET_init_f()){
		return J_NETWORK_NOT_INITIALIZED;
	}

	if(NETIF_NO_ERROR != WIFI_WF200_init_f()){
		return J_NETWORK_NOT_INITIALIZED;
	}

	if(NETIF_NO_ERROR != NETIF_set_default_interface_f(0)){
		WIFI_WF200_deinit_f();
		return J_NETWORK_NOT_INITIALIZED;
	}

	if(NETIF_NO_ERROR != NETIF_initialize_f()){
		WIFI_WF200_deinit_f();
		return J_NETWORK_NOT_INITIALIZED;
	}

	xTaskCreate( xNetDelegateTaskFunction, "NET_DELEGATE", NET_DELEGATE_STACK_SIZE, NULL, NET_DELEGATE_TASK_PRIORITY, NULL );

	return kStatus_Success;

}


#ifdef __cplusplus
	}
#endif
