/*
 * C
 *
 * Copyright 2017-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

/**
 * @file
 * @brief lwip_util implementation over LWIP and FreeRTOS.
 * @author MicroEJ Developer Team
 * @version 0.1.0
 * @date 21 December 2017
 */

#include "lwip_util.h"

#include "LLNET_CONF.h"
#include "LLNET_Common.h"

#include "lwip/opt.h"
#include "lwip/init.h"
#include "lwip/netif.h"
#include "lwip/tcpip.h"
#include "lwip/dhcp.h"
#include "lwip/dns.h"
#include "netif/etharp.h"
//#include "ethernetif.h"
#include <netdb.h>
#include <stddef.h>
#include <string.h>

#include "FreeRTOS.h"
#include "task.h"

#include "wlanif.h"
#include "LLNET_CONF.h"
#include "LLNET_NETWORK_MEM.h"
#include "NETIF_interface.h"

#include "dhcpserver.h"

// Sanity check. Synchronization has to be activated in LWIP configuration file.
#if (SYS_LIGHTWEIGHT_PROT != 1)
	#error LWIP synchronization not activated in lwipopts.h file. Please add "#define SYS_LIGHTWEIGHT_PROT	1" in lwipopts.h header file.
#endif

#define MAX_DHCP_TRIES  4
#define LWIP_LINK_TASK_PRIORITY 9
#define LWIP_DHCP_TASK_PRIORITY 13

#define LWIP_DHCP_POLLING_INTERVAL 250

/* DHCP process states */
#define DHCP_START                 (uint8_t) 1
#define DHCP_WAIT_ADDRESS          (uint8_t) 2
#define DHCP_ADDRESS_ASSIGNED      (uint8_t) 3
#define DHCP_TIMEOUT               (uint8_t) 4
#define DHCP_LINK_DOWN             (uint8_t) 5

struct netif sta_gnetif, ap_gnetif;

uint8_t DHCP_state;

//static TaskHandle_t	link_task_handle;
TaskHandle_t	dhcp_task_handle;

uint8_t dhcp_sleeping = 1;

/* variable used to notify that DNS servers list has changed */
uint8_t dns_servers_list_updated = 1;

extern uint8_t ip_addr0;
extern uint8_t ip_addr1;
extern uint8_t ip_addr2;
extern uint8_t ip_addr3;

extern uint8_t netmask_addr0;
extern uint8_t netmask_addr1;
extern uint8_t netmask_addr2;
extern uint8_t netmask_addr3;

extern uint8_t gw_addr0;
extern uint8_t gw_addr1;
extern uint8_t gw_addr2;
extern uint8_t gw_addr3;

static void User_notification(struct netif *netif);

void set_link_up(uint8_t is_soft_ap)
{
	int32_t dhcpConfEnabled = ((int32_t)&LLNET_DHCP_CONF == LLNET_CONF_TRUE);
	if(is_soft_ap)
	{
		netif_set_link_up(&ap_gnetif);
		netif_set_up(&ap_gnetif);
		dhcpserver_start();
	}else
	{
		netif_set_link_up(&sta_gnetif);
		netif_set_up(&sta_gnetif);
		if (dhcpConfEnabled)
		{
			User_notification( (struct netif *) &sta_gnetif);
		}
	}

}
void set_link_down(uint8_t is_soft_ap)
{
	int32_t dhcpConfEnabled = ((int32_t)&LLNET_DHCP_CONF == LLNET_CONF_TRUE);
	if(is_soft_ap)
	{
		dhcpserver_stop();
		netif_set_link_down(&ap_gnetif);
		netif_set_down(&ap_gnetif);
	}else
	{
		netif_set_link_down(&sta_gnetif);
		netif_set_down(&sta_gnetif);
		if (dhcpConfEnabled)
		{
			User_notification(&sta_gnetif);
		}
	}
}

static void ethernetif_static_ip_config(void);

/**
  * @brief  Reset the network interface ip, netmask and gateway addresses to zero.
  * @param  netif: the network interface
  * @retval None
  */
static void netif_addr_set_zero_ip4(struct netif* netif){
	ip_addr_set_zero_ip4(&netif->ip_addr);
	ip_addr_set_zero_ip4(&netif->netmask);
	ip_addr_set_zero_ip4(&netif->gw);
}

/**
  * @brief  This function is called when the network interface is disconnected.
  * @param  netif: the network interface
  * @retval None
  */
void netif_not_connected(struct netif *netif){
	netif_addr_set_zero_ip4(netif);
	LLNET_DEBUG_TRACE("[INFO] The network cable is not connected \n");
}

/**
  * @brief  Notify the User about the nework interface config status
  * @param  netif: the network interface
  * @retval None
  */
static void User_notification(struct netif *netif)
{
	int32_t dhcpConfEnabled = LLNET_CONF_is_ip_dhcp_enabled() == LLNET_CONF_TRUE;
	if (netif_is_up(netif))
	{
		if(dhcpConfEnabled)
		{
			/* Update DHCP state machine */
			DHCP_state = DHCP_START;
		}else{
			// launch static Network Interface configuration
			ethernetif_static_ip_config();
		}
	}
	else
	{
		netif_not_connected(netif);
		if(dhcpConfEnabled)
		{
			/* Update DHCP state machine */
			DHCP_state = DHCP_LINK_DOWN;
		}
	}

}

/**
  * @brief  DHCP Process
* @param  argument: network interface
  * @retval None
  */
static void DHCP_thread(void const * argument)
{
  unsigned char IPv4_Addr[4] = { 0 };  // IP v4 address, used to 
  char netifName[4] = { 0 };

  struct netif *netif = (struct netif *) argument;
  uint32_t IPaddress;

  // Build NETIF name to be passed to NETIF_ functions
  netifName[0] = netif->name[0];
  netifName[1] = netif->name[1];
  netifName[2] = '0' + netif->num; // Convert to ACSII.
  netifName[3] = 0; // NULL Char at the end.

  for (;;)
  {
  	// check if DHCP thread has to suspend
	if(dhcp_sleeping == 1){
		vTaskSuspend(NULL);
	}

    switch (DHCP_state)
    {
    case DHCP_START:
      {
        memset ( &IPv4_Addr[0], 0, sizeof(IPv4_Addr) );
        NETIF_set_gateway_f(netifName, IPv4_Addr);   
        NETIF_set_netmask_f(netifName, IPv4_Addr);
        netif_addr_set_zero_ip4(netif);
        IPaddress = 0;
        dhcp_start(netif);
        DHCP_state = DHCP_WAIT_ADDRESS;

        LLNET_DEBUG_TRACE("[INFO] DHCP started\n");
      }
      break;

    case DHCP_WAIT_ADDRESS:
      {
        /* Read the new IP address */
        IPaddress = netif->ip_addr.addr;

        if (IPaddress!=0)
        {
          DHCP_state = DHCP_ADDRESS_ASSIGNED;

          /* Stop DHCP */
#ifdef LWIP_2_1_2
          // LwIP 2.1.2 clears existing IP address if dhcp is stopped.
          //dhcp_stop(netif);
#else
          dhcp_stop(netif);
#endif
		  dhcp_sleeping = 1;

		  LLNET_DEBUG_TRACE("[INFO] %c%c%d DHCP IP=%s\n", netif->name[0], netif->name[1], netif->num, inet_ntoa(IPaddress) );
          LLNET_DEBUG_TRACE("[INFO] DHCP GW address of interface set to %"U16_F".%"U16_F".%"U16_F".%"U16_F"\n",
                ip4_addr1_16(netif_ip4_gw(netif)),
                ip4_addr2_16(netif_ip4_gw(netif)),
                ip4_addr3_16(netif_ip4_gw(netif)),
                ip4_addr4_16(netif_ip4_gw(netif)));
          LLNET_DEBUG_TRACE ("[INFO] DHCP netmask of interface set to %"U16_F".%"U16_F".%"U16_F".%"U16_F"\n",
                ip4_addr1_16(netif_ip4_netmask(netif)),
                ip4_addr2_16(netif_ip4_netmask(netif)),
                ip4_addr3_16(netif_ip4_netmask(netif)),
                ip4_addr4_16(netif_ip4_netmask(netif)));

            IPv4_Addr[0] = ip4_addr1(netif_ip4_gw(netif));
            IPv4_Addr[1] = ip4_addr2(netif_ip4_gw(netif));
            IPv4_Addr[2] = ip4_addr3(netif_ip4_gw(netif));
            IPv4_Addr[3] = ip4_addr4(netif_ip4_gw(netif));            
            NETIF_set_gateway_f(netifName, IPv4_Addr);

            IPv4_Addr[0] = ip4_addr1(netif_ip4_netmask(netif));
            IPv4_Addr[1] = ip4_addr2(netif_ip4_netmask(netif));
            IPv4_Addr[2] = ip4_addr3(netif_ip4_netmask(netif));
            IPv4_Addr[3] = ip4_addr4(netif_ip4_netmask(netif));            
            NETIF_set_netmask_f(netifName, IPv4_Addr);

            // set static DNS configuration if required by user
            if(LLNET_CONF_is_dns_dhcp_enabled() != LLNET_CONF_TRUE)
            {
                ip_addr_t dnsaddr;
                if(DNS_MAX_SERVERS > 0)
                {
                    char * static_dns_ip_addr = LLNET_CONF_get_dns_ip_address();
                    if(static_dns_ip_addr != NULL)
                    {
                        dnsaddr.addr = inet_addr(static_dns_ip_addr);
                        dns_setserver(0, &dnsaddr);
                    }
                }
            }
            else
            {
                ip_addr_t const *dnsIPptr = dns_getserver(0);
                IPv4_Addr[0] = ip4_addr1(dnsIPptr);
                IPv4_Addr[1] = ip4_addr2(dnsIPptr);
                IPv4_Addr[2] = ip4_addr3(dnsIPptr);
                IPv4_Addr[3] = ip4_addr4(dnsIPptr);    
                NETIF_set_dns_f(netifName, IPv4_Addr);
            }
            

            // notify DNS servers IP address updated
            dns_servers_list_updated = 1;
        }
        else
        {
          /* DHCP timeout */
//          if (netif->dhcp->tries > MAX_DHCP_TRIES)
//          {
//            DHCP_state = DHCP_TIMEOUT;
//
//            /* Stop DHCP */
//            dhcp_stop(netif);
//			dhcp_sleeping = 1;
//
//			LLNET_DEBUG_TRACE("[INFO] DHCP timeout\n");
//          }
        }
      }
      break;

    default: break;
    }

    /* wait 250 ms */
	TickType_t ticks = LWIP_DHCP_POLLING_INTERVAL / portTICK_PERIOD_MS;
    vTaskDelay(ticks ? ticks : 1);          /* Minimum delay = 1 tick */
  }
}

/**
 * @brief  Setup the network interface
 * @param  None
 * @retval None
 */
static void Netif_Config(void)
{

#if 0
	ip_addr_t ipaddr;
	ip_addr_t netmask;
	ip_addr_t gw;

	ip_addr_set_zero_ip4(&ipaddr);
	ip_addr_set_zero_ip4(&netmask);
	ip_addr_set_zero_ip4(&gw);

	/* Add the network interface */
	netif_add(&gnetif, &ipaddr, &netmask, &gw, NULL, &ethernetif_init, &ethernet_input);

	/* Registers the default network interface */
	netif_set_default(&gnetif);

	if (netif_is_link_up(&gnetif))
	{
		/* When the netif is fully configured this function must be called */
		netif_set_up(&gnetif);
	}
	else
	{
		/* When the netif link is down this function must be called */
		netif_set_down(&gnetif);
	}

	/* Set the link callback function, this function is called on change of link status */
	netif_set_link_callback(&gnetif, ethernetif_update_config);

	/* Create the Ethernet link handler thread */
#if defined(__GNUC__)
	xTaskCreate((TaskFunction_t)ethernetif_set_link, "LWIP Link", configMINIMAL_STACK_SIZE * 5, &gnetif, LWIP_LINK_TASK_PRIORITY, &link_task_handle);
#else
	xTaskCreate((TaskFunction_t)ethernetif_set_link, "LWIP Link", configMINIMAL_STACK_SIZE * 2, &gnetif, LWIP_LINK_TASK_PRIORITY, &link_task_handle);
#endif

#else

	ip_addr_t sta_ipaddr, ap_ipaddr;
	ip_addr_t sta_netmask, ap_netmask;
	ip_addr_t sta_gw, ap_gw;

	ip_addr_set_zero_ip4(&sta_ipaddr);
	ip_addr_set_zero_ip4(&sta_netmask);
	ip_addr_set_zero_ip4(&sta_gw);

	IP_ADDR4(&ap_ipaddr,ip_addr0,ip_addr1,ip_addr2,ip_addr3);
	IP_ADDR4(&ap_netmask,netmask_addr0,netmask_addr1,netmask_addr2,netmask_addr3);
	IP_ADDR4(&ap_gw,gw_addr0,gw_addr1,gw_addr2,gw_addr3);

	/* Add the network interface */
#if NO_SYS
	netif_add(&sta_gnetif,  &sta_ipaddr, &sta_netmask, &sta_gw, NULL, &sta_ethernetif_init, &netif_input);
#else
    netif_add(&sta_gnetif, &sta_ipaddr, &sta_netmask, &sta_gw, NULL, &sta_ethernetif_init, &tcpip_input);
    netif_add(&ap_gnetif, &ap_ipaddr, &ap_netmask, &ap_gw, NULL, &ap_ethernetif_init, &tcpip_input);
#endif

	/* Registers the default network interface */
	netif_set_default(&sta_gnetif);

#endif
    
}


/*
 * Retrieve static IP configuration of the default network interface and set
 * IP parameters (Interface IP address, Netmask and Gateway IP address).
 */
static void ethernetif_static_ip_config()
{
	ip_addr_t ipaddr;
	ip_addr_t netmask;
	ip_addr_t gw;
	ip_addr_t dnsaddr;

	// static IP configuration. Retrieve IP settings from user properties.
	ipaddr.addr = inet_addr(LLNET_CONF_get_device_ip_address());
	netmask.addr = inet_addr(LLNET_CONF_get_netmask());
	gw.addr	= inet_addr(LLNET_CONF_get_gateway_ip_address());
	netif_set_addr(&sta_gnetif, &ipaddr , &netmask, &gw);
	LLNET_DEBUG_TRACE("[INFO] Static IP address assigned: %s\n", inet_ntoa(ipaddr.addr));

	// set static DNS Host IP address.
	if(DNS_MAX_SERVERS > 0)
	{
		char * static_dns_ip_addr = (char*)LLNET_CONF_get_dns_ip_address();
		if(static_dns_ip_addr != NULL)
		{
			dnsaddr.addr = inet_addr(static_dns_ip_addr);
			dns_setserver(0, &dnsaddr);
			// notify DNS servers IP address updated
			dns_servers_list_updated = 1;
		}
	}
}

/**
  * @brief  This function notify user about link status changement.
  * @param  netif: the network interface
  * @retval None
  */
void ethernetif_notify_conn_changed(struct netif *netif)
{
	ip_addr_t ipaddr;
	ip_addr_t netmask;
	ip_addr_t gw;
	int32_t dhcpConfEnabled = LLNET_CONF_is_ip_dhcp_enabled() == LLNET_CONF_TRUE;

	if(netif_is_link_up(netif))
	{
		LLNET_DEBUG_TRACE("[INFO] The network cable is now connected \n");

		if(dhcpConfEnabled)
		{
			/* Update DHCP state machine */
			DHCP_state = DHCP_START;

			netif_set_addr(netif, &ipaddr , &netmask, &gw);

			// resume DHCP thread
			dhcp_sleeping = 0;
			vTaskResume(dhcp_task_handle);
		}else{
			// launch static Network Interface configuration
			ethernetif_static_ip_config();
		}

		/* When the netif is fully configured this function must be called.*/
		netif_set_up(netif);
	}
	else
	{
		if(dhcpConfEnabled){
			/* Update DHCP state machine */
			DHCP_state = DHCP_LINK_DOWN;
		}

		/*  When the netif link is down this function must be called.*/
		netif_set_down(netif);

		netif_not_connected(netif);
	}
}


/**
 * Network initialization. Start network interfaces and configure it.
 * @return 0 if no error occurred, error code otherwise.
 */
int32_t llnet_lwip_init(void)
{
	int32_t dhcpConfEnabled =  LLNET_CONF_is_ip_dhcp_enabled() == LLNET_CONF_TRUE;

    LLNET_NETWORK_HEAP_initialize();

	/* Initialize the LwIP TCP/IP stack */
	tcpip_init(NULL, NULL);

	/* Configure the Network interface */
	Netif_Config();

	/* Notify user about the network interface config */
	User_notification(&sta_gnetif);

	if(dhcpConfEnabled)
	{
		/* Start DHCPClient */
		dhcp_sleeping = 0;
#if defined(__GNUC__)
		xTaskCreate((TaskFunction_t)DHCP_thread, "DHCP", configMINIMAL_STACK_SIZE * 5, &gnetif, LWIP_DHCP_TASK_PRIORITY, &dhcp_task_handle);
#else
		xTaskCreate((TaskFunction_t)DHCP_thread, "DHCP", configMINIMAL_STACK_SIZE * 2, &sta_gnetif, LWIP_DHCP_TASK_PRIORITY, &dhcp_task_handle);
#endif
	}

	return 0;
}
