/*
 * C
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */



/* Toolchain includes */
#include "lwip/err.h"
#include "lwip/opt.h"
#include "netif/etharp.h"

/* wf200 includes */
#include "sl_wfx.h"
#include "sl_wfx_host.h"

#include "wlanif.h"

//#include "fsl_enet.h"
//#include "fsl_phy.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////Macro definitions//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define STA_IFNAME0 's' ///< Station network  interface name 0
#define STA_IFNAME1 't' ///< Station network interface name 1

#define AP_IFNAME0 'a' ///< SoftAp network interface name 0
#define AP_IFNAME1 'p' ///< SoftAp network interface name 1

extern sl_wfx_context_t wifi;
extern sl_wfx_state_t sl_wfx_status;
extern char wlan_ssid[];
extern char wlan_passkey[];
extern sl_wfx_security_mode_t wlan_security;

extern char softap_ssid[];
extern char softap_passkey[];
extern sl_wfx_security_mode_t softap_security;
extern uint8_t softap_channel;

/* Time waiting to check new interface link status */
#define INTERFACE_LINK_STATUS_PERIOD	500

/**@brief Log priority levels */

#define WIFI_LOG_DEBUG     0
#define WIFI_LOG_ERROR     1
#define WIFI_LOG_NONE      2

/**@brief Current log level */
#define WIFI_LOG_LEVEL WIFI_LOG_NONE

#ifndef WIFI_LOG_LEVEL
    #error "WIFI_LOG_LEVEL must be defined"
#endif

/**@brief Debug logger */

#if (WIFI_LOG_DEBUG>= WIFI_LOG_LEVEL)
    #define WIFI_LOG_DEBUG_TRACE       printf("[wifi][DEBUG] ");printf
#else
    #define WIFI_LOG_DEBUG_TRACE(...)   ((void) 0)
#endif

/**@brief Error logger */

#if (WIFI_LOG_ERROR >= WIFI_LOG_LEVEL)
    #define WIFI_LOG_ERROR_TRACE        printf("[wifi][ERROR] ");printf
#else
    #define WIFI_LOG_ERROR_TRACE(...)   ((void) 0)
#endif

/*Private Variables*/
struct netif *netifWifi = NULL;

/* Semaphore to signal wf200 driver available */
SemaphoreHandle_t s_xDriverSemaphore = NULL;

/**
 * @brief Data priority level
 */
typedef enum wf200_priority_e {
  WF200_PRIORITY_BE                                = 0x0,	///< Best Effort
  WF200_PRIORITY_BK                                = 0x1,	///< Background
  WF200_PRIORITY_EE                                = 0x2,	///< Excellent effort
  WF200_PRIORITY_CA                                = 0x3,	///< Critical applications
  WF200_PRIORITY_VI                                = 0x4,	///< Video
  WF200_PRIORITY_VO                                = 0x5,	///< Voice
  WF200_PRIORITY_IC                                = 0x6,	///< Internetwork control
  WF200_PRIORITY_NC                                = 0x7	///< Network control
} wf200_priority_t;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////Functions Prototypes////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * @brief this function initializes the hardware parameters and is called from ethernetif_init()
 *
 *@param[in] netif: the already initialized lwip network interface structure
 *
 * @return none
 */
static void low_level_init(struct netif *netif);

/**
 * @brief this function does the actual transmission of the packets.
 * pbuf contains the packet that is passed to the function. pbuf might be chained.
 *
 * @param[in] netif: the lwip network interface structure
 *
 * @param[in] p: the packet to send
 *
 * @return
 *	ERR_OK if successful
 */
static err_t low_level_output(struct netif *netif, struct pbuf *p);

/**
* @brief this function transfers the receive packets from the wf200 to lwip.
*
* @param[in] netif: the lwip network interface structure
*
* @param[in] rx_buffer: the frame received by the wf200
*
* @return
*    LwIP pbuf filled with received packet, or NULL if and error occurred
*/
static struct pbuf * low_level_input(struct netif *netif, sl_wfx_received_ind_t* rx_buffer);

/**
* @brief this function implements the wf200 received frame callback.
*
* @param[in] rx_buffer: the frame received by the wf200
*
* @return
*    None
*/
void sl_wfx_host_received_frame_callback(sl_wfx_received_ind_t* rx_buffer);

/**
* @brief called at the beginning of the program to set up the network interface.
*
* @param[in] netif: the lwip network interface structure
*
* @return
*    ERR_OK if successful
******************************************************************************/
//err_t ethernetif_init(struct netif *netif);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////Private Functions//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint8_t soft_ap_mode = 0;
static void low_level_init(struct netif *netif)
{
	//printf("%s\n",__func__);

	/*set MAC hardware address length of netif*/
	netif->hwaddr_len = SL_WFX_MAC_ADDR_SIZE;

	/* Check which netif is initialized and set netif MAC hardware address */
	if (netif->name[0] == STA_IFNAME0)
	{
		netif->hwaddr[0] =  wifi.mac_addr_0.octet[0];
		netif->hwaddr[1] =  wifi.mac_addr_0.octet[1];
		netif->hwaddr[2] =  wifi.mac_addr_0.octet[2];
		netif->hwaddr[3] =  wifi.mac_addr_0.octet[3];
		netif->hwaddr[4] =  wifi.mac_addr_0.octet[4];
		netif->hwaddr[5] =  wifi.mac_addr_0.octet[5];
	}
	else
	{
		netif->hwaddr[0] =  wifi.mac_addr_1.octet[0];
		netif->hwaddr[1] =  wifi.mac_addr_1.octet[1];
		netif->hwaddr[2] =  wifi.mac_addr_1.octet[2];
		netif->hwaddr[3] =  wifi.mac_addr_1.octet[3];
		netif->hwaddr[4] =  wifi.mac_addr_1.octet[4];
		netif->hwaddr[5] =  wifi.mac_addr_1.octet[5];
	}

	/*set the interface maximum transfer unit*/
	netif->mtu = 1500;

	/*Accept broadcast address and ARP traffic*/
	netif->flags |= NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP;

	/* create a semaphore used for making driver accesses atomic */
	vSemaphoreCreateBinary(s_xDriverSemaphore);

	/* Set netif link flag */
	netif->flags |= NETIF_FLAG_LINK_UP;

}

static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
	err_t errval = 0;
	struct pbuf *q = NULL;
	sl_wfx_send_frame_req_t* tx_buffer = NULL;
	uint8_t *buffer = NULL;
	uint32_t framelength = 0;
	uint32_t bufferoffset = 0;
	uint32_t byteslefttocopy = 0;
	uint32_t i = 0;
	bufferoffset = 0;
	uint32_t padding = 0;

	for(q = p; q != NULL; q = q->next)
    {
        framelength = framelength + q->len;
    }
    
    if (framelength < 60)
    {
        padding = 60 - framelength;
    }
    else
    {
        padding = 0;
    }

    if( xSemaphoreTake(s_xDriverSemaphore, (portMAX_DELAY)) == pdTRUE)
    {
        sl_wfx_host_allocate_buffer((void**)(&tx_buffer), SL_WFX_TX_FRAME_BUFFER,
                                    SL_WFX_ROUND_UP(framelength+padding, 64) + sizeof(sl_wfx_send_frame_req_t));

        if(tx_buffer)
        {

            buffer = tx_buffer->body.packet_data;

            framelength = 0;
            /* copy frame from pbufs to driver buffers */
            for(q = p; q != NULL; q = q->next)
            {
                /* Get bytes in current lwIP buffer */
                byteslefttocopy = q->len;
                /* Copy the bytes */
                memcpy((uint8_t*)((uint8_t*)buffer + bufferoffset),
                     (uint8_t*)((uint8_t*)q->payload), byteslefttocopy);
                bufferoffset = bufferoffset + byteslefttocopy;
                framelength = framelength + byteslefttocopy;
            }
            for (i=framelength;i< framelength+padding;i++)
            {
              //fill zeros
              buffer[i] = 0;
            }

            /* Transmit to the station or softap interface */
            if (netif->name[0] == AP_IFNAME0)
            {
                sl_wfx_send_ethernet_frame(tx_buffer, framelength + padding, SL_WFX_SOFTAP_INTERFACE, WF200_PRIORITY_BE);
            }
            else
            {

                sl_wfx_send_ethernet_frame(tx_buffer, framelength + padding, SL_WFX_STA_INTERFACE, WF200_PRIORITY_BE);

        //			printf("\r\n");
        //			printf("tx_buffer->header: |  lenght:%d  |  id:%x  |  info:%d  |\n", tx_buffer->header.length, tx_buffer->header.id, tx_buffer->header.info);
        //			printf("tx_buffer->body:  |  packet_data_length:%d  |  frame_type:%x  |  priority:%x  |\n",
        //					tx_buffer->body.packet_data_length, tx_buffer->body.frame_type, tx_buffer->body.priority);
        //			int i = 0;
        //			for(i=0; i<tx_buffer->body.packet_data_length; i++)
        //			{
        //				printf("0x%x ", tx_buffer->body.packet_data[i]);
        //			}
        //			printf("\r\n");
        //			printf("\r\n");
            }

            errval = ERR_OK;
        }
        else
        {
            // tx_buffer allocation fail
            errval = ERR_MEM;
        }
        
        //Free buffer
        sl_wfx_host_free_buffer(tx_buffer, SL_WFX_TX_FRAME_BUFFER);        
    }
    else
    {
        //unable to send
        errval = ERR_TIMEOUT;
    }
    xSemaphoreGive(s_xDriverSemaphore);
    
	return errval;
}

static struct pbuf * low_level_input(struct netif *netif, sl_wfx_received_ind_t* rx_buffer)
{
	struct pbuf *p = NULL;
	struct pbuf *q = NULL;
	uint8_t *buffer = NULL;
	uint32_t bufferoffset = 0;
	uint32_t payloadoffset = 0;
	uint32_t byteslefttocopy = 0;
	/* get received frame */

	/* Obtain the packet by removing the padding. */
//	printf("\r\n");
//	printf("rx_buffer->header: |  lenght:%d  |  id:%x  |  info:%d  |\n", rx_buffer->header.length, rx_buffer->header.id, rx_buffer->header.info);
//	printf("rx_buffer->body:  |  frame_length:%d  |  frame_type:%x  | \n", rx_buffer->body.frame_length, rx_buffer->body.frame_type);
//	int i = 0;
//	for(i=0; i<rx_buffer->body.frame_length; i++)
//	{
//		printf("0x%x ", rx_buffer->body.frame[i]);
//	}
//	printf("\r\n");
//	printf("\r\n");

	buffer = (uint8_t *)&(rx_buffer->body.frame[rx_buffer->body.frame_padding]);

	if (rx_buffer->body.frame_length > 0)
	{
		/* We allocate a pbuf chain of pbufs from the Lwip buffer pool */
		p = pbuf_alloc(PBUF_RAW, rx_buffer->body.frame_length, PBUF_POOL);
	}

	if (p != NULL)
	{
		bufferoffset = 0;
		for(q = p; q != NULL; q = q->next)
		{
	  		byteslefttocopy = q->len;
	  		payloadoffset = 0;
            
	  		/* Copy remaining data in pbuf */
	 		 memcpy((uint8_t*)((uint8_t*)q->payload + payloadoffset),
			 	(uint8_t*)((uint8_t*)buffer + bufferoffset), byteslefttocopy);
	  		bufferoffset = bufferoffset + byteslefttocopy;
		}

	}
	return p;
}

void sl_wfx_host_received_frame_callback(sl_wfx_received_ind_t* rx_buffer)
{
	struct pbuf *p = NULL;
	struct netif *netif;
	/* Check packet interface to send to AP or STA interface */
	if((rx_buffer->header.info & SL_WFX_MSG_INFO_INTERFACE_MASK) ==
			(SL_WFX_STA_INTERFACE << SL_WFX_MSG_INFO_INTERFACE_OFFSET))
	{
		/* Send to station interface */
		netif = &sta_gnetif;
	}else{
		/* Send to softAP interface */
		netif = &ap_gnetif;
	}
	if (netif != NULL)
	{
		p = low_level_input(netif , rx_buffer);
		if (netif->input(p, netif) != ERR_OK)
		{
			if(p != NULL)
			{
				pbuf_free(p);
			}
		}
	}
}

err_t sta_ethernetif_init(struct netif *netif)
{
#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif->hostname = "lwip";
#endif /* LWIP_NETIF_HOSTNAME */

	netif->name[0] = STA_IFNAME0;
	netif->name[1] = STA_IFNAME1;

	netif->output = etharp_output;
	netif->linkoutput = low_level_output;

	/* initialize the hardware */
	low_level_init(netif);
	sta_gnetif = *netif;

	return ERR_OK;
}

err_t ap_ethernetif_init(struct netif *netif)
{
#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif->hostname = "lwip";
#endif /* LWIP_NETIF_HOSTNAME */

	netif->name[0] = AP_IFNAME0;
	netif->name[1] = AP_IFNAME1;

	netif->output = etharp_output;
	netif->linkoutput = low_level_output;

	/* initialize the hardware */
	low_level_init(netif);
	ap_gnetif = *netif;

	return ERR_OK;
}
