/*
 * C
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

// BSP
#include "fsl_debug_console.h"
#include "fsl_dma.h"
#include "fsl_fbdev.h"
#include "fsl_dc_fb_dsi_cmd.h"

// RTOS
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

// Project
#include "mej_debug.h"
#include "mej_log.h"

#include "display_support.h"

#include "vglite_support.h"
#include "vglite_window.h"

#include "board.h"
#include "pin_mux.h"

#include "LLUI_DISPLAY.h"

#include "display_task.h"
#include "display_vglite.h"
#include "display_dma.h"

#include "framerate.h"

#include "power_manager.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

#define DISPLAY_STACK_SIZE       (8 * 1024)
#define DISPLAY_TASK_PRIORITY    (12)                       /** Should be > tskIDLE_PRIORITY & < configTIMER_TASK_PRIORITY */
#define DISPLAY_TASK_STACK_SIZE  (DISPLAY_STACK_SIZE / 4)

// -----------------------------------------------------------------------------
// Types
// -----------------------------------------------------------------------------

/*
 * @brief: Synchronization structure
 *
 * This structure is used to synchronize the framebuffers before a flush.
 * It is used in multiple buffering.
 */
typedef struct s_display_sync {
	int32_t addr;	// Address of the source framebuffer
	int32_t xmin;	// Left-most coordinate of the area to synchronize
	int32_t ymin;	// Top-most coordinate of the area to synchronize
	int32_t xmax;	// Right-most coordinate of the area to synchronize
	int32_t ymax;	// Bottom-most coordinate of the area to synchronize
} __display_sync_t;

// -----------------------------------------------------------------------------
// Static Variables
// -----------------------------------------------------------------------------

/*
 * @brief: Semaphore to synchronize the display flush with MicroUI
 *
 * This semaphore can either be used to start a flush or to wait a
 * hardware rendering operation from the GPU.
 * */
static SemaphoreHandle_t sync_flush;

/*
 * @brief: Flag to indicate if the task should wait for a
 * hardware rendering operation from the GPU or start a flush.
 */
static volatile bool task_flush_or_vg;

/*
 * @brief: pointer to the main VGLite window
 */
static vg_lite_window_t *pWindow;

/*
 * @brief: Structure to synchronize the framebuffers before a flush
 */
static __display_sync_t display_sync;

/*
 * @brief: callback to perform extra flush operations
 */
static display_task_flush_callback_t flush_callback = 0;

// -----------------------------------------------------------------------------
// Internal functions definition
// -----------------------------------------------------------------------------

/*
 * @brief: Task to manage display flushes and synchronize with hardware rendering
 * operations
 */
static void __display_task(void * pvParameters);

/*
 * @brief: Flush current framebuffer to the display
 */
static void __display_task_swap_buffers(void);

/*
 * @brief: Flush current framebuffer and manage simple/double buffering modes
 */
static void __display_task_flush(void);

/*
 * @brief: Wait current VGLite operation and notify MicroUI stack
 */
static void __display_task_finish_vg_lite_operation(void);

// -----------------------------------------------------------------------------
// Project functions
// -----------------------------------------------------------------------------

// See the header file for the function documentation
void display_task_initialize(vg_lite_window_t *window) {
	pWindow = window;

	// MEJ_DEBUG_enable(MEJ_DEBUG_DISPLAY_TASK_LVL);

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_TASK_LVL, "enter\n");

	vSemaphoreCreateBinary(sync_flush);
	xSemaphoreTake(sync_flush, 0);

	BaseType_t ret = xTaskCreate(
			__display_task,
			"Display",
			DISPLAY_TASK_STACK_SIZE,
			NULL,
			DISPLAY_TASK_PRIORITY,
			NULL);

	if (NULL == ret) {
		MEJ_LOG_CRITICAL("failed to create task __display\n");
	}

#if defined FRAMERATE_NATIVE
	framerate_init(250);
#endif

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_TASK_LVL, "leave\n");
}

// See the header file for the function documentation
void display_task_wakeup(bool flush_or_vg) {
	task_flush_or_vg = flush_or_vg;
	xSemaphoreGive(sync_flush);
}

// See the header file for the function documentation
void display_task_flush(
		uint8_t* addr,
		uint32_t xmin,
		uint32_t ymin,
		uint32_t xmax,
		uint32_t ymax) {
	display_sync.addr = (int32_t)addr;
	display_sync.xmin = (int32_t)xmin;
	display_sync.ymin = (int32_t)ymin;
	display_sync.xmax = (int32_t)xmax;
	display_sync.ymax = (int32_t)ymax;

	display_task_wakeup(true);
}

// See the header file for the function documentation
void display_task_register_flush_callback(display_task_flush_callback_t callback) {
	flush_callback = callback;
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

// See the section 'Internal function definitions' for the function documentation

static void __display_task_swap_buffers(void)
{
	if (display_sync.ymin > 0 || display_sync.ymax < FRAME_BUFFER_HEIGHT - 1)
	{         
		// No need to send all content: modify framebuffer information to reduce
		// the number of data to send

		// Retrieve framebuffer information
		const dc_fb_t *dc = pWindow->display->g_fbdev.dc;
		dc_fb_dsi_cmd_handle_t *dcHandle = dc->prvData;
		dc_fb_dsi_cmd_layer_t *pLayer    = &(dcHandle->layers[pWindow->display->g_fbdev.layer]);
		dc_fb_info_t *fbInfo             = &(pLayer->fbInfo);
        
		// Update startY and height
		fbInfo->startY = display_sync.ymin;
		fbInfo->height = display_sync.ymax - display_sync.ymin + 1;

		// Update memory pointer to point on first dirty line
		vg_lite_buffer_t * current_buffer = VGLITE_GetRenderTarget(pWindow);
		uint8_t * original_memory = (uint8_t *) current_buffer->memory;
		current_buffer->memory = (uint8_t *)current_buffer->memory +  (display_sync.ymin) * (fbInfo->strideBytes);

		VGLITE_SwapBuffers(pWindow);

		// Restore original context
		current_buffer->memory = original_memory;
		fbInfo->startY = 0;
		fbInfo->height = FRAME_BUFFER_HEIGHT;
	}
	else
	{
		// just have to swap both buffers
		VGLITE_SwapBuffers(pWindow);
	}
}

static void __display_task_flush(void) {

	vg_lite_buffer_t *current_buffer;

	if (flush_callback != 0) {
		current_buffer = VGLITE_GetRenderTarget(pWindow);
		flush_callback(current_buffer);
	}

	// Two actions:
	// 1- wait for the end of previous swap (if not already done): wait the
	// end of sending of current frame buffer to display
	// 2- start sending of current_buffer to display (without waiting the
	// end)
	__display_task_swap_buffers();

#ifdef FRAMERATE_ENABLED
	/* Increment framerate */
	framerate_increment();
#endif

#if APP_BUFFER_COUNT > 1
	current_buffer = VGLITE_GetRenderTarget(pWindow);

	// Configure frame buffer powering
	// at that point current_buffer is back buffer
	power_manager_framebuffer_power_on(current_buffer->memory);
	power_manager_framebuffer_set_next_off((framebuffer_t *)display_sync.addr);

	display_dma_start(
			(void *) display_sync.addr,
			current_buffer->memory,
			display_sync.ymin,
			display_sync.ymax);
#else
	LLUI_DISPLAY_flushDone(false);
#endif
}

// See the section 'Internal function definitions' for the function documentation
static void __display_task_finish_vg_lite_operation(void) {
	display_vglite_wait_operation();
	LLUI_DISPLAY_notifyAsynchronousDrawingEnd(false);
}

// See the section 'Internal function definitions' for the function documentation
static void __display_task(void * pvParameters) {
	do {
		xSemaphoreTake(sync_flush, portMAX_DELAY);

		if (task_flush_or_vg) {
			__display_task_flush();
		} else {
			__display_task_finish_vg_lite_operation();
		}
	} while (1);
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

