/*
 * C
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include "vg_lite.h"
#include "vglite_path.h"

#include "mej_math.h"
#include "mej_log.h"
#include "mej_disable.h"

#include "ui_drawing_soft.h"
#include "dw_drawing_soft.h"
#include "ui_drawing.h"
#include "dw_drawing.h"

#include "LLUI_DISPLAY.h"

#include "display_task.h"
#include "display_utils.h"
#include "display_vglite.h"

#include "trace_vglite.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * @brief Scale factor to improve drawing quality in vector mode
 */
#define __DRAWING_SCALE_FACTOR      (1 << 1)
#define __DRAWING_SCALE_DIV         (1.0 / __DRAWING_SCALE_FACTOR)

// -----------------------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------------------

/*
 * @brief Variable to store the vglite pathes during the GPU computation.
 *
 * MicroUI design must ensure that only one computation is done at a time.
 */
static union {
	vglite_path_line_t line; // line
	vglite_path_thick_shape_line_t thick_shape_line; // thick shape line
	vglite_path_ellipse_t ellipse[2]; // ellipse (in & out)
	vglite_path_ellipse_arc_t ellipse_arc; // ellipse arc
	vglite_path_rounded_rectangle_t rounded_rectangle[2]; // round rect (in & out)
	vglite_path_thick_ellipse_arc_t thick_ellipse_arc; // anti aliased ellipse arc
} __shape_paths;

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

/*
 * @brief Draws a path to a target buffer.
 *
 * @param[in] target: Pointer to a vg_lite_buffer_t (see vg_lite.h).
 * @param[in] path: Pointer to a vg_lite_path_t (see vg_lite.h).
 * @param[in] fill_rule: Specified fill rule for the path  (see vg_lite.h).
 * @param[in] matrix: Pointer to a 3x3 transformation matrix (see vg_lite.h).
 * @param[in] blend: The blending mode (see vg_lite.h).
 * @param[in] color: The color (see vg_lite.h).
 *
 * @return The drawing status.
 */
static DRAWING_Status __draw_path(
		vg_lite_buffer_t * target,
		vg_lite_path_t * path,
		vg_lite_fill_t fill_rule,
		vg_lite_matrix_t * matrix,
		vg_lite_blend_t blend,
		vg_lite_color_t color);

/*
 * @brief Checks MicroUI clip and updates the graphical engine's drawing limits on demand. This
 * update is useless when calling vg_lite draw_path function because the callback already updates
 * the graphical engine drawing limits. See vg_lite_draw_notify_render_area() in display_vglite.c.
 *
 * @param[in] gc: Graphics context where to draw
 * @param[in] x1 the top-left pixel X coordinate.
 * @param[in] y1 the top-left pixel Y coordinate.
 * @param[in] x2 the bottom-right pixel X coordinate.
 * @param[in] y2 the top-right pixel Y coordinate.
 * @param[in] update_drawing_limits true to update drawing limits
 *
 * @return false when the limits are fully outside the clip, true when at least
 * one pixel fits the clip.
 */
static bool __check_clip(
		MICROUI_GraphicsContext* gc,
		int x1,
		int y1,
		int x2,
		int y2,
		bool update_drawing_limits);

/*
 * @brief Draws an ellipse arc covering the square specified by its diameter.
 * See DRAWING_impl.h for more information
 *
 * @param[in] gc: the MicroUI GraphicsContext target.
 * @param[in] x: the x coordinate of the upper-left corner of the square where the ellipse arc is drawn
 * @param[in] y: the y coordinate of the upper-left corner of the square where the ellipse arc is drawn
 * @param[in] diameter_w: the horizontal diameter of the ellipse arc to draw
 * @param[in] diameter_h: the vertical diameter of the ellipse arc to draw
 * @param[in] start_angle_deg: the beginning angle of the arc to draw
 * @param[in] arc_angle_deg: the angular extent of the arc from start_angle_deg
 * @param[in] fill: the flag indicating if the ellipse arc will be filled
 * - true: the ellipse arc will be filled
 * - false: the ellipse arc will not be filled
 *
 * @return the drawing status.
 */
static DRAWING_Status __draw_ellipse_arc(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_w,
		int diameter_h,
		float start_angle,
		float arc_angle,
		bool fill);

/*
 * @brief Draws a ellipse covering the square specified by its diameter.
 * See DRAWING_impl.h for more information
 *
 * @param[in] gc: the MicroUI GraphicsContext target.
 * @param[in] x: the x coordinate of the upper-left corner of the square where the ellipse is drawn
 * @param[in] y: the y coordinate of the upper-left corner of the square where the ellipse is drawn
 * @param[in] diameter_out_w: the horizontal outer diameter of the ellipse to draw
 * @param[in] diameter_out_h: the vertical outer diameter of the ellipse to draw
 * @param[in] diameter_in_w: the horizontal inner diameter of the ellipse to draw
 * @param[in] diameter_in_h: the vertical inner diameter of the ellipse to draw
 *
 * @return the drawing status.
 */
static DRAWING_Status __draw_ellipse(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_out_w,
		int diameter_out_h,
		int diameter_in_w,
		int diameter_in_h);

/*
 * @brief Fills a ellipse covering the square specified by its diameter.
 * See DRAWING_impl.h for more information
 *
 * @param[in] gc: the MicroUI GraphicsContext target.
 * @param[in] x: the x coordinate of the upper-left corner of the square where the ellipse is drawn
 * @param[in] y: the y coordinate of the upper-left corner of the square where the ellipse is drawn
 * @param[in] diameter_w: the horizontal diameter of the ellipse to draw
 * @param[in] diameter_h: the vertical diameter of the ellipse to draw
 *
 * @return the drawing status.
 */
static DRAWING_Status __fill_ellipse(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_w,
		int diameter_h);

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
/*
 * @brief Draws a line between two points
 * See DRAWING_impl.h for more information
 *
 * @param[in] gc the MicroUI GraphicsContext target.
 * @param[in] x1: the horizontal coordinate of the start of the line
 * @param[in] y1: the vertical coordinate of start of the line
 * @param[in] x2: the horizontal coordinate of the end of the line
 * @param[in] y2: the vertical coordinate of end of the line
 *
 * @return the drawing status.
 */
static DRAWING_Status __draw_line(
		MICROUI_GraphicsContext* gc,
		int x1,
		int y1,
		int x2,
		int y2);
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

/*
 * @brief Draws a thick line between two points
 * See DRAWING_impl.h for more information
 *
 * @param[in] gc the MicroUI GraphicsContext target.
 * @param[in] x1: the horizontal coordinate of the start of the line
 * @param[in] y1: the vertical coordinate of start of the line
 * @param[in] x2: the horizontal coordinate of the end of the line
 * @param[in] y2: the vertical coordinate of end of the line
 * @param[in] thickness: the line thickness.
 * @param[in] start cap representation of start of the line
 * @param[in] end cap representation of end of the line
 *
 * @return the drawing status.
 */
static DRAWING_Status __thick_line(
		MICROUI_GraphicsContext* gc,
		int x1,
		int y1,
		int x2,
		int y2,
		int thickness,
		DRAWING_Cap start,
		DRAWING_Cap end);

/*
 * @brief Draws a thick arc covering the square specified by its diameter.
 * See dw_drawing.h for more information
 *
 * @param[in] gc: the MicroUI GraphicsContext target.
 * @param[in] x: the x coordinate of the upper-left corner of the square where the arc is drawn
 * @param[in] y: the y coordinate of the upper-left corner of the square where the arc is drawn
 * @param[in] diameter_w: the horizontal diameter of the ellipse arc to draw
 * @param[in] diameter_h: the vertical diameter of the ellipse arc to draw
 * @param[in] start_angle_deg: the beginning angle of the arc to draw
 * @param[in] arc_angle_deg: the angular extent of the arc from start_angle_deg
 * @param[in] thickness: the arc thickness.
 * @param[in] fade: the fade to apply.
 * @param[in] start: cap representation of start of shape
 * @param[in] end: cap representation of end of shape
 *
 * @return the drawing status.
 */
static DRAWING_Status __draw_thick_shape_ellipse_arc(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_w,
		int diameter_h,
		float start_angle_deg,
		float arc_angle,
		int thickness,
		int fade,
		DRAWING_Cap start,
		DRAWING_Cap end);

/*
 * @brief Rotates an image using the GPU
 *
 * @param[in] gc: Graphics context where to draw
 * @param[in] img: Source image to draw
 * @param[in] src: Source buffer
 * @param[in] x: Horizontal coordinate of the image reference anchor top-left point.
 * @param[in] y: Vertical coordinate of the image reference anchor top-left point.
 * @param[in] xRotation: Horizontal coordinate of the rotation center
 * @param[in] yRotation: Vertical coordinate of the rotation center
 * @param[in] angle: Angle that must be used to rotate the image
 * @param[in] alpha: The opacity level to apply while drawing the rotated image
 * @param[in] filter: Quality of drawing.
 *
 * @return The drawing status.
 */
static DRAWING_Status __rotate_image(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		vg_lite_buffer_t* src,
		int x,
		int y,
		int xRotation,
		int yRotation,
		float angle_deg,
		int alpha,
		vg_lite_filter_t filter);

/*
 * @brief Scales an image using the GPU
 *
 * @param[in] gc: Graphics context where to draw.
 * @param[in] img: Source image to draw.
 * @param[in] src: Source buffer
 * @param[in] x: Horizontal coordinate of the image reference anchor top-left point.
 * @param[in] y: Vertical coordinate of the image reference anchor top-left point.
 * @param[in] factorX: Horizontal scaling factor.
 * @param[in] yRotation: Vertical scaling factor.
 * @param[in] alpha: The opacity level to apply while drawing the image
 * @param[in] filter: Quality of drawing.
 *
 * @return The drawing status.
 */
static DRAWING_Status __scale_image(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		vg_lite_buffer_t* src,
		int x,
		int y,
		float factorX,
		float factorY,
		int alpha,
		vg_lite_filter_t filter);

/*
 * @brief Flushs the last vg_lite operation and wakes up the display task
 * to notify MicroUI when the operation will be finished
 *
 * @param[in] vg_lite_error: the vg_lite error code of the last operation
 *
 * @return The drawing status.
 */
static DRAWING_Status __check_operation(vg_lite_error_t vg_lite_error);

/*
 * @brief Get color to give to vglite library according destination's foreground
 * color, image format and opacity level.
 *
 * @param[in] gc: the MicroUI GraphicsContext of destination
 * @param[in] img: the MicroUI Image of source
 * @param[in] alpha: the application opacity
 *
 * @return The vg_lite color.
 */
static vg_lite_color_t __get_vglite_color(MICROUI_GraphicsContext* gc, MICROUI_Image* img, jint alpha);

// -----------------------------------------------------------------------------
// MicroUI custom functions
// -----------------------------------------------------------------------------

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawLine(
		MICROUI_GraphicsContext* gc,
		jint x1, jint y1,
		jint x2, jint y2) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawLine(gc, x1, y1, x2, y2);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	// TODO try to crop region (x1 may lower than x2 and y1 may be lower than y2)
	if (!__check_clip(
			gc,
			0, 0,
			gc->image.width, gc->image.height, false)) {
		return DRAWING_DONE;
	}

	return __draw_line(gc, x1, y1, x2, y2);
}

#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawHorizontalLine(
		MICROUI_GraphicsContext* gc,
		jint x1, jint x2, jint y) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawLine(gc, x1, y, x2, y);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x1, y, x2, y, false)) {
		return DRAWING_DONE;
	}

	return __draw_line(gc, x1, y, x2, y);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawVerticalLine(
		MICROUI_GraphicsContext* gc,
		jint x, jint y1, jint y2) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawLine(gc, x, y1, x, y2);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x, y1, x, y2, false)) {
		return DRAWING_DONE;
	}

	return __draw_line(gc, x, y1, x, y2);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillRectangle(
		MICROUI_GraphicsContext* gc,
		jint x1, jint y1, jint x2, jint y2) {

	vg_lite_error_t vg_lite_ret;
	vg_lite_buffer_t *buffer;
	vg_lite_rectangle_t rectangle;

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillRectangle(gc, x1, y1, x2, y2);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x1, y1, x2, y2, true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	rectangle.x = x1;
	rectangle.y = y1;
	rectangle.width = (x2-x1+1);
	rectangle.height = (y2-y1+1);
	buffer = display_vglite_configure_destination(gc);

	TRACE_VGLITE_INIT_GPU_START(CLEAR);
	vg_lite_ret = vg_lite_clear(buffer, &rectangle, gc->foreground_color);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawRoundedRectangle(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height,
		jint arc_width, jint arc_height) {

	int path_offset;
	vg_lite_path_t shape_vg_path;
	vg_lite_matrix_t matrix;

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawRoundedRectangle(gc, x, y, width, height, arc_width,  arc_height);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x, y, x + width, y + height, true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);
	// Compute the rounded rectangle shape path
	shape_vg_path.path = &__shape_paths.rounded_rectangle[0];
	shape_vg_path.path_length = sizeof(__shape_paths.rounded_rectangle);
	path_offset = 0;
	path_offset = vglite_compute_rounded_rectangle(
			&shape_vg_path, path_offset,
			x, y, width, height, arc_width, arc_height, false);

	path_offset = vglite_compute_rounded_rectangle(
			&shape_vg_path, path_offset,
			x + 1, y + 1,
			width - 2, height - 2,
			arc_width - 1, arc_height - 1,
			true);

	if (0 > path_offset) {
		MEJ_LOG_ERROR("Error during path computation");
		return DRAWING_DONE;
	}

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_EVEN_ODD,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillRoundedRectangle(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height,
		jint arc_width, jint arc_height) {

	int path_offset;
	vg_lite_path_t shape_vg_path;
	vg_lite_matrix_t matrix;

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillRoundedRectangle(gc, x, y, width, height, arc_width,  arc_height);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x, y, x + width, y + height, true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);
	// Compute the rounded rectangle shape path
	shape_vg_path.path = &__shape_paths.rounded_rectangle[0];
	shape_vg_path.path_length = sizeof(__shape_paths.rounded_rectangle);
	path_offset = vglite_compute_rounded_rectangle(&shape_vg_path, 0,
			x, y, width, height, arc_width, arc_height, true);

	if (0 > path_offset) {
		MEJ_LOG_ERROR("Error during path computation");
		return DRAWING_DONE;
	}

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_EVEN_ODD,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawCircleArc(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint diameter,
		jfloat start_angle,
		jfloat arc_angle) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawCircleArc(gc, x, y, diameter, start_angle, arc_angle);
		return DRAWING_DONE;
	}

	if (MEJ_ABS(arc_angle) >= 360) {
		return UI_DRAWING_drawCircle(gc, x, y, diameter);
	}

	return __draw_ellipse_arc(
			gc,
			x, y,
			diameter, diameter,
			start_angle, arc_angle,
			false);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawEllipseArc(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint width,
		jint height,
		jfloat start_angle,
		jfloat arc_angle) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawEllipseArc(
				gc,
				x, y,
				width, height,
				start_angle, arc_angle);
		return DRAWING_DONE;
	}

	if (MEJ_ABS(arc_angle) >= 360) {
		return UI_DRAWING_drawEllipse(gc, x, y, width, height);
	}

	return __draw_ellipse_arc(
			gc,
			x, y,
			width, height,
			start_angle, arc_angle,
			false);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillCircleArc(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint diameter,
		jfloat start_angle, jfloat arc_angle) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillCircleArc(
				gc,
				x, y,
				diameter,
				start_angle, arc_angle);
		return DRAWING_DONE;
	}

	if (MEJ_ABS(arc_angle) >= 360) {
		return UI_DRAWING_fillCircle(gc, x, y, diameter);
	}

	return __draw_ellipse_arc(
			gc,
			x, y,
			diameter, diameter,
			start_angle, arc_angle,
			true);
}

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillEllipseArc(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height,
		jfloat start_angle, jfloat arc_angle) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillEllipseArc(
				gc,
				x, y,
				width, height,
				start_angle, arc_angle);
		return DRAWING_DONE;
	}

	if (MEJ_ABS(arc_angle) >= 360) {
		return UI_DRAWING_fillEllipse(
				gc,
				x, y,
				width, height);
	}

	return __draw_ellipse_arc(
			gc,
			x, y,
			width, height,
			start_angle, arc_angle,
			true);
}

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawEllipse(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawEllipse(gc, x, y, width, height);
		return DRAWING_DONE;
	}

	return __draw_ellipse(gc, x, y, width, height, width - 2, height - 2);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillEllipse(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillEllipse(gc, x, y, width, height);
		return DRAWING_DONE;
	}

	return __fill_ellipse(gc, x, y, width, height);
}

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawCircle(
		MICROUI_GraphicsContext* gc, jint x, jint y, jint diameter) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_drawCircle(gc, x, y, diameter);
		return DRAWING_DONE;
	}

	return __draw_ellipse(
			gc,
			x, y,
			diameter, diameter,
			diameter - 2, diameter - 2);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_fillCircle(
		MICROUI_GraphicsContext* gc, jint x, jint y, jint diameter) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		UI_DRAWING_SOFT_fillCircle(gc, x, y, diameter);
		return DRAWING_DONE;
	}

	return __fill_ellipse(gc, x, y, diameter, diameter);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickFadedPoint(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint thickness,
		jint fade) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()
			|| 	(fade > 1)) // No radial gradient with hardware
	{
		DW_DRAWING_SOFT_drawThickFadedPoint(gc, x, y, thickness,fade);
		return DRAWING_DONE;
	}

	return __fill_ellipse(gc, x - thickness/2, y - thickness/2, thickness, thickness);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickFadedLine(
		MICROUI_GraphicsContext* gc,
		jint x1,
		jint y1,
		jint x2,
		jint y2,
		jint thickness,
		jint fade,
		DRAWING_Cap start,
		DRAWING_Cap end) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()
			|| 	(fade > 1)) // No radial gradient with hardware
	{
		DW_DRAWING_SOFT_drawThickFadedLine(
				gc, x1, y1, x2, y2, thickness, fade, start, end);
		return DRAWING_DONE;
	}

	return __thick_line(
			gc,
			x1,
			y1,
			x2,
			y2,
			thickness,
			start,
			end);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickFadedCircle(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint diameter, jint thickness, jint fade) {

	int diameter_out;
	int diameter_in;

	if (!display_vglite_is_hardware_rendering_enabled()
			|| 	(fade > 1)) { // No radial gradient with hardware
		DW_DRAWING_SOFT_drawThickFadedCircle(
				gc, x, y, diameter, thickness, fade);
		return DRAWING_DONE;
	}

	x -= thickness / 2;
	y -= thickness / 2;

	diameter_out = diameter + thickness;
	diameter_in = diameter - thickness;

	return __draw_ellipse(
			gc,
			x, y,
			diameter_out, diameter_out,
			diameter_in, diameter_in);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickFadedCircleArc(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint diameter,
		jfloat start_angle_deg,
		jfloat arc_angle,
		jint thickness,
		jint fade,
		DRAWING_Cap start,
		DRAWING_Cap end) {

	if (!display_vglite_is_hardware_rendering_enabled()
			|| 	(fade > 1)) // No radial gradient with hardware
	{
		DW_DRAWING_SOFT_drawThickFadedCircleArc(	gc,
				x,
				y,
				diameter,
				start_angle_deg,
				arc_angle,
				thickness,
				fade,
				start,
				end );
		return DRAWING_DONE;
	}

	return __draw_thick_shape_ellipse_arc(
			gc,
			x, y,
			diameter, diameter,
			start_angle_deg, arc_angle,
			thickness,
			fade,
			start, end);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickFadedEllipse(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height,
		jint thickness, jint fade) {

	int diameter_out_w;
	int diameter_out_h;
	int diameter_in_w;
	int diameter_in_h;

	if (!display_vglite_is_hardware_rendering_enabled()
			|| 	(fade > 1)) { // No radial gradient with hardware
		DW_DRAWING_SOFT_drawThickFadedEllipse(
				gc,
				x, y,
				width, height,
				thickness, fade);
		return DRAWING_DONE;
	}

	x -= thickness / 2;
	y -= thickness / 2;

	diameter_out_w = width + thickness;
	diameter_out_h = height + thickness;
	diameter_in_w = width - thickness;
	diameter_in_h = height - thickness;

	return __draw_ellipse(
			gc,
			x, y,
			diameter_out_w, diameter_out_h,
			diameter_in_w, diameter_in_h);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickLine(
		MICROUI_GraphicsContext* gc,
		jint x1,
		jint y1,
		jint x2,
		jint y2,
		jint thickness) {

	// Check if rendering should be done in software
	if (!display_vglite_is_hardware_rendering_enabled()) {
		DW_DRAWING_SOFT_drawThickLine(gc, x1, y1, x2, y2, thickness);
		return DRAWING_DONE;
	}

	return __thick_line(	gc,
			x1,
			y1,
			x2,
			y2,
			thickness,
			DRAWING_ENDOFLINE_NONE,
			DRAWING_ENDOFLINE_NONE);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickCircle(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint diameter,
		jint thickness) {

	int diameter_out;
	int diameter_in;

	if (!display_vglite_is_hardware_rendering_enabled()) {
		DW_DRAWING_SOFT_drawThickCircle(gc, x, y, diameter, thickness);
		return DRAWING_DONE;
	}

	x -= thickness / 2;
	y -= thickness / 2;

	diameter_out = diameter + thickness;
	diameter_in = diameter - thickness;

	return __draw_ellipse(
			gc,
			x, y,
			diameter_out, diameter_out,
			diameter_in, diameter_in);
}

DRAWING_Status DW_DRAWING_drawThickEllipse(
		MICROUI_GraphicsContext* gc,
		jint x, jint y,
		jint width, jint height,
		jint thickness) {

	int diameter_out_w;
	int diameter_out_h;
	int diameter_in_w;
	int diameter_in_h;

	if (!display_vglite_is_hardware_rendering_enabled()) {
		DW_DRAWING_SOFT_drawThickEllipse(gc, x, y, width, height, thickness);
		return DRAWING_DONE;
	}

	x -= thickness / 2;
	y -= thickness / 2;

	diameter_out_w = width + thickness;
	diameter_out_h = height + thickness;
	diameter_in_w = width - thickness;
	diameter_in_h = height - thickness;

	return __draw_ellipse(
			gc,
			x, y,
			diameter_out_w, diameter_out_h,
			diameter_in_w, diameter_in_h);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawThickCircleArc(
		MICROUI_GraphicsContext* gc,
		jint x,
		jint y,
		jint diameter,
		jfloat start_angle_deg,
		jfloat arc_angle,
		jint thickness) {

	if (!display_vglite_is_hardware_rendering_enabled()) {
		DW_DRAWING_SOFT_drawThickCircleArc(
				gc, x, y, diameter, start_angle_deg, arc_angle, thickness);
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	return __draw_thick_shape_ellipse_arc(
			gc,
			x, y,
			diameter, diameter,
			start_angle_deg, arc_angle,
			thickness,
			1,
			DRAWING_ENDOFLINE_NONE,
			DRAWING_ENDOFLINE_NONE);
}

// See the header file for the function documentation
DRAWING_Status UI_DRAWING_drawImage(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint region_x,
		jint region_y,
		jint width,
		jint height,
		jint x,
		jint y,
		jint alpha) {

	vg_lite_error_t vg_lite_ret;
	vg_lite_buffer_t src;
	uint32_t blit_rect[4];

#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		UI_DRAWING_SOFT_drawImage(
				gc, img, region_x, region_y, width, height, x, y, alpha);
		return DRAWING_DONE;
	}
#endif

#ifdef MEJ_DISABLED_FOR_PERFORMANCE_IMAGE_RGB565
	if (MICROUI_IMAGE_FORMAT_RGB565 == img->format && 0xff == alpha){	// CPU (memcpy) is faster than GPU
		UI_DRAWING_SOFT_drawImage(
				gc, img, region_x, region_y, width, height, x, y, alpha);
		return DRAWING_DONE;
	}
#endif

	if (!display_vglite_is_hardware_rendering_enabled() || !display_vglite_configure_source(&src, img)){
		UI_DRAWING_SOFT_drawImage(
				gc, img, region_x, region_y, width, height, x, y, alpha);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(
			gc,
			x,
			y,
			x + width - 1,
			y+ height - 1,
			true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);
	vg_lite_translate(x, y, &matrix);

	blit_rect[0] = region_x;
	blit_rect[1] = region_y;
	blit_rect[2] = width;
	blit_rect[3] = height;

	TRACE_VGLITE_INIT_GPU_START(BLIT_RECT);
	vg_lite_ret = vg_lite_blit_rect(
			dst,
			&src,
			blit_rect,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			__get_vglite_color(gc, img, alpha),
			VG_LITE_FILTER_POINT);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawFlippedImage(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint region_x,
		jint region_y,
		jint width,
		jint height,
		jint x,
		jint y,
		DRAWING_Flip transformation,
		jint alpha) {

	vg_lite_error_t vg_lite_ret;
	vg_lite_buffer_t src;


#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		DW_DRAWING_SOFT_drawFlippedImage(
				gc, img,
				region_x, region_y,
				width, height,
				x, y,
				transformation, alpha);
		return DRAWING_DONE;
	}
#endif

#ifdef MEJ_DISABLED_FOR_PERFORMANCE_IMAGE_RGB565
	if (MICROUI_IMAGE_FORMAT_RGB565 == img->format && DRAWING_FLIP_NONE == transformation && 0xff == alpha){	// CPU (memcpy) is faster than GPU
		DW_DRAWING_SOFT_drawFlippedImage(
				gc, img,
				region_x, region_y,
				width, height,
				x, y,
				transformation, alpha);
		return DRAWING_DONE;
	}
#endif

	if (	!display_vglite_is_hardware_rendering_enabled()
			|| !display_vglite_configure_source(&src, img)
			|| region_x != 0		// GPU Limitation: origin different
			|| region_y != 0) {	// than (0, 0) not supported
		DW_DRAWING_SOFT_drawFlippedImage(
				gc, img,
				region_x, region_y,
				width, height,
				x, y,
				transformation, alpha);
		return DRAWING_DONE;
	}

	// Check if there is something to draw and clip drawing limits
	// TODO try to crop region
	if (!__check_clip(
			gc,
			0, 0,
			gc->image.width, gc->image.height,
			true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	// Exclude bottom right region
	src.width = width;
	src.height = height;

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);

	vg_lite_translate(x + img->width/2, y + img->height/2, &matrix);

	switch (transformation) {
	case DRAWING_FLIP_NONE:
		break;
	case DRAWING_FLIP_90:
		vg_lite_rotate(-90, &matrix);
		break;
	case DRAWING_FLIP_180:
		vg_lite_rotate(-180, &matrix);
		break;
	case DRAWING_FLIP_270:
		vg_lite_rotate(-270, &matrix);
		break;
	case DRAWING_FLIP_MIRROR:
		vg_lite_scale(-1, 1, &matrix);
		break;
	case DRAWING_FLIP_MIRROR_90:
		vg_lite_rotate(90, &matrix);
		vg_lite_scale(1, -1, &matrix);
		break;
	case DRAWING_FLIP_MIRROR_180:
		vg_lite_rotate(180, &matrix);
		vg_lite_scale(-1, 1, &matrix);
		break;
	case DRAWING_FLIP_MIRROR_270:
		vg_lite_rotate(270, &matrix);
		vg_lite_scale(1, -1, &matrix);
		break;
	default:
		MEJ_LOG_ERROR("Unexpected transformation: %d\n", transformation);
	}

	vg_lite_translate(-img->width/2, -img->height/2, &matrix);

	TRACE_VGLITE_INIT_GPU_START(BLIT);
	vg_lite_ret = vg_lite_blit(
			dst, &src,
			&matrix, VG_LITE_BLEND_SRC_OVER,
			__get_vglite_color(gc, img, alpha),
			VG_LITE_FILTER_BI_LINEAR);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawRotatedImageNearestNeighbor(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint x,
		jint y,
		jint xRotation,
		jint yRotation,
		jfloat angle,
		jint alpha) {

	vg_lite_buffer_t src;


#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		DW_DRAWING_SOFT_drawRotatedImageNearestNeighbor(gc, img, x, y, xRotation, yRotation, angle, alpha);
		return DRAWING_DONE;
	}
#endif

	// The GCNanoLiteV limitation: Same rendering around 90 degrees (89.0, 89.1 ... 90.9, 91)
	// Render is acceptable since vglite 3.0.4-rev4
	if (!display_vglite_is_hardware_rendering_enabled() || !display_vglite_configure_source(&src, img)) {
		DW_DRAWING_SOFT_drawRotatedImageNearestNeighbor(gc, img, x, y, xRotation, yRotation, angle, alpha);
		return DRAWING_DONE;
	}
	return __rotate_image(gc, img, &src, x, y, xRotation, yRotation, angle, alpha, VG_LITE_FILTER_POINT);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawRotatedImageBilinear(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint x,
		jint y,
		jint xRotation,
		jint yRotation,
		jfloat angle,
		jint alpha) {

	vg_lite_buffer_t src;

#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		DW_DRAWING_SOFT_drawRotatedImageBilinear(gc, img, x, y, xRotation, yRotation, angle, alpha);
		return DRAWING_DONE;
	}
#endif

	// The GCNanoLiteV limitation: Same rendering around 90 degrees (89.0, 89.1 ... 90.9, 91)
	// Render is acceptable since vglite 3.0.4-rev4
	if (!display_vglite_is_hardware_rendering_enabled() || !display_vglite_configure_source(&src, img)) {
		DW_DRAWING_SOFT_drawRotatedImageBilinear(gc, img, x, y, xRotation, yRotation, angle, alpha);
		return DRAWING_DONE;
	}
	return __rotate_image(gc, img, &src, x, y, xRotation, yRotation, angle, alpha, VG_LITE_FILTER_BI_LINEAR);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawScaledImageNearestNeighbor(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint x,
		jint y,
		jfloat factorX,
		jfloat factorY,
		jint alpha) {

	vg_lite_buffer_t src;

#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		DW_DRAWING_SOFT_drawScaledImageNearestNeighbor(gc, img, x, y, factorX, factorY, alpha);
		return DRAWING_DONE;
	}
#endif

	if (!display_vglite_is_hardware_rendering_enabled() || !display_vglite_configure_source(&src, img)) {
		DW_DRAWING_SOFT_drawScaledImageNearestNeighbor(gc, img, x, y, factorX, factorY, alpha);
		return DRAWING_DONE;
	}

	return __scale_image(gc, img, &src, x, y, factorX, factorY, alpha, VG_LITE_FILTER_POINT);
}

// See the header file for the function documentation
DRAWING_Status DW_DRAWING_drawScaledImageBilinear(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		jint x,
		jint y,
		jfloat factorX,
		jfloat factorY,
		jint alpha) {

	vg_lite_buffer_t src;

#ifdef MEJ_DISABLED_FOR_MSAA_LIMITATION
	if (LLUI_DISPLAY_isTransparent(img)){	// No MSAA with hardware
		DW_DRAWING_SOFT_drawScaledImageBilinear(gc, img, x, y, factorX, factorY, alpha);
		return DRAWING_DONE;
	}
#endif

	if (!display_vglite_is_hardware_rendering_enabled() || !display_vglite_configure_source(&src, img)) {
		DW_DRAWING_SOFT_drawScaledImageBilinear(gc, img, x, y, factorX, factorY, alpha);
		return DRAWING_DONE;
	}

	return __scale_image(gc, img, &src, x, y, factorX, factorY, alpha, VG_LITE_FILTER_BI_LINEAR);
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __draw_path(
		vg_lite_buffer_t * target,
		vg_lite_path_t * path,
		vg_lite_fill_t fill_rule,
		vg_lite_matrix_t * matrix,
		vg_lite_blend_t blend,
		vg_lite_color_t color) {

	vg_lite_error_t vg_lite_ret;

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	TRACE_VGLITE_INIT_GPU_START(DRAW);
	vg_lite_ret =  vg_lite_draw(target, path, fill_rule, matrix, blend, color);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

// See the section 'Internal function definitions' for the function documentation
static bool __check_clip(
		MICROUI_GraphicsContext* gc,
		int x1,
		int y1,
		int x2,
		int y2,
		bool update_drawing_limits) {

	if (LLUI_DISPLAY_isClipEnabled(gc))
	{
		if (LLUI_DISPLAY_clipRectangle(gc, &x1, &y1, &x2, &y2))
		{
			// drawing fully or partially fits the clip:

			// enable scissor for next vglite drawing
			vg_lite_enable_scissor();
			vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

			if (update_drawing_limits) {
				// update flush limits (cropped to the clip)
				LLUI_DISPLAY_setDrawingLimits(gc, x1, y1, x2, y2);
			}

			// perform drawing
			return true;
		}
		else
		{
			// drawing is fully outside the clip, nothing to draw
			return false;
		}
	}
	else
	{
		// clip is disabled :

		// disable scissor (vglite lib already crops to the buffer bounds)
		vg_lite_disable_scissor();

		if (update_drawing_limits) {
			// update flush limits
			LLUI_DISPLAY_setDrawingLimits(gc, x1, y1, x2, y2);
		}

		// perform drawing
		return true;
	}
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __draw_ellipse_arc(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_w,
		int diameter_h,
		float start_angle_deg,
		float arc_angle_deg,
		bool fill) {

	int path_offset;
	int radius_out_w;
	int radius_out_h;
	int radius_in_w;
	int radius_in_h;
	vg_lite_path_t shape_vg_path;
	vg_lite_matrix_t matrix;

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x, y, x + diameter_w, y + diameter_h, false)) {
		return DRAWING_DONE;
	}

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);

	radius_out_w = diameter_w / 2;
	radius_out_h = diameter_h / 2;
	radius_in_w = radius_out_w - 1;
	radius_in_h = radius_out_h - 1;

	vg_lite_translate(x + radius_out_w, y + radius_out_h, &matrix);

	radius_out_w *= __DRAWING_SCALE_FACTOR;
	radius_out_h *= __DRAWING_SCALE_FACTOR;
	radius_in_w *= __DRAWING_SCALE_FACTOR;
	radius_in_h *= __DRAWING_SCALE_FACTOR;

	vg_lite_scale(__DRAWING_SCALE_DIV, __DRAWING_SCALE_DIV, &matrix);

	// 0 degrees:
	// - MicroUI: 3 o'clock.
	// - RT595 VGLite implementation : 0 o'clock.
	start_angle_deg = 90-start_angle_deg;

	shape_vg_path.path = &__shape_paths.ellipse_arc;
	shape_vg_path.path_length = sizeof(__shape_paths.ellipse_arc);
	path_offset = vglite_compute_ellipse_arc(
			&shape_vg_path,
			radius_out_w, radius_out_h,
			radius_in_w, radius_in_h,
			start_angle_deg, -arc_angle_deg, fill);

	if (0 > path_offset) {
		MEJ_LOG_ERROR("Error during path computation");
		return DRAWING_DONE;
	}

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_EVEN_ODD,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __draw_ellipse(
		MICROUI_GraphicsContext* gc,
		int x, int y,
		int diameter_out_w, int diameter_out_h,
		int diameter_in_w, int diameter_in_h) {

	int path_offset;
	int radius_w;
	int radius_h;
	vg_lite_path_t shape_vg_path;
	vg_lite_matrix_t matrix;
	vg_lite_buffer_t* dst;

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(gc, x, y, x + diameter_out_w, y + diameter_out_h, false)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);

	radius_w = diameter_out_w / 2;
	radius_h = diameter_out_h / 2;

	vg_lite_translate(x + radius_w, y + radius_h, &matrix);

	radius_w *= __DRAWING_SCALE_FACTOR;
	radius_h *= __DRAWING_SCALE_FACTOR;

	vg_lite_scale(__DRAWING_SCALE_DIV, __DRAWING_SCALE_DIV, &matrix);
	// Compute the rounded rectangle shape path
	shape_vg_path.path = &__shape_paths.ellipse[0];
	shape_vg_path.path_length = sizeof(__shape_paths.ellipse);
	path_offset = vglite_compute_ellipse(
			&shape_vg_path, 0,
			radius_w, radius_h,
			false);

	radius_w = diameter_in_w / 2;
	radius_h = diameter_in_h / 2;

	radius_w *= __DRAWING_SCALE_FACTOR;
	radius_h *= __DRAWING_SCALE_FACTOR;

	path_offset = vglite_compute_ellipse(
			&shape_vg_path,
			path_offset,
			radius_w, radius_h,
			true);

	if (0 > path_offset) {
		MEJ_LOG_ERROR("Error during path computation");
		return DRAWING_DONE;
	}

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_EVEN_ODD,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __fill_ellipse(
		MICROUI_GraphicsContext* gc,
		int x, int y,
		int diameter_w, int diameter_h) {

	int radius_w;
	int radius_h;
	vg_lite_matrix_t matrix;
	vg_lite_path_t shape_vg_path;
	vg_lite_buffer_t* dst;

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(	gc,
			x,
			y,
			x + diameter_w - 1,
			y + diameter_h - 1, 
			false)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	radius_w = diameter_w / 2;
	radius_h = diameter_h / 2;

	dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);
	vg_lite_translate(x + radius_w, y + radius_h, &matrix);

	// Compute the thick shape path
	shape_vg_path.path = &__shape_paths.ellipse;
	shape_vg_path.path_length = sizeof(__shape_paths.ellipse);
	vglite_compute_filled_ellipse(&shape_vg_path, radius_w, radius_h, &matrix);

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_NON_ZERO,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}

#ifndef MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS
// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __draw_line(
		MICROUI_GraphicsContext* gc,
		int x1, int y1,
		int x2, int y2) {

	vg_lite_path_t shape_vg_path;

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);
	vg_lite_translate(x1, y1, &matrix);

	// Compute the thick shape path
	shape_vg_path.path = &__shape_paths.line;
	shape_vg_path.path_length = sizeof(__shape_paths.line);
	vglite_compute_line(&shape_vg_path, x2 - x1, y2 - y1);

	// Draw the point with the GPU
	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_NON_ZERO,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color
	);
}
#endif // MEJ_DISABLED_FOR_PERFORMANCE_SIMPLE_DRAWINGS

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __thick_line(
		MICROUI_GraphicsContext* gc,
		int x1,
		int y1,
		int x2,
		int y2,
		int thickness,
		DRAWING_Cap start,
		DRAWING_Cap end) {

	int caps;
	vg_lite_path_t shape_vg_path;
	vg_lite_matrix_t matrix;
	vg_lite_buffer_t* dst;

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	int min_x = MIN(x1, x2);
	int min_y = MIN(y1, y2);
	int max_x = MAX(x1, x2);
	int max_y = MAX(y1, y2);

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(
			gc,
			min_x - thickness / 2,
			min_y - thickness / 2,
			max_x + thickness / 2,
			max_y + thickness / 2, 
			false)) {
		return DRAWING_DONE;
	}

	dst = display_vglite_configure_destination(gc);

	vg_lite_identity(&matrix);
	vg_lite_translate(x1, y1, &matrix);

	shape_vg_path.path = &__shape_paths.thick_shape_line;
	shape_vg_path.path_length = sizeof(__shape_paths.thick_shape_line);

	caps = 0;
	caps |= MEJ_VGLITE_PATH_SET_CAPS_START(start);
	caps |= MEJ_VGLITE_PATH_SET_CAPS_END(end);

	vglite_compute_thick_shape_line(
			&shape_vg_path, x2 - x1, y2 - y1, thickness, caps, &matrix);

	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_NON_ZERO,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __draw_thick_shape_ellipse_arc(
		MICROUI_GraphicsContext* gc,
		int x,
		int y,
		int diameter_w,
		int diameter_h,
		float start_angle_deg,
		float arc_angle,
		int thickness,
		int fade,
		DRAWING_Cap start,
		DRAWING_Cap end) {

	int translate_x;
	int translate_y;
	int caps;
	vg_lite_path_t shape_vg_path;

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(
			gc,
			x - thickness,
			y - thickness,
			x + diameter_w + 2 * thickness - 1,
			y + diameter_h + 2 * thickness - 1, 
			false)) {
		return DRAWING_DONE;
	}

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	// Tunning for CircleArcs rendering like software algorithm
	diameter_w += 1;
	diameter_h += 1;
	thickness += 1;
	translate_x = x + diameter_w / 2;
	translate_y = y + diameter_h / 2;

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);
	vg_lite_translate(translate_x, translate_y, &matrix);
	vg_lite_rotate(-start_angle_deg, &matrix);
	vg_lite_scale(__DRAWING_SCALE_DIV, __DRAWING_SCALE_DIV, &matrix);

	caps = 0;
	caps |= MEJ_VGLITE_PATH_SET_CAPS_START(start);
	caps |= MEJ_VGLITE_PATH_SET_CAPS_END(end);

	// Compute the thick shape path
	shape_vg_path.path = &__shape_paths.thick_ellipse_arc;
	shape_vg_path.path_length = sizeof(__shape_paths.thick_ellipse_arc);

	vglite_compute_thick_shape_ellipse_arc(
			&shape_vg_path,
			__DRAWING_SCALE_FACTOR*diameter_w,
			__DRAWING_SCALE_FACTOR*diameter_h,
			__DRAWING_SCALE_FACTOR*thickness,
			0,
			arc_angle,
			caps);

	return __draw_path(
			dst,
			&shape_vg_path,
			VG_LITE_FILL_NON_ZERO,
			&matrix,
			VG_LITE_BLEND_SRC_OVER,
			gc->foreground_color);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __rotate_image(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		vg_lite_buffer_t* src,
		int x,
		int y,
		int xRotation,
		int yRotation,
		float angle_deg,
		int alpha,
		vg_lite_filter_t filter) {

	vg_lite_error_t vg_lite_ret;

	// Check if there is something to draw and clip drawing limits
	// TODO try to crop region
	if (!__check_clip(
			gc,
			0, 0,
			gc->image.width, gc->image.height,
			true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);
	vg_lite_translate(xRotation, yRotation, &matrix);
	vg_lite_rotate(-angle_deg, &matrix);
	vg_lite_translate(x-xRotation, y-yRotation, &matrix);

	TRACE_VGLITE_INIT_GPU_START(BLIT);
	vg_lite_ret =  vg_lite_blit(
			dst, src,
			&matrix, VG_LITE_BLEND_SRC_OVER,
			__get_vglite_color(gc, img, alpha),
			filter);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __scale_image(
		MICROUI_GraphicsContext* gc,
		MICROUI_Image* img,
		vg_lite_buffer_t* src,
		int x,
		int y,
		float factorX,
		float factorY,
		int alpha,
		vg_lite_filter_t filter) {

	vg_lite_error_t vg_lite_ret;

	// Check if there is something to draw and clip drawing limits
	if (!__check_clip(
			gc, x, y,
			(int) (x + factorX * img->width),
			(int) (y + factorY * img->height),
			true)) {
		return DRAWING_DONE;
	}

	MEJ_DEBUG(MEJ_DEBUG_MICROUI_DRAWINGS_LVL, "Enter\n");

	vg_lite_buffer_t* dst = display_vglite_configure_destination(gc);

	vg_lite_matrix_t matrix;
	vg_lite_identity(&matrix);
	vg_lite_translate(x, y, &matrix);
	vg_lite_scale(factorX, factorY, &matrix);

	TRACE_VGLITE_INIT_GPU_START(BLIT);
	vg_lite_ret =  vg_lite_blit(
			dst, src,
			&matrix, VG_LITE_BLEND_SRC_OVER,
			__get_vglite_color(gc, img, alpha),
			filter);
	TRACE_VGLITE_INIT_GPU_END();

	return __check_operation(vg_lite_ret);
}

// See the section 'Internal function definitions' for the function documentation
static DRAWING_Status __check_operation(vg_lite_error_t vg_lite_error) {

	DRAWING_Status ret;

	if (vg_lite_error == VG_LITE_SUCCESS) {
		display_vglite_start_operation();
		vg_lite_flush();
		display_task_wakeup(false);
		ret = DRAWING_RUNNING;
	} else {
		MEJ_LOG_ERROR("vg_lite operation failed with error: %d\n", ret);
		ret = DRAWING_DONE;
	}

	return ret;
}

static vg_lite_color_t __get_vglite_color(MICROUI_GraphicsContext* gc, MICROUI_Image* img, jint alpha)
{
	jint fc;

	switch(img->format)
	{
	case MICROUI_IMAGE_FORMAT_A4:
	case MICROUI_IMAGE_FORMAT_A8:
		if (0xff == alpha)
		{
			fc = gc->foreground_color | 0xff000000;
		}
		else
		{
			fc = gc->foreground_color;
			jint red = alpha * ((fc >> 16) & 0xff) / 0xff;
			jint green = alpha * ((fc >> 8) & 0xff) / 0xff;
			jint blue = alpha * (fc & 0xff) / 0xff;
			fc = (alpha << 24) | (red << 16) | (green << 8) | blue;
		}
		break;
	default:
		fc = alpha * 0x01010101;
		break;
	}

	return (vg_lite_color_t)fc;
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------


