/*
 * C
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <math.h>

#include "display_support.h"

#include "LLUI_PAINTER_impl.h"

#include "LLUI_DISPLAY.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * Retrieves a pixel address according pixel position
 */
#define ADDR16(buf,x,y,w) ((buf) + ((y) * (w) + (x)) * 2)

/*
 * Returns the image buffer width and not the image area width.
 */
#define GET_BUFFER_WIDTH(image_ID) DEMO_BUFFER_WIDTH

/*
 * MicroEJ application natives
 */
#define MICROUI_ROTATION_rotate_circular_image_with_array Java_com_microej_microui_RotateCircular_drawCircularImageWithArray

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

int32_t MICROUI_ROTATION_rotate_circular_image_with_array(MICROUI_GraphicsContext* gc, MICROUI_Image * src, int32_t* circular_area, int32_t angle)
{
	// sanity checks
	if (src->format != MICROUI_IMAGE_FORMAT_RGB565 || gc->image.format != MICROUI_IMAGE_FORMAT_RGB565)
	{
		// not supported format
		return -1;
	}

	if (gc->image.width != src->width || gc->image.height != src->height)
	{
		// graphics contexts must have the same size
		return -1;
	}
	if (src->width != src->height)
	{
		// image must be a square
		return -1;
	}

	if (!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&MICROUI_ROTATION_rotate_circular_image_with_array)) {
		return -1;
	}

	// retrieve src & dest characteristics
	uint8_t* src_address = (uint8_t*)LLUI_DISPLAY_getBufferAddress(src);
	uint8_t* dest_address = (uint8_t*)LLUI_DISPLAY_getBufferAddress(&gc->image);
	int32_t area_width = gc->image.width;
	int32_t area_height = gc->image.height;
	int32_t src_stride_in_pxl = LLUI_DISPLAY_getStrideInPixels(src);
	int32_t dest_stride_in_pxl = LLUI_DISPLAY_getStrideInPixels(&gc->image);

	// get angle in radian
	float angleRad = ((float)angle * 3.1415927f) / (float)180.0f;
	volatile float cosAngle = (float)cosf(angleRad);
	volatile float sinAngle = (float)(-sinf(angleRad));

	// retrieve top-left coordinate on source (prevent to perform some multiplications in loops)
	int32_t xCosMul = -area_width / 2;
	int32_t yCosMul = -area_height / 2;
	volatile float xSrcFromSrcOriginBase = (float)xCosMul * cosAngle + (float)yCosMul * sinAngle + (float)(area_width / 2);
	volatile float ySrcFromSrcOriginBase = (float)(-xCosMul) * sinAngle + (float)yCosMul * cosAngle + (float)(area_height / 2);
	uint32_t circular_area_index = 0;

	for(uint32_t y = 0; y < area_height; y++)
	{
		float xSrcFromSrcOrigin = xSrcFromSrcOriginBase;
		float ySrcFromSrcOrigin = ySrcFromSrcOriginBase;

		// part 1: nothing to draw
		xSrcFromSrcOrigin += (cosAngle * circular_area[circular_area_index]);
		ySrcFromSrcOrigin -= (sinAngle * circular_area[circular_area_index]);
		uint32_t x = circular_area[circular_area_index];
		++circular_area_index;

		// part 2: have to draw
		for(; x < circular_area[circular_area_index]; x++)
		{
			if (((int32_t)xSrcFromSrcOrigin >= 0) && ((int32_t)ySrcFromSrcOrigin >= 0) && ((int32_t)xSrcFromSrcOrigin < area_width) && ((int32_t)ySrcFromSrcOrigin < area_height))
			{
				// retrieve source pixel address
				uint16_t* src_pix_addr = (uint16_t*)ADDR16(src_address, (int32_t)xSrcFromSrcOrigin, (int32_t)ySrcFromSrcOrigin, src_stride_in_pxl);

				// retrieve destination pixel address according LCD buffer width
				uint16_t* dest_pix_addr = (uint16_t*)ADDR16(dest_address, x, y, dest_stride_in_pxl);

				// copy source into destination
				*dest_pix_addr = *src_pix_addr;
			}
			// else: source pixel is out of bounds: nothing to draw on destination

			xSrcFromSrcOrigin += cosAngle;
			ySrcFromSrcOrigin -= sinAngle;
		}
		++circular_area_index;

		// part 3: nothing to draw
		xSrcFromSrcOrigin += (cosAngle * (circular_area[circular_area_index] - x + 1));
		ySrcFromSrcOrigin -= (sinAngle * (circular_area[circular_area_index] - x + 1));
		x = circular_area[circular_area_index];
		++circular_area_index;                          

		// part 4: have to draw
		for(; x < circular_area[circular_area_index]; x++)
		{
			if (((int32_t)xSrcFromSrcOrigin >= 0) && ((int32_t)ySrcFromSrcOrigin >= 0) && ((int32_t)xSrcFromSrcOrigin < area_width) && ((int32_t)ySrcFromSrcOrigin < area_height))
			{
				// retrieve source pixel address
				uint16_t* src_pix_addr = (uint16_t*)ADDR16(src_address, (int32_t)xSrcFromSrcOrigin, (int32_t)ySrcFromSrcOrigin, src_stride_in_pxl);

				// retrieve destination pixel address according LCD buffer width
				uint16_t* dest_pix_addr = (uint16_t*)ADDR16(dest_address, x, y, dest_stride_in_pxl);

				// copy source into destination
				*dest_pix_addr = *src_pix_addr;
			}
			// else: source pixel is out of bounds: nothing to draw on destination

			xSrcFromSrcOrigin += cosAngle;
			ySrcFromSrcOrigin -= sinAngle;
		}
		++circular_area_index;

		// go to next line
		xSrcFromSrcOriginBase += sinAngle;
		ySrcFromSrcOriginBase += cosAngle;
	}

	if (LLUI_DISPLAY_isLCD(&gc->image))
	{
		// update flush limits: full screen has been updated
		LLUI_DISPLAY_setDrawingLimits(gc, 0, 0, area_width - 1, area_height - 1);
	}

	LLUI_DISPLAY_setDrawingStatus(DRAWING_DONE);

	return 0; // no error
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------
