/**************************************************************************//**
 * Copyright 2018, Silicon Laboratories Inc.
 *
 * Copyright 2019-2021 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *****************************************************************************/

/**************************************************************************//**
 * Host specific implementation: LPC546XX + FreeRTOS
 *****************************************************************************/

#include "sl_wfx.h"
#include "sl_wfx_host.h"
#include "sl_wfx_host_pin.h"
#include "BRD802xx_pds.h"
#include "FreeRTOS.h"
#include "queue.h"
#include "semphr.h"
#include <stdlib.h>
#include <string.h>
#include "fsl_gpio.h"
#include "netif/etharp.h"
#include "dhcpserver.h"

//#define SPI_HOST_DEBUG

#define WFX_RESET_DELAY_IN_MS   100    // delay to stay in and after reset in milliseconds.

#define MAX_NUMB_OF_SCANS 32

#if   defined( SL_WFX_ALPHA_KEY )
#include "sl_wfx_wf200_A0.h"
#elif defined( SL_WFX_PROD_KEY )
#include "sl_wfx_wf200_C0.h"
#else
#error Must define either SL_WFX_ALPHA_KEY/SL_WFX_PROD_KEY
#endif

extern uint8_t soft_ap_mode;
extern TaskHandle_t	dhcp_task_handle;
extern uint8_t dhcp_sleeping;
extern int   memcpy_s(void *_Restrict_dst, size_t, const void *_Restric_src, size_t);
extern void User_notification(struct netif *netif);

/**
  * @brief  This function notify user about link status changement.
  * @param  netif: the network interface
  * @retval None
  */
__weak void ethernetif_notify_conn_changed(struct netif *netif)
{
  /* NOTE : This is function could be implemented in user file
            when the callback is needed,
  */
}

sl_wfx_state_t sl_wfx_status;
QueueHandle_t eventQueue;
uint8_t scan_index = 0;
uint8_t waited_event_id = 0;
SemaphoreHandle_t HostLockUnlockSemaphore;

sl_wfx_scan_result_ind_body_t gpst_available_ap_list[MAX_NUMB_OF_SCANS];

EventGroupHandle_t gst_wf200_event_group;

struct
{
    uint32_t sl_wfx_firmware_download_progress;
}host_context;


/* Initialization phase*/
sl_status_t sl_wfx_host_init(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  host_context.sl_wfx_firmware_download_progress = 0;
  eventQueue = xQueueCreate(1, sizeof( uint8_t ));

  if(0 == eventQueue)
  {
    PRINTF("[ERROR] xQueueCreate fail at %s\n", __func__);
    return SL_STATUS_FAIL;
  }

  /* Create mutex needed for  sl_wfx_host_lock and sl_wfx_host_unlock APIs*/
  HostLockUnlockSemaphore = xSemaphoreCreateMutex();
  xSemaphoreGive(HostLockUnlockSemaphore);

  /* Create the wifi event group */
  gst_wf200_event_group = xEventGroupCreate();
  if(gst_wf200_event_group == NULL)
  {
	  PRINTF("[ERROR]Event groups failed to be created in %s!\n", __func__);
  }

  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_get_firmware_data(const uint8_t** data, uint32_t data_size)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  *data = &sl_wfx_firmware[host_context.sl_wfx_firmware_download_progress];
  host_context.sl_wfx_firmware_download_progress += data_size;
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_get_firmware_size(uint32_t* firmware_size)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  *firmware_size = sizeof(sl_wfx_firmware);
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_get_pds_data(const char **pds_data, uint16_t index)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  *pds_data = wf200_pds[index];
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_get_pds_size(uint16_t *pds_size)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  *pds_size = SL_WFX_ARRAY_COUNT(wf200_pds);
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_deinit(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  return SL_STATUS_OK;
}

/* GPIO interface */
sl_status_t sl_wfx_host_reset_chip(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif

  GPIO_WritePinOutput(GPIO, SL_WFX_RESET_PORT, SL_WFX_RESET_GPIO, 0);
  TickType_t ticks = WFX_RESET_DELAY_IN_MS / portTICK_PERIOD_MS;
  vTaskDelay(ticks);

  // hold pin high to get chip out of reset
  GPIO_WritePinOutput(GPIO, SL_WFX_RESET_PORT, SL_WFX_RESET_GPIO, 1);
  vTaskDelay(ticks);
  
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_hold_in_reset(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
    GPIO_WritePinOutput(GPIO, SL_WFX_RESET_PORT, SL_WFX_RESET_GPIO, 0);
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_set_wake_up_pin(uint8_t state)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif

  if ( state > 0 )
  {
	  GPIO_WritePinOutput(GPIO, SL_WFX_WUP_PORT, SL_WFX_WUP_GPIO, 1);
  }
  else
  {
	  GPIO_WritePinOutput(GPIO, SL_WFX_WUP_PORT, SL_WFX_WUP_GPIO, 0);
  }
  
  return SL_STATUS_OK;

}

sl_status_t sl_wfx_host_wait_for_wake_up(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  vTaskDelay(2000);
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_wait_for_confirmation(uint8_t confirmation_id ,uint32_t timeout, void** event_payload_out)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  uint8_t posted_event_id;
  waited_event_id = confirmation_id;
  for(uint32_t i = 0; i < timeout; i++)
  {
    if(xQueueReceive(eventQueue, &(posted_event_id), 1))
    {
            if (waited_event_id == posted_event_id)
            {
#ifdef SPI_HOST_DEBUG
            PRINTF("event %#08X \n\r", posted_event_id);
#endif

                    if (event_payload_out != NULL)
                    {
                      *event_payload_out = sl_wfx_context->event_payload_buffer;
                    }
                    return SL_STATUS_OK;
            }
      }
  }
  PRINTF("[WARNING] Timeout in %s\n", __func__);
  return SL_STATUS_TIMEOUT;
}

sl_status_t sl_wfx_host_wait(uint32_t wait_time)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  TickType_t ticks = wait_time / portTICK_PERIOD_MS;
  vTaskDelay(ticks);
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_post_event(sl_wfx_generic_message_t *network_rx_buffer)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
  PRINTF("[DEBUG] id is %x\n", network_rx_buffer->header.id);
#endif
  switch(network_rx_buffer->header.id){
  /******** INDICATION ********/
  case SL_WFX_CONNECT_IND_ID:
    {
      sl_wfx_connect_ind_t* connect_indication = (sl_wfx_connect_ind_t*) network_rx_buffer;
      sl_wfx_connect_callback(connect_indication->body.mac, connect_indication->body.status);
      break;
    }
  case SL_WFX_DISCONNECT_IND_ID:
    {
      sl_wfx_disconnect_ind_t* disconnect_indication = (sl_wfx_disconnect_ind_t*) network_rx_buffer;
      sl_wfx_disconnect_callback(disconnect_indication->body.mac, disconnect_indication->body.reason);
      break;
    }
  case SL_WFX_START_AP_IND_ID:
    {
      sl_wfx_start_ap_ind_t* start_ap_indication = (sl_wfx_start_ap_ind_t*) network_rx_buffer;
      sl_wfx_start_ap_callback(start_ap_indication->body.status);
      break;
    }
  case SL_WFX_STOP_AP_IND_ID:
    {
      sl_wfx_stop_ap_callback();
      break;
    }
  case SL_WFX_RECEIVED_IND_ID:
    {
      sl_wfx_received_ind_t* ethernet_frame = (sl_wfx_received_ind_t*) network_rx_buffer;
      if (ethernet_frame->body.frame_type == 0)
      {
        sl_wfx_host_received_frame_callback( ethernet_frame );
      }
      break;
    }
  case SL_WFX_SCAN_RESULT_IND_ID:
    {
      sl_wfx_scan_result_ind_t* scan_result = (sl_wfx_scan_result_ind_t*) network_rx_buffer;
      sl_wfx_scan_result_callback(&scan_result->body);
      break;
    }
  case SL_WFX_SCAN_COMPLETE_IND_ID:
    {
      sl_wfx_scan_complete_ind_t* scan_complete = (sl_wfx_scan_complete_ind_t*) network_rx_buffer;
      sl_wfx_scan_complete_callback(scan_complete->body.status);
      break;
    }
  case SL_WFX_AP_CLIENT_CONNECTED_IND_ID:
    {
      sl_wfx_ap_client_connected_ind_t* client_connected_indication = (sl_wfx_ap_client_connected_ind_t*) network_rx_buffer;
      sl_wfx_client_connected_callback(client_connected_indication->body.mac);
      break;
    }
  case SL_WFX_AP_CLIENT_REJECTED_IND_ID:
    {
      break;
    }
  case SL_WFX_AP_CLIENT_DISCONNECTED_IND_ID:
    {
      sl_wfx_ap_client_disconnected_ind_t* ap_client_disconnected_indication = (sl_wfx_ap_client_disconnected_ind_t*) network_rx_buffer;
      sl_wfx_ap_client_disconnected_callback(ap_client_disconnected_indication->body.reason, ap_client_disconnected_indication->body.mac);
      break;
    }
  case SL_WFX_EXCEPTION_IND_ID:
    {
      PRINTF("Firmware Exception\r\n");
      break;
    }
  case SL_WFX_ERROR_IND_ID:
    {
      PRINTF("Firmware Error\r\n");
      break;
    }
  /******** CONFIRMATION ********/
  case SL_WFX_SEND_FRAME_CNF_ID:
    {
      if (sl_wfx_context->used_buffers > 0)
      {
        sl_wfx_context->used_buffers--;
      }
      break;
    }
  default:
	{
	  //Confirmation are sent to the event queue
	}
  }    
  if (waited_event_id == network_rx_buffer->header.id)
  {
    if(network_rx_buffer->header.length < 512){
      /* Post the event in the queue */
      memcpy(sl_wfx_context->event_payload_buffer, (void*) network_rx_buffer,network_rx_buffer->header.length);
      xQueueOverwrite(eventQueue, ( void * ) &network_rx_buffer->header.id);
    }
  }
  return SL_STATUS_OK;
}

/* Memory management */
sl_status_t sl_wfx_host_allocate_buffer(void** buffer, sl_wfx_buffer_type_t type, uint32_t buffer_size)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  SL_WFX_UNUSED_PARAMETER(type);
  *buffer = pvPortMalloc(buffer_size);
  if(*buffer == NULL)
  {
	  PRINTF("[ERROR]buffer alocated is NULL\n");
	  return SL_STATUS_FAIL;
  }
  return SL_STATUS_OK;
}

sl_status_t sl_wfx_host_free_buffer(void* buffer, sl_wfx_buffer_type_t type)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  SL_WFX_UNUSED_PARAMETER(type);
  vPortFree(buffer);
  return SL_STATUS_OK;
}

/* Frame hook */
sl_status_t sl_wfx_host_transmit_frame(void* frame, uint32_t frame_len)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  return sl_wfx_data_write(frame, frame_len);
}

sl_status_t sl_wfx_host_setup_waited_event(uint8_t event_id)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  waited_event_id = event_id;
  return SL_STATUS_OK;
}

/** Callback for individual AP discovered
 */
void sl_wfx_scan_result_callback(sl_wfx_scan_result_ind_body_t* scan_result)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  if(scan_index < MAX_NUMB_OF_SCANS)
  {
	  if(memcpy_s((char*)gpst_available_ap_list[scan_index].ssid_def.ssid,
                   sizeof(gpst_available_ap_list[scan_index].ssid_def.ssid),
                  (const char*)(scan_result->ssid_def.ssid),
                   sizeof(gpst_available_ap_list[scan_index].ssid_def.ssid)))
      {
         PRINTF("[ERROR] Copy fail at %s %d\n", __FILE__, __LINE__);
      }

      gpst_available_ap_list[scan_index].ssid_def.ssid_length = scan_result->ssid_def.ssid_length;
	  gpst_available_ap_list[scan_index].rcpi = scan_result->rcpi;
	  gpst_available_ap_list[scan_index].mac[0] = scan_result->mac[0];
	  gpst_available_ap_list[scan_index].mac[1] = scan_result->mac[1];
	  gpst_available_ap_list[scan_index].mac[2] = scan_result->mac[2];
	  gpst_available_ap_list[scan_index].mac[3] = scan_result->mac[3];
	  gpst_available_ap_list[scan_index].mac[4] = scan_result->mac[4];
	  gpst_available_ap_list[scan_index].mac[5] = scan_result->mac[5];
	  gpst_available_ap_list[scan_index].channel = scan_result->channel;
	  gpst_available_ap_list[scan_index].security_mode = scan_result->security_mode;
	  scan_index++;
  }else
  {
	  /*Do nothing. The list is full.*/
      PRINTF("[Warning] Scan result list is full. More that %d network discovered.\n", MAX_NUMB_OF_SCANS);
  }

}

/** Callback to indicate the scan completion
 */
void sl_wfx_scan_complete_callback(uint32_t status)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  xEventGroupSetBits(gst_wf200_event_group, SL_WFX_SCAN_COMPLETE_IND_ID);

//#define DISPLAY_SCANNED_SSIDS
#ifdef DISPLAY_SCANNED_SSIDS
  PRINTF("Scan completed, found %d items.\n", scan_index);
  for ( uint8_t i = 0  ; i < scan_index ; i++ )
  {
     PRINTF("%d.%s\n", i, gpst_available_ap_list[i].ssid_def.ssid );
  }
#endif

  scan_index = 0;
}

/** Callback triggered when a connection is established as station
 */
void sl_wfx_connect_callback(uint8_t* mac, uint32_t status)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif

  if(status == 0)
  {
    xEventGroupSetBits(gst_wf200_event_group, SL_WFX_CONNECT_IND_ID);
    
	/*set the mode and mac addr*/
	soft_ap_mode = STATION_MODE;

	// resume DHCP thread
	dhcp_sleeping = 0;

	if(NULL != dhcp_task_handle)
	{
		vTaskResume(dhcp_task_handle);
	}

    sl_wfx_status |= SL_WFX_STA_INTERFACE_CONNECTED;
    //PRINTF("Connected\r\n");
    

  }else{
    PRINTF("Connection attempt failed\r\n");
  }
}

/** Callback triggered when connection is disconnected
 */
void sl_wfx_disconnect_callback(uint8_t* mac, uint16_t reason)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  
  xEventGroupSetBits(gst_wf200_event_group, SL_WFX_DISCONNECT_IND_ID);
  sl_wfx_status &= ~SL_WFX_STA_INTERFACE_CONNECTED;
  PRINTF("Disconnected reason=%u\r\n", reason);
}

/** Callback triggered when a softap is started
 */
void sl_wfx_start_ap_callback(uint32_t status)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif

  if(status == 0)
  {
	xEventGroupSetBits(gst_wf200_event_group, SL_WFX_START_AP_IND_ID);
    
    sl_wfx_status |= SL_WFX_AP_INTERFACE_UP;
    /*set the mode*/
    soft_ap_mode = SOFTAP_MODE;
    PRINTF("AP started\r\n");
  }else{
    PRINTF("AP start failed\r\n");
  }
}

/** Callback triggered when a softap is stopped
 */
void sl_wfx_stop_ap_callback(void)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
  
  xEventGroupSetBits(gst_wf200_event_group, SL_WFX_STOP_AP_IND_ID);
  sl_wfx_status &= ~SL_WFX_AP_INTERFACE_UP;
  PRINTF("AP stopped\r\n");
}

/** Callback triggered when a client connects
 */
void sl_wfx_client_connected_callback(uint8_t* mac)
{
  PRINTF("Client connected\r\n");
}

/** Callback triggered when a client disconnects
 */
void sl_wfx_ap_client_disconnected_callback(uint32_t status, uint8_t* mac)
{
  PRINTF("Client disconnected\r\n");
}
sl_status_t sl_wfx_host_sleep_grant(sl_wfx_host_bus_transfer_type_t type,
                                    sl_wfx_register_address_t address,
                                    uint32_t length)
{
#ifdef SPI_HOST_DEBUG
  PRINTF("[DEBUG] %s\n", __func__);
#endif
	return SL_STATUS_OK;
}

// sl_wfx_host_lock and sl_wfx_host_unlock need to be tested
/** Called when the driver needs to lock its access
*/
sl_status_t sl_wfx_host_lock(void)
{
	if( HostLockUnlockSemaphore != NULL )
	{
		if(xSemaphoreTake(HostLockUnlockSemaphore, portMAX_DELAY) == pdTRUE)
		{
			return SL_STATUS_OK;
		}
	}

	return SL_STATUS_FAIL;
	// return SL_STATUS_OK;
}

/** Called when the driver needs to unlock its access
*/
sl_status_t sl_wfx_host_unlock(void)
{
	if( HostLockUnlockSemaphore != NULL )
	{
		if(xSemaphoreGive(HostLockUnlockSemaphore) == pdTRUE )
		{
			return SL_STATUS_OK;
		}
	}
	return SL_STATUS_FAIL;
	//return SL_STATUS_OK;
}

extern int VPRINTF(const char *formatString, va_list ap);
void sl_wfx_host_log(const char *string, ...)
{
	va_list  args;

	va_start( args, string );
	VPRINTF( string, args );
	va_end( args );
}
