/*
 * Java
 *
 * Copyright 2021 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.kernel;

import java.io.IOException;
import java.io.InputStream;

/**
 * Input stream to load a feature.
 */
public class FeatureInputStream extends InputStream {

	/**
	 * Waits for a features input stream to be available and returns it.
	 *
	 * @return a feature input stream or null.
	 */
	public static InputStream getInputStream() {
		return isFeatureAvailable() ? new FeatureInputStream() : null;
	}

	private FeatureInputStream() {
		// private constructor
	}

	@Override
	public int read() throws IOException {
		byte[] b = new byte[1];
		int read = read(b, 0, 1);
		return (read > 0) ? b[0] : -1;
	}

	@Override
	public int read(byte[] out, int outOffset, int outLen) throws IOException {
		// Input sanity check
		if (out == null || outOffset < 0 || outLen > (out.length - outOffset) || outLen < 0) {
			throw new IllegalArgumentException();
		}
		return readIntoArray(out, outOffset, outLen);
	}

	@Override
	public void close() throws IOException {
		closeFeature();
	}

	/**
	 * Checks whether a feature is available.
	 *
	 * @return <code>true</code> if the feature is available, <code>false</code> otherwise.
	 */
	private static native boolean isFeatureAvailable();

	/**
	 * Reads a chunk of the feature into the given array.
	 *
	 * @param outOffset
	 *            write start offset in array
	 * @param length
	 *            number of bytes to read
	 * @return the number of bytes read or -1 if EOF
	 */
	private static native int readIntoArray(byte[] array, int outOffset, int length);

	/**
	 * Closes the associated resources.
	 */
	private static native void closeFeature();

	/**
	 * Initializes the stack. Called once before any other function.
	 */
	public static native void init();
}
