/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.microui;

import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;

/**
 * This class allows to draw an image with a rotation using an accelerated algorithm. This algorithm is optimized when
 * the following characteristics are respected (otherwise an {@link IllegalArgumentException} is thrown).
 * <ul>
 * <li>The image to draw has exactly the same size than the destination</li>
 * <li>The image is a square</li>
 * <li>The image format is {@link ej.microui.display.ResourceImage.OutputFormat#RGB565}</li>
 * <li>The display format is {@link ej.microui.display.ResourceImage.OutputFormat#RGB565}</li>
 * <li>No mirror can be applied</li>
 * <li>No global transparent (alpha) can be applied</li>
 * <li>Only a circular area is drawn (not the full image)</li>
 * </ul>
 */
public class RotateCircular {

	private static final int IMAGE_HEIGHT_RATIO = 4;

	private RotateCircular() {
		// private constructor: this class cannot be instantiated
	}

	/**
	 * Rotates an image using nearest neighbor algorithm. This algorithm is faster than <code>bilinear</code> algorithm
	 * but its rendering is more simple.
	 * <p>
	 * Rotates only a subpart of image. This sub part is defined by the given circular area. This area has been
	 * previously extracted from image to render using the external tool <code>CircularRotationExtractor</code>.
	 *
	 * @param gc
	 *            the graphics context where draw the rotated image.
	 * @param image
	 *            the image to render
	 * @param circularArea
	 *            the circular area to rotate
	 * @param angle
	 *            the rotation angle
	 */
	public static void drawNearestNeighborCircularImage(GraphicsContext gc, Image image, int[] circularArea,
			int angle) {
		if (circularArea.length != image.getHeight() * IMAGE_HEIGHT_RATIO) {
			// circular area must contain 4 elements for each line
			throw new IllegalArgumentException();
		}
		checkResult(drawCircularImageWithArray(gc.getSNIContext(), image.getSNIContext(), circularArea, angle));
	}

	private static void checkResult(int result) {
		if (result != 0) {
			// image and/or graphics context are not in RGB888 format
			throw new IllegalArgumentException();
		}
	}

	private static native int drawCircularImageWithArray(byte[] gc, byte[] img, int[] circularArea, int angle);
}
