/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.path.basic;

import static com.verisilicon.vivante.vg_lite_matrix.vg_lite_identity;

import com.microej.vivante.utils.GradientPath;
import com.verisilicon.vivante.vg_lite_gradient;
import com.verisilicon.vivante.vg_lite_matrix;

import ej.drawing.ShapePainter.Cap;

/**
 * A path that represents a circle arc with a gradient.
 */
public class CircleArcGradient extends CircleArc {

	private static final float GRADIENT_BUFFER_SIZE = 256F;

	private float gradientAngle;

	private final boolean fixedGradient;

	/**
	 * Creates a path that represents a circle arc with a gradient.
	 *
	 * <p>
	 * When setting <code>fixedGradient</code> to <code>false</code>, the gradient angle is the set to be equal to the
	 * arc angle, otherwise it is set to the angle specified in the <code>gradientStyle</code>.<br>
	 *
	 * @param diameter
	 *            the diameter of the circular shape.
	 * @param thickness
	 *            the thickness of the shape.
	 * @param startAngle
	 *            the beginning angle of the arc.
	 * @param cap
	 *            the cap to use for the start and end of the shape.
	 * @param fixedGradient
	 *            <code>true</code> to set a fixed gradient, <code>false</code> otherwise.
	 */
	/* package */ CircleArcGradient(GradientPath path, int diameter, int thickness, int startAngle, Cap cap,
			boolean fixedGradient, int gradientAngle) {
		super(path, diameter, thickness, startAngle, cap);
		this.fixedGradient = fixedGradient;
		this.gradientAngle = gradientAngle;
	}

	/**
	 * Gets the VGLite gradient of the vector image.
	 *
	 * @return the VGLite gradient.
	 */
	public vg_lite_gradient getGradient() {
		return ((GradientPath) getPath()).getGradient();
	}

	@Override
	public void close() {
		super.close();
		getGradient().close();
	}

	@Override
	protected void updateMatrices() {
		super.updateMatrices();

		vg_lite_gradient gradient = getGradient();
		assert gradient != null;
		vg_lite_matrix gradientMatrix = gradient.getMatrix();
		vg_lite_identity(gradientMatrix);
		vg_lite_matrix.vg_lite_translate(getCenterX(), getCenterY(), gradientMatrix);
		vg_lite_matrix.vg_lite_rotate(-this.gradientAngle, gradientMatrix);
		int diameter = getDiameter();
		int thickness = getThickness();
		float radius = (diameter + thickness) / 2f;
		vg_lite_matrix.vg_lite_translate(-radius, 0f, gradientMatrix);
		float scaleX = (diameter + thickness) / GRADIENT_BUFFER_SIZE;
		vg_lite_matrix.vg_lite_scale(scaleX, 1f, gradientMatrix);
	}

	@Override
	public void setArcAngle(float arcAngle) {
		if (!this.fixedGradient) {
			this.gradientAngle = getStartAngle() + arcAngle;
		}
		super.setArcAngle(arcAngle);
	}
}
