/*
 * Java
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.path.mascot;

import com.microej.demo.watch.util.path.VectorGraphics;
import com.microej.vivante.utils.SvgImage;
import com.verisilicon.vivante.vg_lite_matrix;

/**
 * A vector graphics that represents the NXP logo.
 */
public class Nxp extends VectorGraphics {

	private static final int LOGO_HALF_WIDTH = 125;

	private static final int LOGO_HALF_HEIGHT = 45;

	private static final float SCALE = 0.53f;

	private static final int OFFSET = 32;

	private final int width;

	private final float speedFactor;

	private float position;

	private boolean back;

	/**
	 * Creates the NXP logo by describing its path.
	 *
	 * @param widgetWidth
	 *            the width of the widget containing the logo
	 * @param speedFactor
	 *            a scalar to tune the logo speed
	 */
	public Nxp(int widgetWidth, float speedFactor) {
		super(SvgImage.getSvgImage("nxp.svg")); //$NON-NLS-1$
		this.width = widgetWidth - (OFFSET * 2);
		this.speedFactor = speedFactor;
		initialize();
	}

	private void initialize() {
		nextFrame();
	}

	/**
	 * Tells whether the logo is moving on to the front.
	 *
	 * @return true if the logo is moving from back to front, false otherwise
	 */
	public boolean isMovingOnFront() {
		return !this.back;
	}

	/**
	 * Moves to the next frame of the mascot animation.
	 */
	public void nextFrame() {
		vg_lite_matrix matrix = getMatrix();
		vg_lite_matrix.vg_lite_identity(matrix);

		final float ratio = getRatio();

		// when the logo touches a border:
		// - the width is reduced to 2 pixels (devided by the half of its width),
		// - the height is devided by 3
		float scaleX = (1 + ((LOGO_HALF_WIDTH - 1) * ratio)) / LOGO_HALF_WIDTH;
		float scaleY = (1 + (2 * ratio)) / 3f;

		if (this.back) {
			// when moving in the back, the ratio is inverted, the width increases and the height decreases as it moves
			// to the center.
			final float backRatio = 1 + ratio * 8f;
			scaleX /= -backRatio;
			scaleY /= backRatio;
		}

		final float x = OFFSET + this.position - LOGO_HALF_WIDTH * scaleX;
		final float y = 155 + LOGO_HALF_HEIGHT - (LOGO_HALF_HEIGHT * scaleY);

		vg_lite_matrix.vg_lite_translate(x, y, matrix);
		vg_lite_matrix.vg_lite_scale(SCALE * scaleX, SCALE * scaleY, matrix);

		this.position += this.speedFactor * (this.back ? -2 : 1);

		if (this.position <= 0) {
			this.back = false;
		} else if (this.position >= this.width) {
			this.back = true;
		}
	}

	private float getRatio() {
		if (this.position <= (this.width / 2)) {
			// the logo is on the left side, it grows when moving from left to right
			return (2f * this.position) / this.width;
		} else {
			// the logo is on the right side, it decreases when moving from left to right
			return (2f * (this.width - this.position)) / this.width;
		}
	}
}
