/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.demo.watch.watchface.flower.path;

import com.microej.demo.watch.util.path.GradientPathGraphics;
import com.microej.demo.watch.util.path.GradientStyle;
import com.microej.vivante.utils.GradientPath;
import com.microej.vivante.utils.PathBufferFP32;
import com.verisilicon.vivante.vg_lite;
import com.verisilicon.vivante.vg_lite_gradient;
import com.verisilicon.vivante.vg_lite_matrix;
import com.verisilicon.vivante.vg_lite_path;

/**
 * A vector graphics that represents the gradient hand of the flower watchface.
 */
public class FlowerSecondGradient extends GradientPathGraphics {

	private float angle;

	private final int gradientAngle;

	private int rotationCenterX;

	private int rotationCenterY;

	/**
	 * Creates the hand by describing its path and gradient.
	 *
	 * @param gradientStyle
	 *            the gradient information.
	 * @param diameter
	 *            face diameter
	 */
	public FlowerSecondGradient(GradientStyle gradientStyle, int diameter) {
		super(createPath(diameter), gradientStyle);
		this.gradientAngle = gradientStyle.getAngle();
	}

	private static GradientPath createPath(int diameter) {

		PathBufferFP32 pathBuffer = new PathBufferFP32();
		pathBuffer.move(0, -diameter);
		pathBuffer.line(-diameter, -diameter);
		pathBuffer.line(-diameter, diameter);
		pathBuffer.line(0, diameter);

		return new GradientPath(pathBuffer.toVGLitePath(-diameter, -diameter, 0, diameter, vg_lite_path.VG_LITE_HIGH),
				new vg_lite_gradient(), vg_lite.VG_LITE_FILL_EVEN_ODD);
	}

	/**
	 * Sets the rotation center point coordinates.
	 *
	 * @param x
	 *            the rotation center x-coordinate to set.
	 * @param y
	 *            the rotation center y-coordinate to set.
	 *
	 */
	public void setRotationCenter(int x, int y) {
		this.rotationCenterX = x;
		this.rotationCenterY = y;
	}

	@Override
	protected void updateMatrices() {
		vg_lite_matrix pathMatrix = getMatrix();
		vg_lite_matrix.vg_lite_identity(pathMatrix);
		vg_lite_matrix.vg_lite_translate(this.rotationCenterX, this.rotationCenterY, pathMatrix);
		vg_lite_matrix.vg_lite_rotate(this.angle, pathMatrix);

		vg_lite_gradient gradient = getGradient();
		vg_lite_matrix gradientMatrix = gradient.getMatrix();
		vg_lite_matrix.vg_lite_identity(gradientMatrix);
		vg_lite_matrix.vg_lite_translate(this.rotationCenterX, this.rotationCenterY, gradientMatrix);
		vg_lite_matrix.vg_lite_rotate(this.angle, gradientMatrix);
		vg_lite_matrix.vg_lite_translate(0, -this.rotationCenterY, gradientMatrix);
		vg_lite_matrix.vg_lite_rotate(this.gradientAngle, gradientMatrix);
	}

	/**
	 * Sets the angle of the gradient.
	 *
	 * @param angle
	 *            the angle of rotation.
	 */
	public void setAngle(float angle) {
		this.angle = angle;
		updateMatrices();
	}
}
