/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.model;

import java.util.Deque;
import java.util.LinkedList;
import java.util.Objects;

/**
 * A child is a model for GIML containers. Children are declared thought methods.
 */
public class Child {

	private final String _name; // NOSONAR
	private final String _prefix; // NOSONAR
	private final int _cardinality; // NOSONAR
	private final boolean _ordered; // NOSONAR
	private final boolean _disabled; // NOSONAR
	private final Deque<Attribute> _attributes; // NOSONAR

	/**
	 * Instantiates a new child.
	 *
	 * @param name
	 *            the name of this child element (can be {@code null} or empty)
	 * @param prefix
	 *            the prefix of the method that declared this child.
	 * @param cardinality
	 *            the number of time this child can appear in the enclosing GIML element.
	 * @param isOrdered
	 *            if child order must be preserved
	 * @param attributes
	 *            this child attributes
	 */
	public Child(String name, String prefix, int cardinality, boolean isOrdered, boolean disabled,
			Deque<Attribute> attributes) {
		this._name = name;
		this._prefix = prefix;
		this._cardinality = cardinality;
		this._ordered = isOrdered;
		this._disabled = disabled;
		this._attributes = attributes;
	}

	/**
	 * Gets this child name.
	 *
	 * @return this child name. Can be {@code null} or empty.
	 */
	public String getName() {
		return this._name;
	}

	/**
	 * Gets if this child name is {@code null} or empty.
	 *
	 * @return {@code true} if this child name is {@code null} or empty, {@code false} otherwise.
	 */
	public boolean hasEmptyName() {
		return this._name == null || this._name.trim().isEmpty();
	}

	/**
	 * Gets this child prefix. Prefix comes from this child declaration.
	 *
	 * @return the prefix of this child.
	 */
	public String getPrefix() {
		return this._prefix;
	}

	/**
	 * Gets the number of times this child can appears in the enclosing GIML element.
	 *
	 * @return the cardinality of this child.
	 */
	public int getCardinality() {
		return this._cardinality;
	}

	/**
	 * Gets if a list of this child must be ordered, i.e. if child declaration order must be preserved.
	 *
	 * @return {@code true} if order must be preserved, {@code false} otherwise.
	 */
	public boolean isOrdered() {
		return this._ordered;
	}

	/**
	 * Gets if this child is disabled for this element. Useful to override element's parent behavior.
	 *
	 * @return @return {@code true} if child is disabled, {@code false} otherwise.
	 */
	public boolean isDisabled() {
		return this._disabled;
	}

	/**
	 * Gets child attributes.
	 *
	 * @return this child attributes.
	 */
	public Deque<Attribute> getAttributes() {
		return new LinkedList<Attribute>(this._attributes);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof Child)) {
			return false;
		}

		Child other = (Child) obj;
		boolean namesAreEqual = (this._name == null && other._name == null) || this._name.equalsIgnoreCase(other._name);
		return namesAreEqual && this._prefix.equalsIgnoreCase(other._prefix) && this._cardinality == other._cardinality
				&& this._ordered == other._ordered && this._attributes.equals(other._attributes);
	}

	@Override
	public int hashCode() {
		return Objects.hash(this._name, this._prefix, this._cardinality, this._ordered, this._attributes);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append('\'').append(this._name).append('\'').append('(');
		builder.append("prefix=").append(this._prefix);
		builder.append(",cardinality=").append(this._cardinality);
		builder.append(",ordered=").append(this._ordered);
		builder.append(",disabled=").append(this._disabled);
		builder.append(')');
		return builder.toString();
	}
}
