/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.generator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.Normalizer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.fedorahosted.tennera.jgettext.catalog.parse.ParseException;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;

import com.is2t.giml.binary.GIMLWriter;
import com.is2t.giml.binary.GIMLWriterCompacted;
import com.is2t.giml.description.Description;
import com.is2t.giml.description.DescriptionFactory;
import com.is2t.giml.description.Type;
import com.is2t.giml.jsgenerator.nodes.InlinedCode;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Node;
import com.is2t.nls.StringEncoder;
import com.is2t.nls.generator.NLSGenerator;
import com.is2t.nls.generator.NLSGeneratorException;
import com.is2t.nls.generator.NLSLocale;
import com.is2t.nls.generator.PoParser;

/**
 * This class {@code GIMLGenerator} represents a GIML generator used to process
 * a GIML root document.
 * <p>
 * The generator generates a GIML binary and factories source code used to
 * decode the binary and to create the GIML objects of the document.
 *
 */
public class GIMLGenerator implements GIMLGeneratorConstants {

	private static final String NLS_FUNCTION_PREFIX = "_(";
	private static final String NLS_FUNCTION_SUFFIX = ")";
	private static final String GIMLNLS_HEADER = "gimlnls";
	private static final String COM_IS2T_NLS_PACKAGE = "com/is2t/nls";

	static final STGroup GimlSTGroup;
	static final Logger GimlLogger;

	private GIMLWriter writer;
	private final File gimlGeneratedSourcePackageDir;
	private final Node gimlRoot;
	private final List<GIMLFactory> factories;
	private final File gimlGeneratedSourceDir;
	private final File gimlGeneratedPropertiesPackageDir;

	private Set<NLSLocale> nlsLocales;
	private final Map<String, Integer> nlsMessagesKeyIdPair;
	private int nlsKeyIdCount = 0;
	private final PoParser gimlPoParser;
	private final Map<String, Integer> rootNodeIds;
	private long gimlBinaryId;
	private static int ROOT_IDENTIFIER = 0;
	private static final int NO_ROOT_IDENTIFIER = -1;

	/**
	 * Javascript code entry points, referenced by giml files
	 */
	private final Map<Integer, InlinedCode> codeIds;
	private int nextCodeId = 0;
	private String javaScriptClassFullyQualifiedName;

	static {
		/* initialize the ST Group file used for code generation */
		GimlSTGroup = new STGroupFile(GIML_STRING_TEMPLATE_FILE_NAME);
		GimlLogger = Logger.getLogger("GIML");
		GimlLogger.setLevel(Level.ALL);
	}

	/**
	 * Creates a new GIML generator for the given GIML root node.
	 *
	 * @param gimlRoot
	 *            the GIML root node.
	 * @param outputDir
	 *            the generator output directory.
	 * @throws IllegalArgumentException
	 *             if the given root or the output directory is
	 *             <code>null</code>.
	 * @throws FileNotFoundException
	 *             if the GIML binary file can not be created.
	 */
	public GIMLGenerator(Node gimlRoot, File outputDir) throws IllegalArgumentException, FileNotFoundException {
		this(gimlRoot, outputDir, null);
	}

	public GIMLGenerator(Node gimlRoot, File outputDir, File poFilesDir)
			throws IllegalArgumentException, FileNotFoundException {
		if (gimlRoot == null || outputDir == null) {
			throw new IllegalArgumentException("null argument");
		}
		this.gimlRoot = gimlRoot;
		this.gimlGeneratedSourceDir = outputDir;

		// create output folders
		this.gimlGeneratedSourcePackageDir = new File(this.gimlGeneratedSourceDir, GIML_GENERATED_SRC_PACKAGE_DIR);
		this.gimlGeneratedPropertiesPackageDir = new File(this.gimlGeneratedSourceDir, "properties");
		createOutputGeneratedDir(this.gimlGeneratedSourcePackageDir);
		createOutputGeneratedDir(this.gimlGeneratedPropertiesPackageDir);

		// GIML factory pool
		this.factories = new ArrayList<>();
		this.gimlPoParser = new PoParser(poFilesDir, "");
		this.nlsLocales = new HashSet<NLSLocale>();
		this.nlsMessagesKeyIdPair = new HashMap<String, Integer>();
		this.rootNodeIds = new HashMap<String, Integer>();
		this.codeIds = new HashMap<>();
	}

	/**
	 * Generates a GIML binary and the GIML factories source code.
	 * <p>
	 * The GIML binary and the factories source code are generated in the output
	 * directory
	 *
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws GIMLGeneratorException
	 *             if a GIML exception occurs.
	 */
	public void generate() throws IOException, GIMLGeneratorException {
		// create the GIML writer
		FileOutputStream writerOutputStream = new FileOutputStream(
				new File(this.gimlGeneratedSourcePackageDir, GIML_BINARY_FILE_NAME));
		this.writer = new GIMLWriterCompacted(writerOutputStream);

		generateIntern();

	}

	/**
	 * Generates the GIML binary and GIML factories source code using the given
	 * writer.
	 *
	 * @param writer
	 *            the GIML writer.
	 * @throws IOException
	 *             if an I/O error occurs.
	 * @throws GIMLGeneratorException
	 *             if a GIML exception occurs.
	 * @throws NullPointerException
	 *             if the writer is <code>null</code>.
	 */
	public void generate(GIMLWriter writer) throws IOException, GIMLGeneratorException {
		if (writer == null) {
			throw new NullPointerException("null writer");
		}
		this.writer = writer;
		generateIntern();
	}

	private void generateIntern() throws IOException, GIMLGeneratorException {

		// on first, process *.po files to manage all translations
		processPOFiles();

		// collect all root nodes ids
		processRootNodeIds();

		// set the unique ID with the current time
		this.gimlBinaryId = System.currentTimeMillis();

		// process giml nodes
		List<Node> children = this.gimlRoot.getChildren();
		for (Node node : children) {

			// process GIML child node
			processNode(node);

		}
		// generate the GIML NLS immutable file
		generateGIMLNLS();
		// generate all factories
		generateFactories();
		// generate root id interface
		generateRootIdsInterface();
		// generate giml.system.properties file
		generateGIMLSystemPropertiesFile();
		// generate giml.types.list file
		generateGIMLTypesListFile();
		// generate giml.resources.list file
		gerenateGIMLResourcesListFile();
		// complete the writer
		this.writer.complete(this.gimlBinaryId);
	}

	// collect all root nodes ids and generate an identifier for each one
	private void processRootNodeIds() throws GIMLGeneratorException {
		List<Node> children = this.gimlRoot.getChildren();
		for (Node node : children) {
			String id = node.getAttributeValue(ID_ATTRIBUTE);
			if (id == null) {
				throw new GIMLGeneratorException(
						String.format("Missing required attribute 'id' on the root node '%s'", (node.getName())));
			}
			addRootIdentifier(id);
		}
	}

	private void processPOFiles() throws FileNotFoundException, ParseException, IOException {

		// parse po files
		this.gimlPoParser.parse();

		this.nlsLocales = this.gimlPoParser.getLocales();
		if (this.nlsLocales.size() == 0) {
			// no locales found; add a default locale
			this.nlsLocales.add(new NLSLocale("default", "default"));
		}

		// init nls messages key/id pair
		Set<String> messagesKeys = this.gimlPoParser.getMessagesKeys();
		Iterator<String> iterator = messagesKeys.iterator();
		this.nlsKeyIdCount = 0;
		while (iterator.hasNext()) {
			String messageKey = iterator.next();
			this.nlsMessagesKeyIdPair.put(messageKey, new Integer(++this.nlsKeyIdCount));
		}
	}

	private void processNode(Node node) throws GIMLGeneratorException {
		GIMLFactory factory = new GIMLFactory(node);
		boolean isRoot = factory.description.isRoot();
		String id = node.getAttributeValue(ID_ATTRIBUTE);
		if (isRoot && id == null) {
			throw new GIMLGeneratorException(
					String.format("Missing required attribute 'id' on the root node '%s'", (node.getName())));
		}

		if (isRoot) {
			// mark the root id position
			// id attribute value is case-insensitive
			int identifier = getRootIdentifier(id);
			if(identifier == NO_ROOT_IDENTIFIER){
				//never happen
				throw new RuntimeException("no identifier generated for the root id "+id);
			}
			this.writer.markPosition(identifier);
		}

		// write the factory id
		int factoryId = addFactory(factory);
		this.writer.writeVarUInt(factoryId);

		// on first, process attributes
		processAttributes(node, factory.description, false, isRoot);

		// process node elements according to the factory type
		switch (factory.factotyType) {
		case NoElement:
			// leaf
			return;
		case OrderedElements:
			processOrderedElements(node);
			break;
		case NotOrderedElements:
			processNotOrderedElements(node);
			break;
		case OrderedArrayElements:
		case NotOrderedArrayElements:
			processArrayElements(node);
			break;
		default:
			throw new RuntimeException();
		}
	}

	private boolean addRootIdentifier(String id) {
		if (id == null) {
			throw new NullPointerException("null root id");
		}

		Iterator<String> iterator = this.rootNodeIds.keySet().iterator();
		while (iterator.hasNext()) {
			String rootId = iterator.next();
			// case insensitive comparison
			if (rootId.equalsIgnoreCase(id)) {
				// id already exists
				return false;
			}
		}
		this.rootNodeIds.put(id, Integer.valueOf(ROOT_IDENTIFIER++));
		return true;
	}

	/**
	 * Get the root (integer) identifier from the String identifier
	 *
	 * @param id
	 *            the string identifier
	 * @return the integer identifier, of {@link #NO_ROOT_IDENTIFIER} if the
	 *         identifier is unknown
	 */
	public int getRootIdentifier(String id) {
		Iterator<String> iterator = this.rootNodeIds.keySet().iterator();
		while (iterator.hasNext()) {
			String rootId = iterator.next();
			// case insensitive comparison
			if (rootId.equalsIgnoreCase(id)) {
				return this.rootNodeIds.get(rootId).intValue();
			}
		}
		return NO_ROOT_IDENTIFIER;
	}

	/**
	 * Get all root identifiers
	 *
	 * @return the map between the GIML identifier name and the internal unique id
	 */
	public Map<String, Integer> getAllRootIndentifiers() {
		return rootNodeIds;
	}

	private int addFactory(GIMLFactory factory) {
		// check if the factory already exists
		int factoryId = this.factories.indexOf(factory);
		if (factoryId == -1) {
			// not exist, add to the factories pool
			factoryId = this.factories.size();
			if (!this.factories.add(factory)) {
				throw new RuntimeException("Can not add factory");
			}
		}
		return factoryId;
	}

	private void processOrderedElements(Node node) throws GIMLGeneratorException {

		List<Node> elements = node.getChildren();
		Description description = DescriptionFactory.getDescription(node.getName());
		processElements(elements, description);
	}

	private void processNotOrderedElements(Node node) throws GIMLGeneratorException {
		// sort element by name
		List<Node> sortedElements = GIMLGeneratorUtils.sortNodeChildrenByName(node);
		Description description = DescriptionFactory.getDescription(node.getName());
		processElements(sortedElements, description);

	}

	private void processArrayElements(Node node) throws GIMLGeneratorException {
		Description description = DescriptionFactory.getDescription(node.getName());
		String[] sortedDescriptionElementsArray = GIMLGeneratorUtils.sortIngnoreCase(description.getElements());

		for (String descElement : sortedDescriptionElementsArray) {
			List<Node> elementsWithSameName = node.getChildrenByName(descElement);
			int size = elementsWithSameName.size();
			// serialize element count
			this.writer.writeVarUInt(size);
			processElements(elementsWithSameName, description);
		}

	}

	private void processElements(List<Node> elements, Description description) throws GIMLGeneratorException {
		for (Node element : elements) {
			processElement(element, description);
		}
	}

	private void processElement(Node element, Description description) throws GIMLGeneratorException {
		// only one child by element
		Node child = element.getChildren().get(0);
		processNode(child);
		// serialize the element attributes
		processAttributes(element, description, true, false);
	}

	private void checkRequiredAttributes(Node node, Description description, boolean isElementNode)
			throws GIMLGeneratorException {
		String nodeName = node.getName();
		Collection<String> attributes;
		if (!isElementNode) {
			attributes = description.getAttributes();
		} else {
			attributes = description.getElementAttributes(nodeName);
		}
		for (String attributeName : attributes) {
			Attribute attribute = node.getAttribute(attributeName);
			if (attribute == null) {
				// check if the attribute is required
				// all attributes of an element node are already required
				if (isElementNode || (description.isRequiredAttribute(attributeName)
						&& getAssociatedCodeAttribute(description, attributeName, node) == null)) {
					throw new GIMLGeneratorException(String.format("%s: missing required attribute '%s' on '%s' node",
							description.getTag(), attributeName, nodeName));
				}
			}
			else {
				if (getAssociatedCodeAttribute(description, attributeName, node) != null) {
					throw new GIMLGeneratorException(
							String.format("%s: duplicate attribute and code attribute declaration '%s' on '%s' node",
									description.getTag(), attributeName, nodeName));
				}
			}
		}
	}

	/**
	 * Gets the associated code attribute of the given attribute if it exists in the
	 * given node.
	 *
	 * @return null if the description does not describe an associated code
	 *         attribute or if the associated code attribute is not declared in the
	 *         given node
	 */
	private Attribute getAssociatedCodeAttribute(Description description, String attributeName, Node node) {
		String codeAttributeName = attributeName + "code"; // lower case normalization
		if (!description.getAttributes().contains(codeAttributeName)) {
			return null;
		}
		return node.getAttribute(codeAttributeName);
	}

	private void processAttributes(Node node, Description description, boolean isElementNode, boolean isRoot)
			throws GIMLGeneratorException {

		// check required attributes
		checkRequiredAttributes(node, description, isElementNode);

		String nodeName = node.getName();
		Attribute[] attributesArray = GIMLGeneratorUtils.sortNodeAttributesByName(node);
		Attribute idAttribute = null;
		int rootIdentifier = 0;
		// serialize attributes values
		for (Attribute attribute : attributesArray) {
			if (!isElementNode && attribute.getName().equalsIgnoreCase(ID_ATTRIBUTE)) {

				// check if the value of the id attribute is not a NLS String
				checkNonNLSValue(nodeName, attribute);
				// save the id attribute
				idAttribute = attribute;
				if (isRoot) {
					// check if the type of the id attribute of the node is a
					// String
					checkStringType(nodeName, attribute, description, isElementNode);
					// write the root identifier value (will be read for the code
					// generation of the description initialization)
					rootIdentifier = getRootIdentifier(attribute.getValue());
					if(rootIdentifier == NO_ROOT_IDENTIFIER){
						//never happen
						throw new RuntimeException("no identifier generated for the root id "+attribute.getValue());
					}
					this.writer.writeNonNLSString(String.valueOf(rootIdentifier));
				}
			} else {
				writeAttributeValue(nodeName, attribute, description, isElementNode);
			}
		}

		if (idAttribute != null) {
			// write the id attribute or the identifier (if the node is root)
			// this value will be read to register the GIML object
			if(isRoot){
				this.writer.writeNonNLSString(String.valueOf(rootIdentifier));
			}else{
				this.writer.writeNonNLSString(idAttribute.getValue());
			}
		}
	}

	private void checkStringType(String nodeName, Attribute attribute, Description description, boolean isElementNode)
			throws GIMLGeneratorException {

		Type attrType = getAttributeType(nodeName, attribute.getName(), description, isElementNode);
		// check String type
		if (attrType.getBasetype() != com.is2t.giml.description.Type.Basetype.String) {
			throw new GIMLGeneratorException(String.format(
					"Illegal type for the attribute '%s' in '%s' description. Expected: '%s', found: '%s'",
					attribute.getName(), description.getTag(), com.is2t.giml.description.Type.Basetype.String.getName(),
					attrType.getBasetype().getName()));
		}
	}

	private void checkNonNLSValue(String nodeName, Attribute attribute) throws GIMLGeneratorException {
		// check non NLS value
		String attributeValue = attribute.getValue();
		if (getNLSString(attributeValue) != null) {
			throw new GIMLGeneratorException(
					String.format("The attribute '%s' of '%s' node does not support internationalization",
							attribute.getName(), nodeName));
		}
	}

	public static Type getAttributeType(String nodeName, String attributeName, Description nodeDescription,
			boolean isElementNode) throws GIMLGeneratorException {
		Type attrType;

		try {
			if (isElementNode) {
				attrType = nodeDescription.getElementAttributeType(nodeName, attributeName);
			} else {
				attrType = nodeDescription.getAttributeType(attributeName);
			}
		} catch (IllegalArgumentException e) {
			throw new GIMLGeneratorException(String.format("No attribute named '%s' in '%s' description", attributeName,
					nodeDescription.getTag()));
		}

		if (attrType == null) {
			throw new GIMLGeneratorException(
					String.format("No type provided for the attribute '%s' in '%s' description", attributeName,
							nodeDescription.getTag()));
		}

		switch (attrType.getBasetype()) {
		case Boolean:
		case Byte:
		case Short:
		case Integer:
		case Long:
		case Float:
		case Double:
		case String:
		case RefId:
		case Code:
			break;
		default:
			throw new GIMLGeneratorException(String.format(
					"Unsupported attribute type '%s' in '%s' description. Only base types and Reference type are supported",
					attrType.getBasetype().getName(), nodeDescription.getTag()));
		}

		return attrType;
	}

	/*
	 * write attribute into GIML binary value according to the attribute type
	 */
	private void writeAttributeValue(String nodeName, Attribute attribute, Description nodeDescription,
			boolean isElementNode) throws GIMLGeneratorException {

		String attributeName = attribute.getName();
		String attributeValue = attribute.getValue();
		Type attrType = getAttributeType(nodeName, attributeName, nodeDescription, isElementNode);

		switch (attrType.getBasetype()) {
		case Boolean:
			this.writer.writeBoolean(Boolean.valueOf(attributeValue).booleanValue());
			break;
		case Byte:
			this.writer.writeVarSInt(Byte.decode(attributeValue).byteValue());
			break;
		case Short:
			this.writer.writeVarSInt(Short.decode(attributeValue).shortValue());
			break;
		case Integer:
			this.writer.writeVarSInt(Integer.decode(attributeValue).intValue());
			break;
		case Long:
			this.writer.writeVarLong(Long.decode(attributeValue).longValue());
			break;
		case Float:
			this.writer.writeFloat(Float.parseFloat(attributeValue));
			break;
		case Double:
			this.writer.writeDouble(Double.parseDouble(attributeValue));
			break;
		case String:
			Integer nlsKeyId = processNLS(attributeValue);
			if (nlsKeyId != null) {
				// write the NLS key id value
				this.writer.writeNLSStringId(nlsKeyId);

			} else {
				// non NLS string
				this.writer.writeNonNLSString(attributeValue);
			}
			break;
		case RefId:
			// reference type not support NLS string
			checkNonNLSValue(nodeName, attribute);

			if (addRootIdentifier(attributeValue)) {
				// the reference id is added, it means that the id was not found
				// warning
				GimlLogger.warning(String.format("Reference %s not found", attributeValue));
			}
			int identifier = getRootIdentifier(attributeValue);
			if(identifier == NO_ROOT_IDENTIFIER){
				//never happen
				throw new RuntimeException("no identifier generated for the root id "+attribute.getValue());
			}
			// write the reference identifier value
			this.writer.writeNonNLSString(String.valueOf(identifier));
			break;
		case Code: {
			int codeId = registerCode(attributeValue);
			this.writer.writeVarSInt(codeId);
			break;
		}
		default:
			throw new GIMLGeneratorException(String.format(
					"Unsupported attribute type '%s' in '%s' description. Only base types and Reference type are supported",
					attrType.getBasetype().getName(), nodeDescription.getTag()));
		}
	}

	/**
	 * Process a NLS string (macroexpansion of <code>_(xxx)"</code> attribute).
	 *
	 * @return the NLS key or <code>null</code> if the String has not been expanded
	 *         into a NLS string
	 */
	private Integer processNLS(String str) {
		String originalNLSString = getNLSString(str);
		if (originalNLSString != null) {
			// NLS string
			String nlsMsgKey = gimlStringToKey(originalNLSString);
			processMissingTranslations(nlsMsgKey, originalNLSString);
			Integer nlsKeyId = this.nlsMessagesKeyIdPair.get(nlsMsgKey);
			assert nlsKeyId != null;
			return nlsKeyId;
		}
		else {
			return null;
		}
	}

	private void processMissingTranslations(String nlsMsgKey, String originalNLSString) {
		Integer nlsKeyId = this.nlsMessagesKeyIdPair.get(nlsMsgKey);
		if (nlsKeyId == null) {
			// add missing nls key/id pair
			nlsKeyId = new Integer(++this.nlsKeyIdCount);
			this.nlsMessagesKeyIdPair.put(nlsMsgKey, nlsKeyId);
		}

		for (NLSLocale locale : this.nlsLocales) {
			if (locale.getTranslation(nlsMsgKey) == null) {
				// no translation found for the msg key
				// add the original string as default translation for this key
				locale.addTranslation(nlsMsgKey, originalNLSString);
			}
		}
	}

	private String getNLSString(String attributeValue) {
		if (attributeValue.startsWith(NLS_FUNCTION_PREFIX) && attributeValue.endsWith(NLS_FUNCTION_SUFFIX)) {
			return attributeValue.substring(2, attributeValue.length() - 1);
		} else if (attributeValue.startsWith("gettext(") && attributeValue.endsWith(")")) {
			return attributeValue.substring(8, attributeValue.length() - 1);
		}
		return null;
	}

	/*
	 * generate GIML NLS
	 */
	private void generateGIMLNLS() throws IOException, GIMLGeneratorException {
		// generate only if NLS messages exist
		if (existingNLSMessages()) {
			try {
				NLSGenerator gimlnlsGenerator = new NLSGenerator(this.nlsLocales, this.nlsMessagesKeyIdPair,
						this.gimlGeneratedSourceDir, new StringEncoder(), GIMLNLS_HEADER, COM_IS2T_NLS_PACKAGE);
				gimlnlsGenerator.generate();
			} catch (NLSGeneratorException e) {
				throw new GIMLGeneratorException(e.getMessage());
			}
		}
	}

	/*
	 * generate GIML factories
	 */
	private void generateFactories() throws IOException, GIMLGeneratorException {
		// template for factory provider generation
		ST factoryProvider = GimlSTGroup.getInstanceOf("GIMLFactoryProvider");
		factoryProvider.add("package", GIML_GENERATED_SRC_PACKAGE);
		factoryProvider.add("class", GIML_FACTORY_PROVIDER_IMPL_NAME);
		factoryProvider.add("gimlBinaryId", this.gimlBinaryId);

		int factoriesCount = this.factories.size();
		for (int factoryId = -1; ++factoryId < factoriesCount;) {
			GIMLFactory factory = this.factories.get(factoryId);
			ST factoryTemplate = factory.generate();
			// add factory infos
			factoryProvider.addAggr("factories.{id, methodName, method}", factoryId, factory.uniqueID, factoryTemplate);
		}

		for (int codeId : codeIds.keySet()) {
			InlinedCode code = codeIds.get(codeId);
			String codeMethodName = javaScriptClassFullyQualifiedName + '.' + code.getFunctionName();
			factoryProvider.addAggr("codeAttributes.{id, methodName}", codeId, codeMethodName);
		}

		// generate the factory provider
		File factoryProviderFile = new File(this.gimlGeneratedSourcePackageDir,
				GIML_FACTORY_PROVIDER_IMPL_NAME + ".java");
		writeGeneratedGIMLCode(factoryProvider, factoryProviderFile);
	}



	private void generateRootIdsInterface() throws IOException {
		// generate Root ids interface
		ST rootIdInterface = GimlSTGroup.getInstanceOf("GIMLRootIdsInterface");
		rootIdInterface.add("interfaceName", ROOT_IDS_INTERFACE_NAME);
		rootIdInterface.add("package", GIML_GENERATED_SRC_PACKAGE);
		Iterator<String> iterator = this.rootNodeIds.keySet().iterator();
		Map<String, String> idKeysValues = new HashMap<String, String>();
		while (iterator.hasNext()) {
			String idValue = iterator.next();
			String idKey = gimlStringToKey(idValue);
			Integer identifier = rootNodeIds.get(idValue);
			idKeysValues.put(idKey, Integer.toString(identifier.intValue()));
		}
		rootIdInterface.add("idKeysValues", idKeysValues);
		File rootIdsInterfaceFile = new File(this.gimlGeneratedSourcePackageDir, ROOT_IDS_INTERFACE_NAME + ".java");
		GIMLGenerator.writeGeneratedGIMLCode(rootIdInterface, rootIdsInterfaceFile);
	}

	/*
	 * generate giml.system.properties file
	 */
	private void generateGIMLSystemPropertiesFile() throws IOException {
		Map<String, String> properties = new HashMap<String, String>();
		// add GIML factory implementation property
		properties.put(GIML_FACTORY_PROVIDER_INTERFACE,
				GIML_GENERATED_SRC_PACKAGE + "." + GIML_FACTORY_PROVIDER_IMPL_NAME);
		// add GIML binary file property
		properties.put(GIML_BINARY_PROPERTY_NAME, "/" + GIML_GENERATED_SRC_PACKAGE_DIR + GIML_BINARY_FILE_NAME);
		properties.put(GIML_RUNTIME_VERSION_PROPERTY_NAME, GIML_RUNTIME_VERSION);
		ST GIMLPropertyTemplate = GimlSTGroup.getInstanceOf("GIMLProperties");
		GIMLPropertyTemplate.add("properties", properties);
		File GIMLPropertyFile = new File(this.gimlGeneratedPropertiesPackageDir, GIML_SYSTEM_PROPERTIES_FILE_NAME);
		writeGeneratedGIMLCode(GIMLPropertyTemplate, GIMLPropertyFile);
	}

	/*
	 * generate giml.types.list
	 */
	private void generateGIMLTypesListFile() throws IOException {
		ST typesListTemplate = GimlSTGroup.getInstanceOf("GIMLList");
		typesListTemplate.add("entries", GIML_GENERATED_SRC_PACKAGE + "." + GIML_FACTORY_PROVIDER_IMPL_NAME);
		File GIMLTypesListFile = new File(this.gimlGeneratedPropertiesPackageDir, GIML_TYPES_LIST_FILE_NAME);
		writeGeneratedGIMLCode(typesListTemplate, GIMLTypesListFile);
	}

	/*
	 * generate giml.resources.list
	 */
	private void gerenateGIMLResourcesListFile() throws IOException {
		ST resourcesListTemplate = GimlSTGroup.getInstanceOf("GIMLList");
		resourcesListTemplate.add("entries", GIML_GENERATED_SRC_PACKAGE_DIR + GIML_BINARY_FILE_NAME);
		File GIMLResourcesListFile = new File(this.gimlGeneratedPropertiesPackageDir, GIML_RESOURCES_LIST_FILE_NAME);
		writeGeneratedGIMLCode(resourcesListTemplate, GIMLResourcesListFile);

	}

	private boolean existingNLSMessages() {
		return this.nlsMessagesKeyIdPair.size() > 0;
	}

	public static void writeGeneratedGIMLCode(ST factoryTemplate, File file) throws IOException {
		BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF-8"));
		out.write(factoryTemplate.render());
		out.flush();
		out.close();
	}

	static void createOutputGeneratedDir(File output) {
		if (output.isDirectory()) {
			return;
		}
		if (!output.mkdirs()) {
			throw new RuntimeException(
					String.format("Can not create the output directory: %s", output.getAbsolutePath()));
		}
	}

	/*
	 * Computes a key from the given string value. All characters that is not a
	 * letter or a digit are replaced by an underscore.
	 */
	static String gimlStringToKey(String string) {
		string = Normalizer.normalize(string, Normalizer.Form.NFD);
		string = string.replaceAll("\\p{IncombiningDiacriticalMarks}+", "");
		int len = string.length();
		StringBuffer key = new StringBuffer();
		char firstChar = string.charAt(0);
		if (Character.isLetter(firstChar) || firstChar == '$' || firstChar == '_') {
			key.append(firstChar);
		} else {
			key.append('_');
			if (Character.isDigit(firstChar)) {
				key.append(firstChar);
			}
		}
		for (int i = 0; ++i < len;) {
			char c = string.charAt(i);
			if (Character.isLetterOrDigit(c) || firstChar == '$' || firstChar == '_') {
				key.append(c);
			} else {
				key.append('_');
			}
		}
		return key.toString();
	}

	/**
	 * Register a new piece of code entry point.
	 *
	 * @param code
	 *            the piece of code
	 * @return a unique id that reference this piece of code
	 */
	private int registerCode(String code) {
		code = processNLSOnJavaScriptCode(code);

		int codeId = nextCodeId;
		String functionName = "GIMLCodeAttributeInlinedFunction" + toStringCodeId(codeId);
		codeIds.put(codeId, new InlinedCode(code, functionName));
		++nextCodeId;
		return codeId;
	}

	/**
	 * Replace NLS expansion into JavaScript code
	 *
	 * @param code
	 *            the code with NLS expansion
	 * @return the code with NLS expansion resolved to a global function
	 *         <code>nls(int)</code>
	 */
	private String processNLSOnJavaScriptCode(String code) {
		StringBuilder sb = new StringBuilder();
		int startIndex = 0;
		while (true) {
			int nlsStartIndex = code.indexOf(NLS_FUNCTION_PREFIX, startIndex);
			if (nlsStartIndex == -1) {
				sb.append(code, startIndex, code.length());
				break;
			}
			sb.append(code, startIndex, nlsStartIndex);
			int nlsEndIndex = code.indexOf(NLS_FUNCTION_SUFFIX, nlsStartIndex);
			Integer key;
			if (nlsEndIndex != -1 && (key = processNLS(code.substring(nlsStartIndex, nlsEndIndex + 1))) != null) {
				// NLS String: replace by a global function + continue after the suffix
				boolean stringLiteral = isIncludedAsStringLiteral(code, nlsStartIndex, nlsEndIndex);

				// replace a string literal by string expression if any
				if (stringLiteral) {
					// remove end quote
					sb.deleteCharAt(sb.length() - 1);
				}
				// generate the new String expression
				sb.append("nls(" + key + ")");
				if (stringLiteral) {
					++nlsEndIndex; // skip String end quote
				}
				startIndex = nlsEndIndex + NLS_FUNCTION_SUFFIX.length();
			} else {
				// not a NLS string - continue just after the prefix
				sb.append(NLS_FUNCTION_PREFIX);
				startIndex = nlsStartIndex + NLS_FUNCTION_PREFIX.length();
			}
		}
		return sb.toString();
	}

	private boolean isIncludedAsStringLiteral(String code, int nlsStartIndex, int nlsEndIndex) {
		if (nlsStartIndex == 0) {
			return false;
		}
		if (nlsEndIndex == code.length() - 1) {
			return false;
		}
		char previousChar = code.charAt(nlsStartIndex - 1);
		char nextChar = code.charAt(nlsEndIndex + 1);

		if (((previousChar == '\'' && nextChar == '\'')) || ((previousChar == '"' && nextChar == '"'))) {
			return true;
		}
		return false;
	}

	/**
	 * A string representation of the given id on the same number of characters
	 *
	 * @param codeId
	 * @return a fixed size length
	 */
	private String toStringCodeId(int codeId) {
		StringBuilder sb = new StringBuilder();
		if (codeId < 10) {
			sb.append('0');
		}
		if (codeId < 100) {
			sb.append('0');
		}
		if (codeId < 1000) {
			sb.append('0');
		}
		return sb.append(codeId).toString();
	}

	/**
	 * @param name
	 *            fully qualified name of the JavaScript class name
	 */
	public void setJavaScriptClassFullyQualifiedName(String name) {
		this.javaScriptClassFullyQualifiedName = name;
	}

	/**
	 * @return all inlined JavaScript code writtent in GIML files
	 */
	public InlinedCode[] getInlinedJavaScriptCodes() {
		Collection<InlinedCode> codeVect = codeIds.values();
		InlinedCode[] codes = new InlinedCode[codeVect.size()];
		codeVect.toArray(codes);
		Arrays.sort(codes, new Comparator<InlinedCode>() {
			// alphabetical sort of codeIds functions for better lisibility
			@Override
			public int compare(InlinedCode o1, InlinedCode o2) {
				return o1.getFunctionName().compareTo(o2.getFunctionName());
			};
		});
		return codes;
	}
}