/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.runtime.reader;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import com.is2t.memory.Memory;
import com.is2t.memory.MemoryHelper;
import com.is2t.memory.StringDecoder;

import ej.giml.GIMLReader;
import ej.nls.NLS;

/**
 * Implementations of the {@link GIMLReader} that read GIML binary data generated in a compact format.
 */
public class GIMLReaderJava extends AbstractGIMLReader {

	private static final int TRUE = 1;

	/** Size in bytes of the GIML header */
	private static final int HEADER_SIZE = 12;

	private final Map<Integer, Integer> objectMap;

	private final Memory memory;

	/**
	 * Creates a {@link GIMLReader} that reads the data from the given stream.
	 *
	 * @throws IOException
	 */
	public GIMLReaderJava(String gimlBinaryPath) throws IOException {
		this.memory = MemoryHelper.newMemory(gimlBinaryPath);
		this.objectMap = new HashMap<>();
	}

	@Override
	public void initialize(NLS nls) {
		super.initialize(nls);
		readObjectMap();
	}

	private void readObjectMap() {
		this.memory.seek(HEADER_SIZE);
		int nbBuckets = readUInt16();

		// skip pointers table
		this.memory.skip(nbBuckets * 2);

		for (int bucketIndex = 0; bucketIndex < nbBuckets; bucketIndex++) {
			int nbEntries = this.memory.readUnsignedByte();
			for (int entryIndex = 0; entryIndex < nbEntries; entryIndex++) {
				int objectIdentifier = readUInt24();
				int objectOffset = readUInt24();

				int objectAddr = objectOffset + this.dataPtr;
				this.objectMap.put(Integer.valueOf(objectIdentifier), Integer.valueOf(objectAddr));
			}
		}
	}

	@Override
	public int readVarSInt() {
		int chunck = this.memory.readSignedByte();
		int res = (chunck >> 1);

		while (true) {
			if ((chunck & 1) == 0) {
				return res;
			}
			chunck = this.memory.readUnsignedByte();
			res = (res << 7) | (chunck >> 1);
		}
	}

	@Override
	public int readVarUInt() {
		int chunck = this.memory.readUnsignedByte();
		int res = chunck >> 1;

		while (true) {
			if ((chunck & 1) == 0) {
				return res;
			}
			chunck = this.memory.readUnsignedByte();
			res = (res << 7) | (chunck >> 1);
		}
	}

	@Override
	public long readVarLong() {
		int chunck = this.memory.readSignedByte();
		long res = chunck >> 1;

		while (true) {
			if ((chunck & 1) == 0) {
				return res;
			}
			chunck = this.memory.readUnsignedByte();
			res = (res << 7) | (chunck >> 1);
		}
	}

	@Override
	public boolean readBoolean() {
		int value = this.memory.readUnsignedByte();
		return value == TRUE;
	}

	@Override
	public float readFloat() {
		int bits = (this.memory.readUnsignedByte() << 24) | (this.memory.readUnsignedByte() << 16)
				| (this.memory.readUnsignedByte() << 8) | (this.memory.readUnsignedByte());

		return Float.intBitsToFloat(bits);
	}

	@Override
	public double readDouble() {
		return Double.longBitsToDouble(readLong());
	}

	@Override
	protected long readLong() {
		long bits = (((long) this.memory.readUnsignedByte()) << 56) | (((long) this.memory.readUnsignedByte()) << 48)
				| (((long) this.memory.readUnsignedByte()) << 40) | (((long) this.memory.readUnsignedByte()) << 32)
				| (((long) this.memory.readUnsignedByte()) << 24) | (((long) this.memory.readUnsignedByte()) << 16)
				| (((long) this.memory.readUnsignedByte()) << 8) | (this.memory.readUnsignedByte());

		return bits;
	}

	@Override
	public void seek(int positionIdentifier) throws IllegalArgumentException {
		Integer position = this.objectMap.get(Integer.valueOf(positionIdentifier));
		if (position == null) {
			throw new IllegalArgumentException("Unknown ID: " + positionIdentifier);
		}
		this.memory.seek(position.intValue());
	}

	private int readUInt24() {
		return (this.memory.readUnsignedByte() << 16) | (this.memory.readUnsignedByte() << 8)
				| (this.memory.readUnsignedByte());
	}

	@Override
	protected int readUInt16() {
		return (this.memory.readUnsignedByte() << 8) | (this.memory.readUnsignedByte());
	}

	@Override
	protected String decodeString(int stringAddr) {
		return StringDecoder.decodeString(this.memory, stringAddr);
	}

	@Override
	protected int startAddress() {
		return 0;
	}

}
