/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser.test.fileutils;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.junit.Test;

import com.is2t.giml.xml.parser.test.AbstractTest;
import com.is2t.giml.xml.parser.utils.FileUtils;

import junit.framework.Assert;

/**
 * Unit test for {@link FileUtils} class.
 */
public class FileUtilsTest extends AbstractTest {

	private static final String DEFAULT_EXTENSION = ".ext";

	@Test
	public void testIsValidFileValidFile() throws IOException {
		FileUtils.checkValidFile(createTempFile("temp-file"));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidFileNullArg() throws IOException {
		FileUtils.checkValidFile(null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidFileValidFolder() throws IOException {
		FileUtils.checkValidFile(createTempDir("temp-folder"));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidFileNonExistingFile() throws IOException {
		FileUtils.checkValidFile(new File("/does/not/exists.ext"));
	}

	@Test
	public void testIsValidDirectoryValidDir() throws IOException {
		FileUtils.checkValidDirectory(createTempDir("temp-dir"));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidDirectoryNullArg() throws IOException {
		FileUtils.checkValidDirectory((File) null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidDirectoryValidFile() throws IOException {
		FileUtils.checkValidDirectory(createTempFile("temp-file"));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testIsValidDirectoryNonExistingFile() throws IOException {
		FileUtils.checkValidDirectory(new File("/does/not/exists/"));
	}

	@Test
	public void testIsValidFile() throws IOException {
		String filePrefix = "testIsValidFile";

		File file01 = createTempFile(filePrefix + "File01");
		File file02 = null;
		File dir01 = createTempDir(filePrefix + "Dir01");
		File dir02 = null;

		Assert.assertTrue(FileUtils.isValidFile(file01));
		Assert.assertFalse(FileUtils.isValidFile(file02));
		Assert.assertFalse(FileUtils.isValidFile(dir01));
		Assert.assertFalse(FileUtils.isValidFile(dir02));
	}

	@Test
	public void testIsValidDirectory() throws IOException {
		String filePrefix = "testIsValidDirectory";

		File file01 = createTempFile(filePrefix + "File01");
		File file02 = null;
		File dir01 = createTempDir(filePrefix + "Dir01");
		File dir02 = null;

		Assert.assertFalse(FileUtils.isValidDirectory(file01));
		Assert.assertFalse(FileUtils.isValidDirectory(file02));
		Assert.assertTrue(FileUtils.isValidDirectory(dir01));
		Assert.assertFalse(FileUtils.isValidDirectory(dir02));
	}

	@Test(expected = IllegalArgumentException.class)
	public void testListFileFromNonExistingDirectory() {
		FileUtils.listFiles(new File("/does/not/exist/"), DEFAULT_EXTENSION);
	}

	@Test(expected = NullPointerException.class)
	public void testListFileFromNullFolder() throws IOException {
		FileUtils.listFiles((File) null, DEFAULT_EXTENSION);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testListFileFromExistingFile() throws IOException {
		File tempFile = createTempFile("temp-file");
		FileUtils.listFiles(tempFile, DEFAULT_EXTENSION);
	}

	@Test
	public void testListFilesEmptyFolder() throws IOException {
		File tempDir = createTempDir("temp-dir");
		Assert.assertEquals(FileUtils.listFiles(tempDir, DEFAULT_EXTENSION).size(), 0);
	}

	@Test
	public void testListFilesValidFolder() throws IOException {
		File tempDir = createTempDir("temp-dir");
		File file1 = new File(tempDir, "file1" + DEFAULT_EXTENSION);
		File file2 = new File(tempDir, "file2" + DEFAULT_EXTENSION);
		file1.createNewFile();
		file2.createNewFile();
		File folder = new File(tempDir, "folder");
		folder.mkdir();
		File file3 = new File(folder, "file3.other");
		File file4 = new File(folder, "file4");
		File file5 = new File(folder, "file5" + DEFAULT_EXTENSION);
		file3.createNewFile();
		file4.createNewFile();
		file5.createNewFile();

		List<File> listFiles = FileUtils.listFiles(tempDir, DEFAULT_EXTENSION);
		Assert.assertEquals(listFiles.size(), 3);
		Assert.assertTrue(listFiles.contains(file1));
		Assert.assertTrue(listFiles.contains(file2));
		Assert.assertTrue(listFiles.contains(file5));
	}
}
