/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.AbstractPlainDatagramSocketImpl;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;

import com.is2t.support.DependencyInjectionHelper;
import com.is2t.support.net.util.IntRef;

import ej.lang.ResourceManager;

public final class PlainDatagramSocketImpl extends AbstractPlainDatagramSocketImpl {
	private static IDatagramSocketChannelFactory SocketChannelFactory = (IDatagramSocketChannelFactory) DependencyInjectionHelper
			.getImplementation(IDatagramSocketChannelFactory.class);
	private final IDatagramSocketChannel channel;

	/**
	 * Lock object to serialize threads wanting to receive
	 */
	private final Object receiveLock = new Object();

	/**
	 * Lock object to serialize threads wanting to send
	 */
	private final Object sendLock = new Object();

	private boolean isBound = false;

	private final boolean isMulticast;

	/**
	 * Default do nothing constructor
	 */
	public PlainDatagramSocketImpl(boolean isMulticast) throws IOException {
		this.isMulticast = isMulticast;
		channel = SocketChannelFactory.newDatagramSocketChannel();
	}

	@Override
	protected int getLocalPort() {
		if (isClosed()){
			return -1; //-1 if the socket is closed
		}

		if(isBound){
			return localPort;
		}
		return 0;//0 if the socket is not bound
	}

	@Override
	protected void send(DatagramPacket p) throws IOException {
		synchronized (sendLock) {
			InetAddress remote = p.getAddress();
			int port = p.getPort();
			if (remote == null) {
				throw new NullPointerException();
			}
			if (port <= 0) {
				throw new SocketException("invalid port " + port);
			}
			try {
				channel.send(p.getData(), p.getOffset(),
						p.getLength(), new InetSocketAddress(remote, port));
			} catch (InterruptedIOException ioe) {
				// Ignore; interrupted system call.
			}
		}
	}


	@Override
	protected void bind0(int lport, InetAddress laddr) throws SocketException {
		try {

			channel.bind(new InetSocketAddress(laddr, lport));
			localPort = channel.getLocalAddress().getPort();
			isBound = true;
		} catch (SocketException se) {
			throw se;
		} catch (IOException ioe) {
			SocketException se = new SocketException();
			se.initCause(ioe);
			throw se;
		}

	}

	@Override
	protected void receive0(DatagramPacket p) throws IOException {
		synchronized (receiveLock) {
			if (isClosed()){
				throw new SocketException("Socket is closed");
			}
			SocketAddress addr = null;
			IntRef readLength = new IntRef(-1);
			while (true) {
				try {
					addr = channel.receive(p.getData(),
							p.getOffset(), p.getBufLength(), readLength);
					break;
				} catch (SocketTimeoutException ste) {
					throw ste;
				} catch (InterruptedIOException iioe) {
					// Ignore. Loop.
				}
			}
			if (addr != null){
				p.setSocketAddress(addr);
				p.setReceivedLength(readLength.getValue());
			}
		}

	}

	@Override
	protected void datagramSocketCreate() throws SocketException {
		try {
			channel.initDatagramSocket(isMulticast);
			// register socket impl as new resource in ResourceManager
			ResourceManager rm = ResourceManager.getResourceManager();
			if (rm != null) {
				rm.resourceCreated(this);
			}
		} catch (SocketException se) {
			throw se;
		} catch (IOException ioe) {
			SocketException se = new SocketException();
			se.initCause(ioe);
			throw se;
		}

	}

	@Override
	protected void datagramSocketClose() {
		try {
			if (channel.getState().isValid()){
				channel.close();
			}
			ResourceManager rm = ResourceManager.getResourceManager();
			if (rm != null) {
				rm.resourceReclaimed(this);
			}
		} catch (IOException ioe) {
		}

	}

	@Override
	protected void socketSetOption(int opt, Object val) throws SocketException {
		switch (opt) {
		case IP_TOS:
		case IP_MULTICAST_LOOP:
		case IP_MULTICAST_IF:
		case IP_MULTICAST_IF2:
		case SO_BROADCAST:
		case SO_RCVBUF:
		case SO_SNDBUF:
		case SO_TIMEOUT:
		case SO_REUSEADDR:
			channel.setOption(opt, val);
			return;

		default:
			throw new SocketException("cannot set option " + opt);
		}

	}

	@Override
	protected Object socketGetOption(int opt) throws SocketException {
		if (opt == SO_BINDADDR) {
			try {
				InetSocketAddress local = channel.getLocalAddress();
				if (local == null) {
					return null;
				}
				return local.getAddress();
			} catch (SocketException se) {
				throw se;
			} catch (IOException ioe) {
				SocketException se = new SocketException();
				se.initCause(ioe);
				throw se;
			}
		}
		return channel.getOption(opt);
	}

	@Override
	protected void connect0(InetAddress address, int port)
			throws SocketException {
		try {
			channel.connect(new InetSocketAddress(address, port), 0);
		} catch (SocketTimeoutException e) {
			// never occurred, inifite timeout set
		}

	}

	@Override
	protected void disconnect0(int family) {
		synchronized (this) {
			try {
				if (channel.getState().isValid()) {
					channel.disconnect();
				}
			} catch (IOException ioe) {
			}
		}
	}

	@Override
	protected void setTimeToLive(int ttl) throws IOException {
		channel.setTimeToLive(ttl);
	}

	@Override
	protected int getTimeToLive() throws IOException {
		return channel.getTimeToLive();
	}

	@Override
	protected void join(InetAddress inetaddr, NetworkInterface netIf) throws IOException {
		channel.join(inetaddr, netIf);
	}

	@Override
	protected void leave(InetAddress inetaddr, NetworkInterface netIf) throws IOException {
		channel.leave(inetaddr, netIf);
	}

	/*
	 * returns the socket file descriptor
	 * well known API by DTLS impl
	 */
	public int getNativeFD() {
		try {
			return channel.getState().getNativeFD();
		} catch(IOException e) {
			return -1; // invalid descriptor
		}
	}

}
