/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2016, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.AbstractPlainSocketImpl;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketImpl;
import java.net.SocketTimeoutException;

import com.is2t.support.DependencyInjectionHelper;

import ej.lang.ResourceManager;

public class PlainSocketImpl extends AbstractPlainSocketImpl {

	private static IStreamSocketChannelFactory SocketChannelFactory = (IStreamSocketChannelFactory) DependencyInjectionHelper
			.getImplementation(IStreamSocketChannelFactory.class);

	// FIXME underlying socket channel implementation
	protected IStreamSocketChannel socketChannel;
	private boolean connected;// FIXME added

	private boolean isCreated = false;
	private boolean isServerSocket = false;
	private boolean isBound = false;
	private InetAddress localAddress = null;
	
	/**
	 * Default do nothing constructor.
	 */
	public PlainSocketImpl() {
		try {
			this.socketChannel = SocketChannelFactory.newStreamSocketChannel();
		} catch (IOException e) {
			this.socketChannel = null;
		}
	}

	/**
	 * FIXME added
	 */
	private Object LOCK = new Object();
	private boolean blocking = true;

	private final void configureBlocking(boolean blocking) throws IOException {
		synchronized (LOCK) {
			if (this.blocking != blocking) {
				socketChannel.setBlocking(blocking);
				this.blocking = blocking;
			}
		}

	}

	@Override
	protected InetAddress getInetAddress() {

			if (isServerSocket) {
				// return the local address of the server socket 
				return localAddress;
			}

			// not a server, return the remote address
			return address;

	}

	@Override
	protected int getLocalPort() {
		if(isBound){
			return localport;
		}
		return -1; //-1 if the socket is not bound
	}

	@Override
	protected int getPort() {
		if(connected){
			return port;
		}
		return 0; //0 if the socket is not connected
	}

	@Override
	protected void socketCreate(boolean stream) throws IOException {
		if (isCreated) {
			throw new IOException("socket already created");
		}
		// create a new socket channel
		socketChannel.initSocket(stream);
		// register socket impl as new resource in ResourceManager
		ResourceManager rm = ResourceManager.getResourceManager();
		if (rm != null) {
			rm.resourceCreated(this);
		}
		configureBlocking(true);// FIXME added
		isCreated = true;
	}

	@Override
	protected void serverSocketCreate() throws IOException {
		if (isCreated) {
			throw new IOException("socket already created");
		}
		socketChannel.initServerSocket();
		// register socket impl as new resource in ResourceManager
		ResourceManager rm = ResourceManager.getResourceManager();
		if (rm != null) {
			rm.resourceCreated(this);
		}
		configureBlocking(true);// FIXME added
		isCreated = true;
		isServerSocket = true;
	}

	@Override
	protected void socketConnect(InetAddress address, int port, int timeout)
			throws IOException {

		if (connected) {
			throw new IOException("Socket already connected");
		}

		InetSocketAddress connectAddress = new InetSocketAddress(address, port);

		if (connectAddress.isUnresolved()) {
			throw new IOException("Unresolved address");
		}

		connected = socketChannel.connect(connectAddress, timeout);

		if (!connected) {
			throw new SocketTimeoutException("connect timed out");
		}
		this.address = address;
		this.port = port;
		if(!getSocket().isBound()){
			//The socket was not explicitly bound before the call of connect.
			//The bound is silently done by the underlying native socket implementation.
			//Retrieve the local address on which the socket is bound
			InetSocketAddress local = socketChannel.getLocalAddress();
			localAddress = local.getAddress();
			localport = local.getPort();
			isBound = true;
		}

	}

	@Override
	protected void socketBind(InetAddress address, int port) throws IOException {

		socketChannel.bind(new InetSocketAddress(address, port));
		InetSocketAddress local = socketChannel.getLocalAddress();
		localport = local.getPort();
		localAddress = local.getAddress();;
		isBound = true;
	}

	@Override
	protected void socketListen(int count) throws IOException {
		socketChannel.listen(count);
	}

	@Override
	protected void socketAccept(SocketImpl impl) throws IOException {
		if (!(impl instanceof PlainSocketImpl)) {
			throw new IOException("incompatible SocketImpl: "
					+ impl.getClass().getName());
		}
		PlainSocketImpl that = (PlainSocketImpl) impl;
		IStreamSocketChannel ssch = socketChannel.accept();
		// register socket impl as new resource in ResourceManager
		ResourceManager rm = ResourceManager.getResourceManager();
		if (rm != null) {
			rm.resourceCreated(that);
		}
		that.socketChannel.setNativeFD(ssch.getState().getNativeFD());
		that.setOption(SO_REUSEADDR, Boolean.TRUE);
		// Reset the inherited timeout.
		that.setOption(SO_TIMEOUT, Integer.valueOf(0));
		InetSocketAddress thatPeer = that.socketChannel.getPeerAddress();
		InetSocketAddress thatLocal = that.socketChannel.getLocalAddress();
		that.address = thatPeer.getAddress();
		that.port = thatPeer.getPort();
		that.localAddress = thatLocal.getAddress();
		that.localport = thatLocal.getPort();
		that.isCreated = true;
		that.connected = true;
		that.isBound = true;

	}

	@Override
	protected int socketAvailable() throws IOException {
		return socketChannel.available();
	}

	@Override
	protected void socketClose0(boolean useDeferredClose) throws IOException {
		if (socketChannel.getState().isValid()) {
			socketChannel.close();
			ResourceManager rm = ResourceManager.getResourceManager();
			if (rm != null) {
				rm.resourceReclaimed(this);
			}
		}
	}

	@Override
	protected void socketSetOption(int cmd, boolean on, Object value)
			throws SocketException {
		switch (cmd) {
		case IP_TOS:
		case IP_MULTICAST_LOOP:
		case SO_BROADCAST:
		case SO_KEEPALIVE:
		case TCP_NODELAY:
		case SO_RCVBUF:
		case SO_SNDBUF:
		case SO_TIMEOUT:
		case SO_REUSEADDR:
		case SO_LINGER:
		case SO_OOBINLINE:
			socketChannel.setOption(cmd, value);
			return;
		default:
			throw new SocketException("Unrecognized TCP option: " + cmd);
		}

	}

	@Override
	protected Object socketGetOption(int opt) throws SocketException {
		if (opt == SO_BINDADDR) {
			try {
				return socketChannel.getLocalAddress().getAddress();
			} catch (IOException ioe) {
				SocketException se = new SocketException();
				se.initCause(ioe);
				throw se;
			}
		}

		// This filters options which are invalid for TCP.
		switch (opt) {
		case IP_TOS:
		case IP_MULTICAST_LOOP:
		case SO_BROADCAST:
		case SO_KEEPALIVE:
		case TCP_NODELAY:
		case SO_RCVBUF:
		case SO_SNDBUF:
		case SO_TIMEOUT:
		case SO_REUSEADDR:
		case SO_LINGER:
		case SO_OOBINLINE:
			return socketChannel.getOption(opt);
		default:
			throw new SocketException("Unrecognized TCP option: " + opt);
		}
	}

	@Override
	protected int socketRead(byte[] b, int off, int len) throws IOException {
		if(isClosedOrPending()){
			throw new SocketException("Socket closed");
		}
		return socketChannel.read(b, off, len);
	}

	@Override
	protected void socketWrite(byte[] b, int off, int len) throws IOException {
		if(isClosedOrPending()){
			throw new SocketException("Socket closed");
		}
		socketChannel.write(b, off, len);

	}

	/*
	 * returns the socket file descriptor 
	 * well known API by SSL impl
	 */
	public int getNativeFD() {
		try {
			return socketChannel.getState().getNativeFD();
		} catch(IOException e) {
			return -1; // invalid descriptor
		}
	}

}
