/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.ConnectException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NoRouteToHostException;
import java.net.SocketException;
import java.net.SocketTimeoutException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.SocketChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;

import ej.bon.Util;
import ej.error.Message;


public class SocketChannel extends Channel implements ISocketChannel {



	public SocketChannel() throws IOException {
		super();
	}

	public SocketChannel(int fd) throws IOException {
		super(fd);
	}

	// Network (socket) specific methods.

	/**
	 * Connect the underlying socket file descriptor to the remote host.
	 *
	 * @param saddr
	 *            The address to connect to.
	 * @param timeout
	 *            The connect timeout to use for blocking connects.
	 * @return True if the connection succeeded; false if the file descriptor is
	 *         in non-blocking mode and the connection did not immediately
	 *         succeed.
	 * @throws IOException
	 *             If an error occurs while connecting.
	 */
	@Override
	public boolean connect(InetSocketAddress saddr, int timeout)
			throws SocketException, SocketTimeoutException {
		int fd;

		InetAddress addr = saddr.getAddress();

		// Translates an IOException into a SocketException to conform
		// to the throws clause.
		try {
			fd = this.nfd.getNativeFD();
		} catch (IOException ioe) {
			throw new SocketException(ioe.getMessage());
		}

		byte[] addrBuffer = addr.getAddress();
		byte[] hostAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		System.arraycopy(addrBuffer, 0, hostAddr, 0, addrBuffer.length);
		int len = addrBuffer.length;

		int ret = 0;

		if (addr instanceof Inet4Address || addr instanceof Inet6Address) {
			long absoluteTimeout = 0;
			if(timeout > 0){
				absoluteTimeout = Util.platformTimeMillis() + timeout;
			}
			ret = SocketChannelNatives.connect(fd, hostAddr, len,
					saddr.getPort(), timeout, false);
			while (ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SocketChannelNatives.connect(fd, hostAddr, len,
						saddr.getPort(), timeout, true);
			}

			// Check if timeout reached
			if(ret < 0 && absoluteTimeout != 0 && Util.platformTimeMillis() >= absoluteTimeout){
				ret = NetErrors.J_ETIMEDOUT;
			}

		} else {
			throw new SocketException(NetErrors.UNSUPPORTED_INTERNET_ADDRESS);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_ETIMEDOUT:
				throw new SocketTimeoutException(Message.at(NetErrors.getInstance(), ret));
			case NetErrors.J_ECONNREFUSED:
				throw new ConnectException(Message.at(NetErrors.getInstance(), ret));
			case NetErrors.J_EHOSTUNREACH:
				throw new NoRouteToHostException(Message.at(NetErrors.getInstance(), ret));
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));

			}
		}
		return ret == 1;
	}

	@Override
	public InetSocketAddress getLocalAddress() throws IOException {
		if (!this.nfd.isValid()) {
			return null;
		}

		int namelen;
		byte[] addr = null;
		int native_fd = this.nfd.getNativeFD();
		byte[] localAddrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];

		namelen = SocketChannelNatives.getLocalAddress(native_fd, localAddrBuffer, localAddrBuffer.length, false);

		//retry native returned with result
		while(namelen == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			namelen = SocketChannelNatives.getLocalAddress(native_fd, localAddrBuffer, localAddrBuffer.length,
					true);
		}

		if (namelen > 0) {
			addr = new byte[namelen];
			System.arraycopy(localAddrBuffer, 0, addr, 0, namelen);
		}


		int port = SocketChannelNatives.getLocalPort(native_fd, false);

		//retry native returned with result
		while(port == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			port = SocketChannelNatives.getLocalPort(native_fd, true);
		}

		if (port < 0) {
			switch (port) {
			case -1: // not bound
				throw new SocketException(NetErrors.SOCKET_NOT_BOUND);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), port));
			}
		}

		switch (namelen) {
		case 0: // not bound
			throw new SocketException(NetErrors.SOCKET_NOT_BOUND);
		case 4:
			return new InetSocketAddress(Inet4Address.getByAddress(addr), port);
		case 16:
			return new InetSocketAddress(Inet6Address.getByAddress(addr), port);
		default: {
			if(namelen < 0){
				throw new SocketException(Message.at(NetErrors.getInstance(), namelen));
			}
			throw new SocketException(NetErrors.INVALID_ADDRESS_LENGTH);
		}
		}
	}

	/**
	 * Returns the socket address of the remote peer this channel is connected
	 * to
	 *
	 * @return The peer address.
	 * @throws IOException
	 */
	@Override
	public InetSocketAddress getPeerAddress() throws IOException {

		if (!this.nfd.isValid()) {
			return null;
		}
		int namelen;
		byte[] addr = null;
		int native_fd = this.nfd.getNativeFD();
		byte[] peerAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		namelen = SocketChannelNatives.getPeerAddress(native_fd, peerAddr, peerAddr.length, false);

		//retry native returned with result
		while(namelen == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			namelen = SocketChannelNatives.getPeerAddress(native_fd, peerAddr, peerAddr.length, true);
		}

		if (namelen > 0) {
			addr = new byte[namelen];
			System.arraycopy(peerAddr, 0, addr, 0, namelen);
		}

		int port = SocketChannelNatives.getPeerPort(native_fd, false);

		//retry native returned with result
		while(port == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			port = SocketChannelNatives.getPeerPort(native_fd, true);
		}

		if (port < 0) {
			throw new SocketException(Message.at(NetErrors.getInstance(), port));
		}
		switch (namelen) {
		case 0: // not connected yet
			throw new SocketException(NetErrors.SOCKET_NOT_CONNECTED);
		case 4:
			return new InetSocketAddress(Inet4Address.getByAddress(addr), port);
		case 16:
			return new InetSocketAddress(Inet6Address.getByAddress(addr), port);
		default: {
			if(namelen < 0){
				throw new SocketException(Message.at(NetErrors.getInstance(), namelen));
			}
			throw new SocketException(NetErrors.INVALID_ADDRESS_LENGTH);

		}
		}
	}


	@Override
	protected void checkSockConnection(boolean onlyOnClose)
			throws SocketException {
		if (this.nfd.isClosed()) {
			// closed socket
			throw new SocketException("Socket closed");
		}
		if (!onlyOnClose) {
			// check if the channel is still connected
			InetSocketAddress peer = null;
			try {
				peer = getPeerAddress();
			} catch (IOException e) {

			} finally {
				if (peer == null) {
					// connection reset
					throw new SocketException("Connection reset");
				}
			}
		}
	}

	/**
	 * Create a new socket. This method will initialize the native file
	 * descriptor state of this instance.
	 *
	 * @param stream
	 *            Whether or not to create a streaming socket, or a datagram
	 *            socket.
	 * @throws IOException
	 *             If creating a new socket fails, or if this channel already
	 *             has its native descriptor initialized.
	 */
	@Override
	public void initSocket(boolean stream) throws IOException {
		if (this.nfd.isValid()) {
			throw new IOException("native FD already initialized");
		}

		this.kind = (stream) ? Kind.SOCK_STREAM : Kind.SOCK_DGRAM;

		int fd = SocketChannelNatives.socket(stream, false);

		//retry native returned with result
		while(fd == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			fd = SocketChannelNatives.socket(stream, true);
		}

		if (fd < 0) {
			throw new IOException(Message.at(NetErrors.getInstance(), fd));
		}
		setNativeFD(fd);
	}

	@Override
	public void initServerSocket() throws java.io.IOException {
		if (this.nfd.isValid()) {
			throw new IOException("native FD already initialized");
		}

		int fd = SocketChannelNatives.serverSocket(false);

		//retry native returned with result
		while(fd == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			fd = SocketChannelNatives.serverSocket(true);
		}

		if (fd < 0) {
			throw new SocketException(Message.at(NetErrors.getInstance(), fd));
		}
		setNativeFD(fd);

	}

	@Override
	public void initMulticastSocket() throws IOException {
		if (this.nfd.isValid()) {
			throw new IOException("native FD already initialized");
		}

		this.kind = Kind.SOCK_DGRAM;

		int fd = SocketChannelNatives.multicastSocket(false);

		//retry native returned with result
		while(fd == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			fd = SocketChannelNatives.multicastSocket(true);
		}

		if (fd < 0) {
			throw new IOException(Message.at(NetErrors.getInstance(), fd));
		}
		setNativeFD(fd);
	}
}
