/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2020, IS2T. EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.x509;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.security.auth.x500.X500Principal;

import com.is2t.support.security.NativePublicKey;
import com.is2t.support.security.natives.X509CertSupportNatives;

public class X509CertImpl extends X509Certificate {

	//certificate content
	byte[] data;

	//encoded format
	EncodedFormat format;

	public X509CertImpl(byte[] data) throws CertificateException{
		this.data = data;
		try{
			format = X509CertSupport.parse(this);
		} catch (Exception e) {
			throw new CertificateException("Unable to initialize", e);
		}
	}

	public EncodedFormat getEncodingFormat(){
		return format;
	}

	/**
	 * Gets the public key from this certificate.
	 *
	 * @return the public key.
	 */
	@Override
	public PublicKey getPublicKey() {
		// TODO: global review of certificate parsing and data extraction.
		byte[] keyDataBuffer = new byte[X509CertSupportNatives.getKeyDataSize()];
		int keyDataLength = X509CertSupportNatives.getKeyData(data, data.length, keyDataBuffer,
				keyDataBuffer.length);

		byte[] keyData = new byte[keyDataLength];
		System.arraycopy(keyDataBuffer, 0, keyData, 0, keyData.length);
		PublicKey publicKey = new NativePublicKey(keyData, "Unknown");
		return publicKey;
	}

	public enum EncodedFormat {
		//Base64 encoded ASCII format
		PEM,
		//Binary format
		DER
	}

	@Override
	public byte[] getEncoded() throws CertificateEncodingException {
		return data;
	}

	@Override
	public void verify(PublicKey key) throws CertificateException, NoSuchAlgorithmException, InvalidKeyException,
	NoSuchProviderException, SignatureException {
		X509CertSupport.verify(this, key);
	}

	@Override
	public void verify(PublicKey key, String sigProvider) throws CertificateException, NoSuchAlgorithmException,
	InvalidKeyException, NoSuchProviderException, SignatureException {
		throw new UnsupportedOperationException();
	}

	@Override
	public String toString() {
		return null;
	}

	/**
	 * Extract the subject X500Principal from an X509Certificate. Called from
	 * java.security.cert.X509Certificate.getSubjectX500Principal().
	 */
	public static X500Principal getSubjectX500Principal(X509Certificate cert) {
		try {
			return getX500Principal(cert, false);
		} catch (Exception e) {
			throw new RuntimeException("Could not parse subject", e);
		}
	}

	/**
	 * Extract the issuer X500Principal from an X509Certificate. Called from
	 * java.security.cert.X509Certificate.getIssuerX500Principal().
	 */
	public static X500Principal getIssuerX500Principal(X509Certificate cert) {
		try {
			return getX500Principal(cert, true);
		} catch (Exception e) {
			throw new RuntimeException("Could not parse issuer", e);
		}
	}

	/**
	 * Extract the subject or issuer X500Principal from an X509Certificate. Parses the encoded form of the cert to
	 * preserve the principal's ASN.1 encoding.
	 */
	private static X500Principal getX500Principal(X509Certificate cert, boolean getIssuer) throws Exception {
		byte[] encoded = cert.getEncoded();
		byte[] principalData = new byte[encoded.length];

		int len = X509CertSupportNatives.getX500PrincipalData(encoded, encoded.length, principalData,
				principalData.length,
				getIssuer);

		byte[] principalBytes = new byte[len];
		System.arraycopy(principalData, 0, principalBytes, 0, len);
		return new X500Principal(principalBytes);
	}
}
