/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.vivante.utils;

import static com.verisilicon.vivante.vg_lite_path.VG_LITE_FP32;

import com.verisilicon.vivante.vg_lite_path;

/**
 * Utility class for building {@link vg_lite_path} in FP32 format.
 *
 * <p>
 * See Section <code>2.4 Vector Path Data Opcodes</code> of <code>VGLite® Graphics API v2.07</code>.
 * </p>
 */
public class PathBufferFP32 extends PathBufferBase {
	private static final int COORD_SIZE = 4;

	private float xOffset;

	private float yOffset;

	@Override
	protected int getUnitSize() {
		return COORD_SIZE;
	}

	@Override
	public int getFormat() {
		return VG_LITE_FP32;
	}

	private static int append(byte[] path, int offset, float v) {
		int i = Float.floatToIntBits(v);

		path[offset++] = (byte) (i);
		path[offset++] = (byte) (i >> 8);
		path[offset++] = (byte) (i >> 16);
		path[offset++] = (byte) (i >> 24);

		return offset;
	}

	private void append(int i) {
		this.path[this.index++] = (byte) (i);
		this.path[this.index++] = (byte) (i >> 8);
		this.path[this.index++] = (byte) (i >> 16);
		this.path[this.index++] = (byte) (i >> 24);
	}

	private static int append(byte[] path, int offset, int i) {
		path[offset++] = (byte) (i);
		path[offset++] = (byte) (i >> 8);
		path[offset++] = (byte) (i >> 16);
		path[offset++] = (byte) (i >> 24);

		return offset;
	}

	@Override
	protected void command(int cmd) {
		extend(1);

		append(cmd);
	}

	/**
	 * Adds a command with 1 point parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x
	 *            the horizontal coordinate of the point parameter
	 * @param y
	 *            the vertical coordinate of the point parameter
	 */
	private void command(int cmd, float x, float y) {
		byte[] path = extend(3);

		int offset;

		offset = append(path, this.index, cmd);
		offset = append(path, offset, x);
		offset = append(path, offset, y);

		this.index = offset;
	}

	/**
	 * Adds a command with 2 points parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x1
	 *            the horizontal coordinate of the first point parameter
	 * @param y1
	 *            the vertical coordinate of the first point parameter
	 * @param x2
	 *            the horizontal coordinate of the second point parameter
	 * @param y2
	 *            the vertical coordinate of the second point parameter
	 */
	private void command(int cmd, float x1, float y1, float x2, float y2) {
		byte[] path = extend(5);

		int offset;

		offset = append(path, this.index, cmd);
		offset = append(path, offset, x1);
		offset = append(path, offset, y1);
		offset = append(path, offset, x2);
		offset = append(path, offset, y2);

		this.index = offset;
	}

	/**
	 * Adds a command with 3 points parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x1
	 *            the horizontal coordinate of the first point parameter
	 * @param y1
	 *            the vertical coordinate of the first point parameter
	 * @param x2
	 *            the horizontal coordinate of the second point parameter
	 * @param y2
	 *            the vertical coordinate of the second point parameter
	 * @param x3
	 *            the horizontal coordinate of the third point parameter
	 * @param y3
	 *            the vertical coordinate of the third point parameter
	 */
	private void command(int cmd, float x1, float y1, float x2, float y2, float x3, float y3) {
		byte[] path = extend(7);

		int offset;

		offset = append(path, this.index, cmd);
		offset = append(path, offset, x1);
		offset = append(path, offset, y1);
		offset = append(path, offset, x2);
		offset = append(path, offset, y2);
		offset = append(path, offset, x3);
		offset = append(path, offset, y3);

		this.index = offset;
	}

	/**
	 * Sets an offset that will be used during drawing.
	 *
	 * @param x
	 *            The horizontal offset
	 * @param y
	 *            The vertical offset
	 */
	public void setOffset(float x, float y) {
		this.xOffset = x;
		this.yOffset = y;
	}

	/**
	 * Applies the offset to an horizontal coordinate
	 *
	 * Note: This method is duplicated in <code>PathBufferS8</code>, <code>PathBufferS16</code>,
	 * <code>PathBufferS32</code> and <code>PathBufferFP32</code>.
	 *
	 * @param x
	 *            The horizontal coordinate to offset
	 * @return the updated coordinate
	 */
	private float applyXOffset(float x) {
		return x + this.xOffset;
	}

	/**
	 * Applies the offset to an vertical coordinate
	 *
	 * Note: This method is duplicated in <code>PathBufferS8</code>, <code>PathBufferS16</code>,
	 * <code>PathBufferS32</code> and <code>PathBufferFP32</code>.
	 *
	 * @param y
	 *            The vertical coordinate to offset
	 * @return the updated coordinate
	 */
	private float applyYOffset(float y) {
		return y + this.yOffset;
	}

	/**
	 * Add a move command to the given point.
	 *
	 * The current point is set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param x
	 *            the horizontal coordinate of the point to move to
	 * @param y
	 *            the vertical coordinate of the point to move to
	 */
	public void move(float x, float y) {
		command(CMD_MOVE, applyXOffset(x), applyYOffset(y));
	}

	/**
	 * Add a move command to the given point.
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param x
	 *            the horizontal coordinate of the point to move to
	 * @param y
	 *            the vertical coordinate of the point to move to
	 */
	public void moveRel(float x, float y) {
		command(CMD_MOVE, x, y);
	}

	/**
	 * Add a line command from the current point to the given point
	 *
	 * A line is drawn from a starting point (current point) to an ending point (c, y).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param x
	 *            the horizontal coordinate of the end of the line
	 * @param y
	 *            the vertical coordinate of the end of the line
	 */
	public void line(float x, float y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_LINE, x, y);
	}

	/**
	 * Add a line command from the current point to the given point
	 *
	 * A line is drawn from a starting point (current point) to an ending point (c, y).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param x
	 *            the horizontal coordinate of the end of the line relative to the current point
	 * @param y
	 *            the vertical coordinate of the end of the line relative to the current point
	 */
	public void lineRel(float x, float y) {
		command(CMD_LINE_REL, x, y);
	}

	/**
	 * Add a quadratic curve from the current point to the given point
	 *
	 * A quadratic curve is drawn from a starting point (current point) to an ending point (x, y) using one control
	 * point (cx, cy).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param cx
	 *            the horizontal coordinate of the control point
	 * @param cy
	 *            the vertical coordinate of the control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void quad(float cx, float cy, float x, float y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_QUAD, cx, cy, x, y);
	}

	/**
	 * Add a quadratic curve from the current point to the given point
	 *
	 * A quadratic curve is drawn from a starting point (current point) to an ending point (x, y) using one control
	 * point (cx, cy).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param cx
	 *            the horizontal coordinate of the control point relative to the current point
	 * @param cy
	 *            the vertical coordinate of the control point relative to the current point
	 * @param x
	 *            the horizontal coordinate of the end of the curve relative to the current point
	 * @param y
	 *            the vertical coordinate of the end of the curve relative to the current point
	 */
	public void quadRel(float cx, float cy, float x, float y) {
		command(CMD_QUAD_REL, cx, cy, x, y);
	}

	/**
	 * Add a cubic curve from the current point to the given point
	 *
	 * A cubic curve is drawn from a starting point (current point) to an ending point (x, y) using two control point
	 * (cx1, cy1) and (cx2, cy2).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param cx1
	 *            the horizontal coordinate of the first control point
	 * @param cy1
	 *            the vertical coordinate of the first control point
	 * @param cx2
	 *            the horizontal coordinate of the second control point
	 * @param cy2
	 *            the vertical coordinate of the second control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void cubic(float cx1, float cy1, float cx2, float cy2, float x, float y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_CUBIC, cx1, cy1, cx2, cy2, x, y);
	}

	/**
	 * Add a cubic curve from the current point to the given point
	 *
	 * A cubic curve is drawn from a starting point (current point) to an ending point (x, y) using two control point
	 * (cx1, cy1) and (cx2, cy2).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param cx1
	 *            the horizontal coordinate of the first control point
	 * @param cy1
	 *            the vertical coordinate of the first control point
	 * @param cx2
	 *            the horizontal coordinate of the second control point
	 * @param cy2
	 *            the vertical coordinate of the second control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void cubicRel(float cx1, float cy1, float cx2, float cy2, float x, float y) {
		command(CMD_CUBIC_REL, cx1, cy1, cx2, cy2, x, y);
	}
}
