/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayInputStreamUtils;
import java.io.Closeable;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;

import com.is2t.microbsp.microui.natives.NSystemDisplay;
import com.is2t.microbsp.microui.natives.NTools;
import com.is2t.vm.support.io.MemoryInputStream;

import ej.annotation.Nullable;
import ej.microui.Log;
import ej.microui.MicroUI;
import ej.microui.MicroUIException;

/**
 * microUI-API
 */
public class ResourceImage extends Image implements Closeable {

	/**
	 * microUI-API
	 */
	/*
	 * Same constants in Format.iceTea
	 */
	public enum OutputFormat {
		/**
		 * microUI-API
		 */
		ARGB8888(2),
		/**
		 * microUI-API
		 */
		RGB888(3),
		/**
		 * microUI-API
		 */
		RGB565(4),
		/**
		 * microUI-API
		 */
		ARGB1555(5),
		/**
		 * microUI-API
		 */
		ARGB4444(6),
		/**
		 * microUI-API
		 */
		A8(8);

		private final int val;

		OutputFormat(int val) {
			this.val = val;
		}
	}

	/**
	 *
	 */
	ResourceImage(byte[] sd, int logOpen, int logType) {
		super(sd, logOpen, logType);
	}

	/**
	 * microUI-API
	 *
	 * @param name
	 *            microUI-API
	 * @return microUI-API
	 */
	public static boolean canLoadImage(String name) {
		assert name != null;
		name = checkImagePrerequisites(name);
		return canLoadImage(name, false, false) || canLoadImage(name, true, false) || canLoadImage(name, false, true)
				|| canLoadImage(name, true, true);
	}

	/**
	 * microUI-API
	 *
	 * @param name
	 *            microUI-API
	 * @return microUI-API
	 */
	public static ResourceImage loadImage(String name) {
		assert name != null;
		return loadImage(name, NSystemDisplay.DEFAULT_FORMAT);
	}

	/**
	 * microUI-API
	 *
	 * @param name
	 *            microUI-API
	 * @param format
	 *            microUI-API
	 * @return microUI-API
	 */
	public static ResourceImage loadImage(String name, @Nullable OutputFormat format) {
		assert name != null;
		return loadImage(name, getFormat(format));
	}

	private static ResourceImage loadImage(String name, int format) {

		logCreateImageStart(Log.TYPE_OPEN_IMAGE_FROM_PATH);
		name = checkImagePrerequisites(name);

		byte[] sd = SImageMetadata.allocate(false);

		// 1- try to load an internal raw image
		ResourceImage image = loadImage(name, sd, format, false, false, Log.IMAGE_CARACTERISTICS_INTERNAL_RAW);
		if (image != null) {
			// done!
			return image;
		}

		// 2- try to load an internal dynamic image
		image = loadImage(name, sd, format, true, false, Log.IMAGE_CARACTERISTICS_INTERNAL_DYNAMIC);
		if (image != null) {
			// done!
			return image;
		}

		// 3- try to load an external raw image
		image = loadImage(name, sd, format, false, true, Log.IMAGE_CARACTERISTICS_EXTERNAL_RAW);
		if (image != null) {
			// done!
			return image;
		}

		// 4- try to load an external dynamic image
		image = loadImage(name, sd, format, true, true, Log.IMAGE_CARACTERISTICS_EXTERNAL_DYNAMIC);
		if (image != null) {
			// done!
			return image;
		}

		throw new MicroUIException(MicroUIException.RESOURCE_INVALID_PATH, name);
	}

	private static String getImagePath(String name, boolean dynamic, boolean externalResource) {
		String fullname = dynamic ? name : NTools.getRawImagePath(name);
		return externalResource ? NTools.getExternalImagePath(fullname) : fullname;
	}

	private static boolean canLoadImage(String name, boolean dynamic, boolean externalResource) {
		String fullname = getImagePath(name, dynamic, externalResource);
		return NSystemDisplay.NO_ERROR == NSystemDisplay.loadImageFromPath(fullname.getBytes(), fullname.length(),
				dynamic, externalResource, false, NSystemDisplay.DEFAULT_FORMAT, null);
	}

	private static @Nullable ResourceImage loadImage(String name, byte[] sd, int format, boolean dynamic,
			boolean externalResource, int logType) {

		String fullname = getImagePath(name, dynamic, externalResource);
		int errorCode = NSystemDisplay.loadImageFromPath(fullname.getBytes(), fullname.length(), dynamic,
				externalResource, false, format, sd);

		switch (errorCode) {
		case NSystemDisplay.NO_ERROR:
			return new ResourceImage(sd, Log.TYPE_OPEN_IMAGE_FROM_PATH, logType);
		case NSystemDisplay.INVALID_PATH:
			// let caller manage the error
			return null;
		default:
			// file found but native exception: throw exception
			NTools.checkError(errorCode, fullname);
			return null; // never here
		}
	}

	/**
	 * microUI-API
	 *
	 * @param stream
	 *            microUI-API
	 * @param size
	 *            microUI-API
	 * @return microUI-API
	 */
	public static ResourceImage loadImage(InputStream stream, int size) {
		assert stream != null;
		return loadImage(stream, size, NSystemDisplay.DEFAULT_FORMAT);
	}

	/**
	 * microUI-API
	 *
	 * @param stream
	 *            microUI-API
	 * @param size
	 *            microUI-API
	 * @param format
	 *            microUI-API
	 * @return microUI-API
	 */
	public static ResourceImage loadImage(InputStream stream, int size, @Nullable OutputFormat format) {
		assert stream != null;
		return loadImage(stream, size, getFormat(format));
	}

	private static ResourceImage loadImage(InputStream stream, final int size, int format) {

		MicroUI.checkRunning();
		checkSecurityManager();
		logCreateImageStart(Log.TYPE_OPEN_IMAGE_FROM_INPUTSTREAM);

		byte[] sd = SImageMetadata.allocate(false);
		int logType;
		int errorCode;

		try {

			if (stream instanceof MemoryInputStream) {
				// On emb, a memory input stream targets a fixed memory region (immutable). On
				// sim, this notion does not exist. MemoryInputStream is available in
				// edc-internal. On sim, no object is an instance of MemoryInputStream. But if
				// in the future this use case appear, HIL protocol must be updated to authorize
				// a MemoryInputStream as a field

				logType = Log.IMAGE_CARACTERISTICS_MEMORYINPUTSTREAM;
				errorCode = NSystemDisplay.loadImageFromMIS((MemoryInputStream) stream, size, format, sd);

			} else if (stream instanceof ByteArrayInputStream) {
				// we can directly use the byte array

				byte[] data = ByteArrayInputStreamUtils.getByteArray((ByteArrayInputStream) stream);
				int offset = ByteArrayInputStreamUtils.getOffset((ByteArrayInputStream) stream);
				stream.skip(size); // skip bytes read in native side, can throw an IOE

				logType = Log.IMAGE_CARACTERISTICS_BYTEARRAYINPUTSTREAM;
				errorCode = NSystemDisplay.loadImageFromBytes(data, offset, size, format, sd);

			} else {
				// we have to read the stream

				DataInputStream dis = stream instanceof DataInputStream ? (DataInputStream) stream
						: new DataInputStream(stream);
				byte[] data = new byte[size];
				try {
					dis.readFully(data); // can throw an IOE
				} finally {
					if (!(stream instanceof DataInputStream)) {
						// have to close the temporary dis
						dis.close();
					}
				}

				logType = Log.IMAGE_CARACTERISTICS_GENERICINPUTSTREAM;
				errorCode = NSystemDisplay.loadImageFromBytes(data, 0, size, format, sd);
			}
		} catch (IOException e) {
			throw new MicroUIException(MicroUIException.RESOURCE_INVALID_FILE);
		}

		NTools.checkError(errorCode, null);
		return new ResourceImage(sd, Log.TYPE_OPEN_IMAGE_FROM_INPUTSTREAM, logType);
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public boolean isClosed() {
		synchronized (this) {
			return this.sd == null;
		}
	}

	@Override
	public void close() {
		synchronized (this) {
			byte[] data = this.sd;
			if (data != null) {
				this.sd = null;
				NSystemDisplay.closeOffScreen(data);
			}
		}
	}

	/**
	 * Return integer format from {@link OutputFormat}
	 */
	private static int getFormat(@Nullable OutputFormat format) {
		return format == null ? NSystemDisplay.DEFAULT_FORMAT : format.val;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	@Override
	public byte[] getSNIContext() {
		if (this.sd == null) {
			// image has been closed
			throw new MicroUIException(MicroUIException.RESOURCE_CLOSED);
		}
		return this.sd;
	}
}
