/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector.font;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.font.GlyphVector;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;

import com.verisilicon.vivante.Matrix;

import ej.microui.display.LLUIDisplay;
import ej.microui.display.LLUIPainter.MicroUIGraphicsContext;

/**
 * Vector Font natives implementation on simulator.
 */
public class VectorFontPainter {

	private static final Point2D ORIGIN = new Point2D.Float(0, 0);

	/**
	 * Draws a character horizontally.
	 *
	 * @param gcArray
	 *            the graphic context
	 * @param fontReference
	 *            the reference of the font used to render the string
	 * @param size
	 *            the height of the characters to draw in pixels
	 * @param character
	 *            the character to draw.
	 * @param matrix
	 *            affine transformation matrix to apply to the glyph, originated on top-left anchor point of the
	 *            character.
	 * @param alpha
	 *            the opacity, 0xFF meaning fully opaque
	 * @param gradientHandle
	 *            a handle to the gradient to apply, 0 meaning that gradient is not used. In this case, use color.
	 */
	public static void drawCharNative(byte[] gcArray, int fontReference, float size, char character, float[] matrix,
			int alpha, int gradientHandle) {
		Font font = VectorFont.get(fontReference, size);

		LLUIDisplay engine = LLUIDisplay.Instance;
		synchronized (engine) {
			MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(gcArray);
			if (gc.requestDrawing()) {

				BufferedImage bi = (BufferedImage) gc.getImage().getRAWImage();
				Graphics2D g2d = (Graphics2D) bi.getGraphics();

				final Paint previousPaint = setColor(gc, g2d, alpha, gradientHandle);

				g2d.setClip(gc.getClipX1(), gc.getClipY1(), gc.getClipX2() - gc.getClipX1(),
						gc.getClipY2() - gc.getClipY1());

				g2d.setFont(font);
				g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
				g2d.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
				NativeUtils.refreshContent(matrix, 0, matrix.length);

				final AffineTransform transform = new AffineTransform(
						new float[] { matrix[0], matrix[3], matrix[1], matrix[4], matrix[2], matrix[5] });

				// Translate the glyph to baseline position.
				transform.translate(0, VectorFont.getBaselinePositionNative(fontReference, size));
				final GlyphVector glyphVector = font.createGlyphVector(g2d.getFontRenderContext(),
						new char[] { character });

				Shape glyph = glyphVector.getGlyphOutline(0);
				glyph = transform.createTransformedShape(glyph);
				g2d.fill(glyph);

				final Rectangle2D bounds = glyph.getBounds2D();
				gc.setDrawingLimits((int) bounds.getMinX(), (int) bounds.getMinY(),
						(int) (bounds.getMinX() + bounds.getWidth() - 1),
						(int) (bounds.getMinY() + bounds.getHeight() - 1));
				g2d.setPaint(previousPaint);

				// convert colors in drawn area
				convertColorToColorToDraw(bi, (int) bounds.getMinX(), (int) bounds.getMinY(),
						(int) bounds.getWidth(), (int) bounds.getHeight());
			}
		}
	}

	/**
	 * Draws a string horizontally.
	 *
	 * @param gcArray
	 *            the graphic context
	 * @param fontReference
	 *            the reference of the font used to render the string
	 * @param size
	 *            the height of the characters to draw in pixels
	 * @param string
	 *            the array of characters containing the string to draw.
	 * @param stringOffset
	 *            the index of the first character of the string in the array
	 * @param stringLength
	 *            the length of the string
	 * @param x
	 *            the top left coordinate of the characters string on x axis
	 * @param y
	 *            the top left coordinate of the characters string on y axis
	 * @param alpha
	 *            the opacity, 0xFF meaning fully opaque
	 * @param gradientHandle
	 *            a handle to the gradient to apply, 0 meaning that gradient is not used. In this case, use color.
	 */
	public static void drawStringNative(byte[] gcArray, int fontReference, float size, char[] string, int stringOffset,
			int stringLength, int x, int y, int alpha, int gradientHandle) {
		Font font = VectorFont.get(fontReference, size);

		LLUIDisplay engine = LLUIDisplay.Instance;
		synchronized (engine) {
			MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(gcArray);
			if (gc.requestDrawing()) {

				BufferedImage bi = (BufferedImage) gc.getImage().getRAWImage();
				Graphics2D g2d = (Graphics2D) bi.getGraphics();

				final Paint previousPaint = setColor(gc, g2d, alpha, gradientHandle);

				g2d.setClip(gc.getClipX1(), gc.getClipY1(), gc.getClipX2() - gc.getClipX1(),
						gc.getClipY2() - gc.getClipY1());

				g2d.setFont(font);
				g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
				g2d.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
				NativeUtils.refreshContent(string, stringOffset, stringLength);

				final FontMetrics metrics = g2d.getFontMetrics();

				int advance_y = VectorFont.getBaselinePositionNative(fontReference, size);

				g2d.drawChars(string, stringOffset, stringLength, x, y + advance_y);

				gc.setDrawingLimits(x, y + Math.abs(metrics.getMaxDescent()),
						x + metrics.charsWidth(string, stringOffset, stringLength),
						y + Math.abs(metrics.getMaxDescent() + advance_y));
				g2d.setPaint(previousPaint);

				// convert colors in drawn area
				convertColorToColorToDraw(bi, x, y, metrics.charsWidth(string, stringOffset, stringLength),
						metrics.getHeight());
			}
		}

	}

	/**
	 * Draws a string on an arc of a circle.
	 *
	 * @param gcArray
	 *            the graphic context
	 * @param fontReference
	 *            the reference of the font used to render the string
	 * @param size
	 *            the height of the characters to draw in pixels
	 * @param string
	 *            the array of characters containing the string to draw.
	 * @param stringOffset
	 *            the index of the first character of the string in the array
	 * @param stringLength
	 *            the length of the string
	 * @param x
	 *            the top left coordinate of the circle on x axis
	 * @param y
	 *            the top left coordinate of the circle on y axis
	 * @param diameter
	 *            the diameter of the circle
	 * @param angleDegrees
	 *            the angle in degrees at which the first character of the string is drawn on the circle
	 * @param alpha
	 *            the opacity, 0xFF meaning fully opaque
	 * @param gradientHandle
	 *            a handle to the gradient to apply, 0 meaning that gradient is not used. In this case, use color.
	 * @param out
	 *            a flag to set the direction of the text
	 */
	public static void drawStringArcNative(byte[] gcArray, int fontReference, float size, char[] string,
			int stringOffset, int stringLength, int x, int y, int diameter, float angleDegrees, int alpha,
			int gradientHandle, boolean out) {

		Font font = VectorFont.get(fontReference, size);
		LLUIDisplay engine = LLUIDisplay.Instance;

		synchronized (engine) {
			MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(gcArray);

			if (gc.requestDrawing()) {

				BufferedImage bi = (BufferedImage) gc.getImage().getRAWImage();
				Graphics2D g2d = (Graphics2D) bi.getGraphics();

				final Paint previousPaint = setColor(gc, g2d, alpha, gradientHandle);

				g2d.setClip(gc.getClipX1(), gc.getClipY1(), gc.getClipX2() - gc.getClipX1(),
						gc.getClipY2() - gc.getClipY1());

				g2d.setFont(font);
				g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
				g2d.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);

				NativeUtils.refreshContent(string, stringOffset, stringLength);

				int radius = diameter / 2;
				int centerX = x + radius;
				int centerY = y + radius;

				// Fix start angle
				angleDegrees = ((out ? -90 : 90) - angleDegrees);

				final FontMetrics metrics = g2d.getFontMetrics();
				int ascent = metrics.getAscent();
				ascent -= metrics.getLeading() / 2;

				int glyphHeight = metrics.getHeight() - metrics.getLeading();

				float advanceY = out ? radius - ((glyphHeight - ascent) * size / glyphHeight)
						: (ascent * size / glyphHeight) - radius;

				final Matrix matrix = new Matrix();
				final GlyphVector glyphVector = font.createGlyphVector(g2d.getFontRenderContext(), string);

				for (int glyphIndex = stringOffset; glyphIndex < stringOffset + stringLength; glyphIndex++) {
					final double charWidth = metrics.charWidth(string[glyphIndex]);

					// reset transformations
					matrix.identity();

					// set the origin to the center of the circle
					matrix.translate(centerX, centerY);

					// set the angle of the glyph
					matrix.rotate(angleDegrees);

					// move the glyph to the arc
					matrix.translate(0, advanceY);

					// adjust baseline angle to the tangent
					float adjustAngle = (float) Math.toDegrees(Math.atan(charWidth / diameter));
					matrix.rotate(out ? -adjustAngle : adjustAngle);

					// get glyph shape and apply transformations
					glyphVector.setGlyphPosition(glyphIndex, ORIGIN);
					Shape glyph = glyphVector.getGlyphOutline(glyphIndex);
					glyph = matrix.transform(glyph);

					g2d.fill(glyph);

					final Rectangle2D bounds = glyph.getBounds2D();

					// convert colors in drawn area
					convertColorToColorToDraw(bi, (int) bounds.getMinX(), (int) bounds.getMinY(),
							(int) bounds.getWidth(), (int) bounds.getHeight());

					angleDegrees = updateAngle(charWidth, radius, glyphHeight, out, angleDegrees);
				}

				gc.setDrawingLimits(x - radius, y - radius, x + radius, y + radius);
				g2d.setPaint(previousPaint);
			}

		}
	}

	private static Paint setColor(MicroUIGraphicsContext gc, Graphics2D g2d, int alpha, int gradientHandle) {
		final Paint previousPaint = g2d.getPaint();
		if (!LinearGradient.setPaint(gradientHandle, g2d)) {
			// Cannot apply gradient (may be undefined), use color instead
			final Color color = new Color(gc.getRenderingColor());
			g2d.setColor(new Color(color.getRed(), color.getGreen(), color.getBlue(), alpha));
		}
		return previousPaint;
	}

	/**
	 * Converts all pixels colors to Display Color in given region.
	 */
	private static void convertColorToColorToDraw(BufferedImage bi, int x, int y, int width, int height) {

		// (width, height) represents first outside pixel
		width += x;
		height += y;

		// sanity check
		x = Math.max(x, 0);
		y = Math.max(y, 0);
		width = Math.min(width, bi.getWidth());
		height = Math.min(height, bi.getHeight());

		for (int i = x; i < width; i++) {
			for (int j = y; j < height; j++) {
				bi.setRGB(i, j, LLUIDisplay.Instance.convertARGBColorToColorToDraw(bi.getRGB(i, j)));
			}
		}
	}

	private static float updateAngle(double advance, int radius, int glyphHeight, boolean out, float angleDegrees) {
		if (out) {
			return angleDegrees - (float) (Math.toDegrees(advance / (radius - glyphHeight)));
		} else {
			return angleDegrees + (float) (Math.toDegrees(advance / (radius - glyphHeight)));
		}
	}

}
